/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#ifndef _PARTICLE_H
#define _PARTICLE_H

#define PG_STATUS_ALIVE 0
#define PG_STATUS_DEAD  1

#define ROBOT_DATA 0
#define ROOM_DATA 1
#define ROBOT_ROOM_DATA 2


/** 
 \typedef double weight_type
 \brief 重みの型
*/
typedef double weight_type; 

/**
   \struct state_type 
   \brief 1つのパーティクルが持つ内部状態のデータ
*/
typedef struct{ 
  double x; 	///<位置(x)
  double y;	///<位置(y)
  double z;	///<位置(z) add by morio 3D拡張
  double v;	///<移動速度
  double theta;	///<ロボットから見たパーティクルの角度(0°から)
  double phi;	///<ロボットから見たパーティクルの角度(0°から) add by morio 3D拡張
  double orient; ///<室内マイクから見たパーティクルの進行方向

} state_type;

/**
   \struct history_data 
   \brief 履歴データ(出力)
*/
typedef struct{ 
  int Frame_ID;		///<処理回数
  double x;		///<位置(x)
  double y;		///<位置(y)
  double z;		///<位置(z)
  double v;		///<速度
  double theta;		///<パーティクルの角度
  double phi;		///<パーティクルの角度 add by Morio
} history_data;

/**
   \struct particle_group
   \brief １つのパーティクルグループの状態の情報を持つ
*/
typedef struct{ 
  int pg_status;		///< 状態 0:ALIVE(生存), 1:DEAD(消去)
  state_type *state;		///<内部状態
  weight_type *weight;		///< 重み(内部状態数と同じ数),各グループごとに合計が１になるような正規化がされる。
   history_data *h;		///< 履歴データ
  int histnum;			///< 履歴数
  int histmax;			///< 履歴データの領域確保済み個数
  int target;			///< 対応づけられている観測値の番号
  int r_target;			///< 対応づけられている観測値の番号(rob)
  int update_FrameID ;  	///< 最後に観測が得られたフレーム番号   
  int index;			///< パーティクルグループ番号(unique)        
    //TFIndex 2016.9.26
  int tfindex[3];		///< TFIndex  
} particle_group;

/**
   \struct particle_group_set
   \brief パーティクルグループのある時刻での状態を表す
*/
typedef struct{ 
  particle_group *pg;		///<パーティクルグループ型
  int pg_num;			///< パーティクルグループの個数。音源数に相当する
  int pg_alive_num;		///< 現時点で生存しているグループの数 
} particle_group_set;


/**
   \struct observation_type
   \brief 観測値(IRMA)のデータ（音源１つ分）
*/
typedef struct{ 
  	double x;		///<音源位置
  	double y;		///<音源位置
	double z;		///<音源位置 add by Morio 3D拡張
	//double theta;	//使用しない　　///<音源の進行方向(未使用)
} observation_type;


/**
   \struct observation_set
   \brief ある時点での観測値の集合(IRMA)
*/
typedef struct{ 
	observation_type *ob;		///<観測値データ(IRMA)
  int ob_num;				///<観測値(音源)の数
  int ob_FrameID;			///< フレーム番号
} observation_set;

/*
typedef struct{ 
  double pow[27][15];
  double time; 
} rmap_ob;

typedef struct{ 
  double pow[72];
  double time; 
} bmap_ob;
/*

/**
   \brief 観測値(REMA)のデータ(音源１つ分)
*/
typedef struct{
  double x;			///<ロボットの位置(x座標) 
  double y;			///<ロボットの位置(y座標)
  double z;			///<ロボットの位置(y座標)
  
  double r_theta;		///<ロボットの頭部方向 (2008/10/03追加)
  double r_phi;			///<ロボットの頭部方向 (2016/9/21追加)
  double theta;			///<定位結果の方向(deg)
  double phi;			///<定位結果の方向(deg)
  double likelihood;		//音源のパワーにより決まるこの定位結果に対する尤度
   
  //TFIndex 2016.9.26
  int tfindex[3];		///< TFIndex  
} robot_observation_type;

/**
   \struct robot_observation_set
   \brief ある時点での観測値の集合(REMA)
*/
typedef struct{ 
  robot_observation_type *r_ob;			///<観測値データ(REMA)
  int ob_num;					///<観測値(音源)の数
  int ob_FrameID;				///<フレーム番号
} robot_observation_set;

/**
   \struct PARAM
   \brief harkmwからのパラメータをまとめる構造体
*/
typedef struct{
	int m_param_particle_num;			///<総パーティクル数
	int m_param_max_num_of_source;			///<音源数の最大値
	double m_param_th_another_source;		///<対応付けの閾値
	double m_param_th_ignore_source;		///<新音源生成・消滅の判断のための閾値
	int m_param_th_remove_source;			///<パーティクルグループ削除までのフレーム数
	double m_param_th_output_range;			///<平均からth以内の距離（または角度）にあるパーティクルを使用する
	double m_param_likelihood_sigma;		///<観測値の尤度を計算するときの分散(room)
	double m_param_likelihood_sigma_robot;		///<観測値の尤度を計算するときの分散(room)
	double m_param_likelihood_robot_weight;		///<robに対する重みの計算
	double m_param_th_random_walk;			///<RandomWalk遷移のための閾値
	double m_param_state_update_sigma;		///<内部状態の遷移に加えられるノイズの分散
	double m_param_sum_weight;			///<近傍の平均　or 全体の平均　の閾値
	double m_param_rob_pow;				///<入力データを音源と判断するかどうかの閾値(room)　(未使用)
	double m_param_room_length;			///<部屋のサイズ[m]（未使用）
	double m_param_room_width;			///<部屋のサイズ[m]（未使用）
	double m_param_room_pow;			///<入力データを音源と判断するかどうかの閾値(rob)　（未使用）
	int m_param_history;
} PARAM;


//20081014追加
#ifdef __cplusplus
extern "C" {
#endif

/** 
 * \brief 新しい音源や、途中で音源が消えるときへの対応, 結構重要パーティクルフィルタの枠組みには入っていない泥臭い部分
 * \param flag 入力データの種類によるフラグ(IRMA or REMA or IRMA+REMA)
 * \param Pnum パーティクル総数
 * \param pg_set パーティクルグループ集合
 * \param ob_set 観測値集合
 * \param r_set ロボット観測値集合
 * \param s_param パラメータ設定
 */
void check_source(const int flag, int Pnum, particle_group_set *pg_set, observation_set *ob_set, robot_observation_set *r_set, const PARAM s_param);

//void check_source_rmap(int Pnum, particle_group_set *pg_set, rmap_ob *new_rmap); 



// * \fn void group_generate(const int flag, int Pnum, particle_group_set *pg_set, observation_set *ob_set, robot_observation_set *r_set, int *used, int rest, const PARAM s_param);

/** 
 * \brief グループを増やす場合の処理
 * \param flag 入力データの種類によるフラグ(IRMA or REMA or IRMA+REMA)
 * \param Pnum パーティクル数
 * \param pg_set パーティクルグループ集合
 * \param ob_set 観測値集合
 * \param r_set ロボット観測値集合
 * \param used 観測値対応付リスト
 * \param rest 未対応観測値数
 * \param s_param パラメータ設定
 */
void group_generate(const int flag, int Pnum, particle_group_set *pg_set, observation_set *ob_set, robot_observation_set *r_set, int *used, int rest, const PARAM s_param);

//void group_generate_rmap(int Pnum, particle_group_set *pg_set, rmap_ob *rmap, int rmap_used[27][15]);
// * \fn void group_delete(int Pnum, particle_group_set *pg_set,  const int *group_used, int ob_FrameID, const PARAM s_param);

/** 
 * \brief グループを減らす場合の処理: 追跡している粒子グループが残った → 一定フレーム以上ならグループを削除
 * \param Pnum パーティクル数
 * \param pg_set パーティクルグループ集合
 * \param group_used 観測値対応付リスト
 * \param ob_FrameID フレーム番号
 * \param s_param パラメータ設定
 */
void group_delete(int Pnum, particle_group_set *pg_set,  const int *group_used, int ob_FrameID, const PARAM s_param);


	// * \fn double state_norm(state_type s1, state_type s2);

/**
 * \brief 粒子間の距離を算出
 * \param s1 状態1(パーティクルグループの位置)
 * \param s2 状態2(１つのパーティクルの位置)
 * \return 粒子間の角度の差
*/
double state_norm(state_type s1, state_type s2);

	// * \fn double state_norm(state_type s1, state_type s2);

/**
 * \brief 粒子間の角度の差を算出
 * \param s1 状態1(パーティクルグループの角度)
 * \param s2 状態2(１つのパーティクルの角度)
 * \return 粒子間の角度の差
 */
double state_diff_angle(state_type s1, state_type s2);

	 //* \fn void state_output(const int flag, int Pnum, const particle_group_set *pg_set, int Frame_ID, const PARAM s_param);

/** 
 * \brief パーティクルの状態(結果)の出力
 * \param flag 入力データの種類によるフラグ(IRMA or REMA or IRMA+REMA)
 * \param Pnum パーティクル総数
 * \param pg_set パーティクルグループ集合
 * \param Frame_ID フレーム番号
 * \param s_param パラメータ構造体
*/
void state_output(const int flag, int Pnum, const particle_group_set *pg_set, int Frame_ID, const PARAM s_param);
//void state_output_rmap(FILE *output, int Pnum,  const particle_group_set *pg_set, double time);


	// * \fn int is_in_sink(double x, double y)

///** 
// * \brief 障害物があるか確かめる
// * \param x X座標
// * \param y YX座標
// */
//int is_in_sink(double x, double y);


	// * \fn state_type state_update(const int flag, state_type old_state, int time_diff, const PARAM s_param)

/** 
 * \brief 内部状態を遷移させる 
 * \param flag 入力データの種類によるフラグ(IRMA or REMA or IRMA+REMA)
 * \param old_state  現在の内部状態
 * \param time_diff  前回更新時からの時刻
 * \param s_param パラメータ構造体
 * \return 更新された状態
 */
state_type state_update(const int flag, state_type old_state, int time_diff, const PARAM s_param);
//state_type state_update_rmap(state_type old_state, double time_diff);


	// * \fn double likelihood(const state_type *state, const observation_type *ob,  const robot_observation_type *r_ob, const PARAM s_param);

/** 
 * \brief stateに対するobの尤度を計算 結構重要
 * \param state 内部状態
 * \param ob 観測値
 * \param r_ob ロボット観測値
 * \param s_param パラメータ構造体
 * \return stateに対するobの尤度
*/
double likelihood(const state_type *state, const observation_type *ob,  const robot_observation_type *r_ob, const PARAM s_param);

//double likelihood_rmap(const state_type *state, rmap_ob *rmap,  double x, double y);


	// * \fn weight_type weight_update(const state_type *state, const observation_type *ob, weight_type old_weight, const robot_observation_type *r_ob, const PARAM s_param)

/**
 * \brief 重みの更新
 * \param state 更新した直後の状態
 * \param ob 対応付された観測値
 * \param old_weight 現在の重み
 * \param r_ob 対応付されたロボット観測値
 * \param s_param パラメータ構造体
 * \return 更新した重み
 */
weight_type weight_update(const state_type *state, const observation_type *ob, weight_type old_weight, const robot_observation_type *r_ob, const PARAM s_param);
//weight_type weight_update_no_target(const state_type *state, const double h_theta,  const observation_type *ob, weight_type old_weight,  const robot_observation_type *r_ob, const PARAM s_param);
//weight_type weight_update_rmap( const state_type *state, rmap_ob *rmap, weight_type old_weight ); 


// * \fn double randn(double mu, double sigma)

/** 
 *  \brief ガウスノイズを発生させる
 *  \param mu 平均
 *  \param sigma 分散
 *  \return ランダムな値(0〜1)
 */
double randn(double mu, double sigma);


	// * \fn void importance_sampling(const int flag, int Pnum, particle_group_set *pg_set, observation_set *ob_set, robot_observation_set *r_set, const PARAM s_param)

/** 
 * \brief 内部状態の更新と重要度の更新
 * \param flag 入力データの種類によるフラグ(IRMA or REMA or IRMA+REMA)
 * \param Pnum 粒子の総数
 * \param pg_set 粒子グループ集合
 * \param ob_set 観測値集合
 * \param r_set  ロボット観測値集合
 * \param s_param パラメータ構造体
 */
void importance_sampling(const int flag, int Pnum, particle_group_set *pg_set, observation_set *ob_set, robot_observation_set *r_set, const PARAM s_param);
//void importance_sampling_rmap(int Pnum, particle_group_set *pg_set, rmap_ob *new_rmap); 


	// * \fn void selection(int Pnum, particle_group_set *pg_set)

/** 
 * \brief パーティクルの増減(グループ内)
 * \param Pnum  全パーティクル数
 * \param pg_set  パーティクルグループ集合
 * \param s_param パラメータ構造体
 */
void selection(int Pnum, particle_group_set *pg_set, const PARAM s_param);


	// * \fn void SIR(int old_Pnum, int new_Pnum, const state_type *state, const weight_type *weight, state_type *new_state, weight_type *new_weight, const PARAM s_param)

/** 
 * \brief Sampling Importance Resampling
 * \param old_Pnum  グループ内パーティクル数
 * \param new_Pnum 
 * \param state 
 * \param weight 
 * \param new_state 
 * \param new_weight 
 * \param s_param パラメータ構造体
 */
void SIR(int old_Pnum, int new_Pnum, const state_type *state, const weight_type *weight, state_type *new_state, weight_type *new_weight, const PARAM s_param);


/** 
 * \brief count=0ならgroup_numを0に初期化
 */
void init_group_num(void); //count=0ならgroup_numを0に初期化

void print_observation_set(observation_set *os);

void print_robot_observation_set(robot_observation_set *os);

void print_state(int Pnum, particle_group_set *pg_set, double time);

#ifdef __cplusplus
}
#endif

#if 0
int get_data(FILE *input, FILE *robot_input, 
	     observation_set *ob_set, robot_observation_set *r_set);
#endif

#endif
