/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include "BF.h"

#include <stdio.h>
#include <map>
#include <iostream>

namespace BF
{
    const double BF::BF_EPS = 1e-10;
    
    BF::BF(int _numFreq, int _lowerfreq, int _upperfreq, int _samplingRate, double _soundSpeed, UtUpdateType _separationMatUpdateType) : 
    numFreq(_numFreq), 
    samplingRate(_samplingRate), 
    soundSpeed(_soundSpeed), 
    transformFunctionDatabase(NULL), 
    separationMatricesDatabase(NULL),
    lowerfreq(_lowerfreq),
    upperfreq(_upperfreq),
    separationMatUpdateType(_separationMatUpdateType)
    {
    }

    BF::~BF()
    {
	if (this->transformFunctionDatabase){
		delete this->transformFunctionDatabase;
		this->transformFunctionDatabase = NULL;
	}
	if (this->separationMatricesDatabase){
		delete this->separationMatricesDatabase;
		this->separationMatricesDatabase = NULL;
	}
    }

    void
    BF::getTransformFunctions(const std::vector<SourceLocation> &sourcePositions, std::vector<Matrix> &transformFunctions) const
    {
		int numSrc = sourcePositions.size();
		std::vector<Matrix> newTransformFunctions;
		std::vector<SourceLocation> newSourcePositions;


	    /* 伝達関数をデータベースから取得する場合、マイク数はデータベースから取得する。 */
	    int numMic = this->transformFunctionDatabase->micLocations.size();

	    /* 伝達関数をデータベースから取得するが、まずは取得結果を覚める領域を確保 */
	    for(int i = 0; i < this->numFreq; i++){
			newTransformFunctions.push_back(Matrix(numMic, numSrc));
	    }
	    /* 続いて、入力の音源・マイクの組を、データベース上どの音源・マイクの組にマップするかを検索 */
	    std::map<int, int> srcMap, micMap;
	    for(int i = 0; i < numSrc; i++){
			int srcIndex = this->transformFunctionDatabase->findSourceIndex(sourcePositions[i]);
			srcMap.insert(std::pair<int, int>(i, srcIndex));
	    }

	    for(int i = 0; i < this->numFreq; i++){
			Matrix &transformFunction = newTransformFunctions[i];
			for(int j = 0; j < numSrc; j++){
			    int jj = srcMap[j];
			    transformFunction.setColumn(j, this->transformFunctionDatabase->Data[i].getColumn(srcMap[j]));
			}
	    }

		/* 結果の格納 */
		newTransformFunctions.swap(transformFunctions);
    }

    void
    BF::executeBeamForming(const Matrix &inputSpectrum, Matrix &outputSpectrum)
    {
	Matrix newInputSpectrum (inputSpectrum);
	Matrix newOutputSpectrum(this->separationMatrices[0].getRowCount(), this->numFreq);

//	for(int i = 0; i < this->numFreq; i++){
	for(int i = lowerfreq; i <= this->upperfreq; i++){
//	    this->separationMatrices[i].dump(stdout, "this->separationMatrices[%d]", i);
//	    inputSpectrum.getColumn(i).dump(stdout, "inputSpectrum.getColumn(%d)", i);
//	    this->separationMatrices[i].multiplied(inputSpectrum.getColumn(i)).dump(stdout, "this->separationMatrices[%d].multiplied(inputSpectrum.getColumn(%d))", i, i);
	    newOutputSpectrum.setColumn(i, this->separationMatrices[i].multiplied(inputSpectrum.getColumn(i)));
	}

//	inputSpectrum.dump(stdout, "in");
//	separationMatrices[0].dump(stdout, "G[0]");
//	newOutputSpectrum.dump(stdout, "out");

	outputSpectrum = newOutputSpectrum;
	newInputSpectrum.swap(this->inputSpectrum);
	newOutputSpectrum.swap(this->outputSpectrum);
    }

    void
    BF::setTransformFunctionDatabase(const TransformFunctionDatabase &transformFunctionDatabase)
    {
	using std::swap;
	TransformFunctionDatabase *newTransformFunctionDatabase = new TransformFunctionDatabase(transformFunctionDatabase);
	swap(this->transformFunctionDatabase, newTransformFunctionDatabase);
	delete newTransformFunctionDatabase;

	// マイク座標の登録
	this->micPositions.clear();
	for (int i = 0 ; i < this->transformFunctionDatabase->micLocations.size() ; i++){
		this->micPositions.push_back(this->transformFunctionDatabase->micLocations[i]);
	}
    }

    void    
    BF::newTransformFunctionDatabase()
    {
	using std::swap;
	TransformFunctionDatabase *newTransformFunctionDatabase = new TransformFunctionDatabase();
	swap(this->transformFunctionDatabase, newTransformFunctionDatabase);
	delete newTransformFunctionDatabase;
    }

	// 分離行列データベースのset
	void
	BF::setSeparationMatricesDatabase(const Database &saparationMatricesDatabase)
	{
		using std::swap;
		Database *newSeparationMatricesDatabase = new Database(saparationMatricesDatabase);
		swap(this->separationMatricesDatabase, newSeparationMatricesDatabase);
		delete newSeparationMatricesDatabase;
	}
    
    void    
    BF::newSeparationMatricesDatabase()
    {
		using std::swap;
		Database *newSeparationMatricesDatabase = new Database();
		swap(this->separationMatricesDatabase, newSeparationMatricesDatabase);
		delete newSeparationMatricesDatabase;
    }

	// 音源ID検索
	int
	BF::getSourceIndexUsingID(std::vector<SourceLocation> sourcePositions, SourceLocation Position)
	{
	    /* 直近フレームの音源から、同じ SourceID を持つものを探す */
    	int j;
	    for(j = 0; j < this->sourcePositions.size(); j++){
			if(Position.sourceID == this->sourcePositions[j].sourceID){
			    break;
			}
		}
	
		if(j < this->sourcePositions.size()){
			/* 見つかった */
			return (j);
		}
		
		/* 見つからなかった */
		return (-1);
	}
    
	// TFデータベースのNeighborsを使用したPOS検索
	int
	BF::getSourceIndexUsingNeighbor(std::vector<SourceLocation> sourcePositions, SourceLocation Position)
	{
		if (transformFunctionDatabase == NULL){
			return (-1);
		}

		// 伝達関数データベースからneighborsリストを引く
		int tfSrcIndex = this->transformFunctionDatabase->findSourceIndex(Position);
		if (tfSrcIndex < 0){
			return (-1);	// 見つからなかった
		}
		if (this->transformFunctionDatabase->sourceNeighbors[tfSrcIndex].size() == 0){
			return (-1);	// 見つからなかった
		}
		int i, j;
		for (i = 0 ; i < this->transformFunctionDatabase->sourceNeighbors[tfSrcIndex].size() ; i++){
			for (j = 0 ; j < sourcePositions.size() ; j++){
				if (sourcePositions[j].id == this->transformFunctionDatabase->sourceNeighbors[tfSrcIndex][i]){
					break;
				}
			}
			
			if (j < sourcePositions.size()){
				// 見つかった
				return (j);
			}
		}
		
		// 見つからなかった
		return (-1);
	
	}

	// TFデータベースのNeighborsを使用したPOS検索
	int
	BF::getSourceIndexUsingNeighbor(std::vector<std::vector<SourceLocation> > sourcePositions, SourceLocation Position, int &type)
	{
		type = -1;

		if (transformFunctionDatabase == NULL){
			return (-1);
		}

		// 伝達関数データベースからneighborsリストを引く
		int tfSrcIndex = this->transformFunctionDatabase->findSourceIndex(Position);
		if (tfSrcIndex < 0){
			return (-1); // 見つからなかった
		}
		if (this->transformFunctionDatabase->sourceNeighbors[tfSrcIndex].size() == 0){
			return (-1); // 見つからなかった
		}
		int i, j, k;
		for (i = 0 ; i < this->transformFunctionDatabase->sourceNeighbors[tfSrcIndex].size() ; i++){
			for (k = 0 ; k < sourcePositions.size() ; k++){ // k=0:直近フレーム, k=1:DB
				for (j = 0 ; j < sourcePositions[k].size() ; j++){
					if (sourcePositions[k][j].id == this->transformFunctionDatabase->sourceNeighbors[tfSrcIndex][i]){
						break;
					}
				}
			
				if (j < sourcePositions[k].size()){
					// 見つかった
					type = k;
					return (j);
				}
			}
		}
		
		// 見つからなかった
		return (-1);
	}

	// 現フレームの音源位置について、直近フレームとデータベースの音源位置インデックスリスト作成
	//	sourceIndicestype: 現フレームの各音源に対するインデックスの情報元, 0:直近フレーム, 1:DB, -1:Not Found
	// 	sourceIndices: typeで指定された情報内でのインデックス
	void
	BF::getSourceIndex(std::vector<SourceLocation> currentPositions, std::vector<int> &sourceIndicestype, std::vector<int> &sourceIndices)
	{
		// 初期化
		sourceIndices.resize(currentPositions.size(), -1);
		sourceIndicestype.resize(currentPositions.size(), -1);

		// POS検索で使用するsourcePositions
		std::vector<std::vector<SourceLocation> > sourcePositions;
		sourcePositions.push_back(this->sourcePositions);
		if (separationMatricesDatabase){ // 分離行列データベースのsourceLocationsを追加
			sourcePositions.push_back(separationMatricesDatabase->sourceLocations);
		}
		
		int i, index, type;
		for (i = 0 ; i < currentPositions.size() ; i++){
			index = -1;
			type = -1;
			
			// 音源ID検索(直近フレームからのみの対応) - UtID, UtID_POS
			if (separationMatUpdateType == UtID || separationMatUpdateType == UtID_POS){
				index = getSourceIndexUsingID(this->sourcePositions, currentPositions[i]);
				type = 0;
			}
			// POS検索(TFデータベースのNeighborsを使用) - UtPOS, ID検索で見つからなかった場合のUtID_POS
			if (index < 0 && (separationMatUpdateType == UtID_POS || separationMatUpdateType == UtPOS)){
				index = getSourceIndexUsingNeighbor(sourcePositions, currentPositions[i], type);
			}
							
			sourceIndices[i] = index;
			if (index >= 0){
				sourceIndicestype[i] = type;
			}
		}
	}


	// 分離行列データベースの座標を更新
	//	currentPositions: 現フレームの音源位置
	//	sourceIndicestype: 現フレームの各音源に対するインデックスの情報元, 0:直近フレーム, 1:DB, -1:Not Found
	//	sourceIndices: 現フレームの各音源に対するDBの音源位置
	void
	BF::updateSeparationMatricesDatabaseSource(std::vector<SourceLocation> currentPositions, std::vector<int> sourceIndicestype, std::vector<int> sourceIndices)
	{
		// IDの場合座標更新しない
		if (separationMatUpdateType == UtID){
			return;
		}

		// サイズチェック
		if (currentPositions.size() != sourceIndicestype.size() ||
		    currentPositions.size() != sourceIndices.size()){
			return;
		}

		// 座標更新
		int index;
		for (int i = 0 ; i < currentPositions.size() ; i++){
			index = sourceIndices[i];

			if (sourceIndicestype[i] == 0 &&
			    index >= 0 && 
			    index < this->sourcePositions.size()){
			    // 直近フレームの結果もデータベースに書き込まれているため、それに対応するものの座標を更新する
			    index = separationMatricesDatabase->findSourceIndex(sourcePositions[index]);
			}
			else if (sourceIndicestype[i] != 1 ||
			    index < 0 || 
			    index >= separationMatricesDatabase->sourceLocations.size()){
			    index = -1;
			}

			if (index >= 0){
 			    // データベースの座標更新
			    separationMatricesDatabase->sourceLocations[index] = currentPositions[i];
			}
		}
		// データベースの座標が重複している場合は先に出現しているものを削除
		for (int i = separationMatricesDatabase->sourceLocations.size()-1 ; i >= 0  ; i--){
			for (int k = i-1 ; k >= 0 ; k--){
				if (separationMatricesDatabase->sourceLocations[i].id == separationMatricesDatabase->sourceLocations[k].id){
					separationMatricesDatabase->sourceLocations.erase(separationMatricesDatabase->sourceLocations.begin() + k);
					separationMatricesDatabase->Data.erase(separationMatricesDatabase->Data.begin() + k);
				}
			}
		}
	}
	
	// 分離行列データベースへの保存
	//     separationMatricesに保存されている全てのデータを保存する
	void 
	BF::updateSeparationMatricesDatabaseAll()
	{
		int i, j, k, index;

		if (separationMatricesDatabase == NULL){
			return;
		}

		// this->separationMatricesはまだ使用されていない
		if (this->separationMatrices.size() == 0){
			return;
		}

		// サイズチェック
		if (this->separationMatrices[0].getRowCount() != this->sourcePositions.size()){
			return;	// 何もしない
		}
		
		// マイク座標はデータベース登録されている？
		if (separationMatricesDatabase->micLocations.size() != this->micPositions.size()){
			separationMatricesDatabase->micLocations.clear();
			for (i = 0 ; i < this->micPositions.size() ; i++){
				separationMatricesDatabase->micLocations.push_back(this->micPositions[i]);
			}
		}
		
		for (i = 0 ; i < this->sourcePositions.size() ; i++){
			// 現在の分離行列データベースに座標が存在する？
			index = separationMatricesDatabase->findSourceIndex(this->sourcePositions[i]);
			if (index < 0){
				// 座標を追加する
//std::cout << "Added to separationMatricesDatabase: " << this->sourcePositions[i].id << "@BF::updateSeparationMatricesDatabaseAll" << std::endl;
				separationMatricesDatabase->sourceLocations.push_back(this->sourcePositions[i]);
				Matrix m(this->micPositions.size(), this->separationMatrices.size());
				separationMatricesDatabase->Data.push_back(m);
				index = separationMatricesDatabase->Data.size() - 1;			
			}
			
			if (index >= 0){
				// データ更新
				// separationMatrices: size_is(numFreq), element_is(Matrix(numSrc, numMic))
				// separationMatricesDatabase->Data: size_is(numSrc), elemnent_is(Matrix(numMic, numFreq))
				for (j = 0 ; j < this->micPositions.size() ; j++){
					for (k = 0 ; k < this->separationMatrices.size() ; k++){
						separationMatricesDatabase->Data[index].setElementAt(j, k, this->separationMatrices[k].getElementAt(i, j));
					}
				}
			}
		}
	}
	
	// 分離行列データベースから値を取得
	//     DBIndex: 読み込むDBのインデックス
	//     currentSourceIndex: currentSeparationMatricesの対象座標インデックス
	//     currentFreqIndex: currentSeparationMatricesの対象周波数インデックス
	//     currentSeparationMatrices: 更新中の分離行列
	bool
	BF::setSeparationMatricesFromDatabase(int DBIndex, int currentSourceIndex, int currentFreqIndex, Matrix &currentSeparationMatrices){

		int j;
		if (separationMatricesDatabase == NULL){
			return false;
		}
		
		// currentSeparationMatrices: size_is(numFreq), element_is(Matrix(numSrc, numMic))
		// separationMatricesDatabase->Data: size_is(numSrc), elemnent_is(Matrix(numMic, numFreq))
		for (j = 0 ; j < this->micPositions.size() ; j++){
			currentSeparationMatrices.setElementAt(currentSourceIndex, j, 													separationMatricesDatabase->Data[DBIndex].getElementAt(j, currentFreqIndex));
		}

		return true;
	}

}
