/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include "BF_Fixed.h"

#include <stdio.h>
#include <iostream>
namespace BF
{
    Fixed::Fixed(int _numFreq, int _lowerfreq, int _upperfreq, int _samplingRate, double _soundSpeed) :
	BF(_numFreq, _lowerfreq, _upperfreq, _samplingRate, _soundSpeed, UtID)
    {
    }

    Fixed::~Fixed()
    {
    }

    /*== From BSsample.m ==

for method = 1:4 % DS, WDS, NULL, ILSE
    if method ~= 3 % except NULL
       G{method} = mkBFcoef(H, method);

    else % Reduce non-target source TFs for Nsrc < NMic
      for ns = 1:NSrc 
        nti = 1 + rem([ns ns-3 ns+3]-1+24, 24); % decimate source index
        Hr = H(:, nti, :); % decimation
        Gn = mkBFcoef(Hr, method); % BF design
        G{method}(ns, :, :) = Gn(1, :, :); % Store BF coef.
      end
    end
    
end

     */

    /*== From mkBFcoef.m ==

switch method
    case 1 % DS
        for nf = 1:NFreq
            for ns = 1:NSrc
              H0 = H(:, ns, nf);
              H0a = abs(H0);
              G(ns, :, nf) = (H0./H0a)'/sum(H0a);
            end

        end

    case 2 % WDS
        for nf = 1:NFreq
            for ns = 1:NSrc
              G(ns, :, nf) = pinv(H(:, ns, nf));
            end
        end
        
    case 3 % NULL
        for nf = 1:NFreq
            G(:, :, nf) = pinv(H(:, :, nf));
        end
        
    case 4 % ILSE
        for nf = 1:NFreq
            for ns = 1:NSrc
              H0 = H(:, ns, nf).'; % Target TF
              g0 = pinv(H0); % Base BF (WDS)
              [U, S, V] = svd(H0);
              Z = V(:, S==0); % Null-space of H0
              Hu = H(:, [(1:ns) (ns+1:end)], nf).'; % Non-target TFs
              y = -pinv(Hu*Z)*Hu*g0; % Indifinite term
              g = g0 + Z*y; % ILSE coefficients
              G(ns, :, nf) = g;
            end
        end
        
    otherwise
        G = [];
end

     */

    void
    Fixed::updateSources(const std::vector<SourceLocation> &sourcePositions)
    {
	int numSrc = sourcePositions.size();

	/*== 伝達関数を再計算する ==*/
	std::vector<Matrix> newHs, newGs;
	this->BF::getTransformFunctions(sourcePositions, newHs);

//	for(int i = lowerfreq; i <= upperfreq ; i++){
	for(int i = 0; i < this->numFreq; i++){
	    const Matrix &H = newHs[i];
	    int numMic = H.getRowCount();

	    // 対象周波数のみ分析
    	if (i == 0 || i == this->numFreq-1 || i < lowerfreq || i > upperfreq) {
	        newGs.push_back(Matrix::O(numSrc, numMic));
		    continue;
	    }

	    Matrix G(numSrc, numMic);

	    switch(this->getBfAlgorithmType()){
	    case BfDS:
		for(int j = 0; j < numSrc; j++){
		    Matrix  H0  = H.getColumn(j);
		    Matrix  H0a = H0.elementwiseAbs();
		    Complex s   = H0a.getElementSum();
		    for(int k = 0; k < numMic; k++){
			if(abs(H0a.getElementAt(k, 0) * s) < BF::BF_EPS)
			{
				H0.setElementAt(k, 0,H0.getElementAt(k, 0) / ((H0a.getElementAt(k, 0) * s) + BF::BF_EPS));
			}
			else
			{
				H0.setElementAt(k, 0, H0.getElementAt(k, 0) / H0a.getElementAt(k, 0) / s);
			}
		    }
		    G.setRow(j, H0.transposed());
		}
		break;
		
	    case BfWDS:
		for(int j = 0; j < numSrc; j++){
		    G.setRow(j, H.getColumn(j).pseudoInverse());
		}
		break;
		
	    case BfNull:
		G = H.pseudoInverse();
		break;
		
	    case BfILSE:
		for(int j = 0; j < numSrc; j++){

		    Matrix H0 = H.getColumn(j).transposed(false);
//		    H0.dump(stdout, "H0");

		    Matrix g0 = H0.pseudoInverse();
//		    g0.dump(stdout, "g0");
			if(numSrc > 1 ){
			    Matrix U, S, V(numMic, numMic);
			    H0.singularValueDecomposition(U, S, V);
//			    U.dump(stdout, "U");
//			    S.dump(stdout, "S");
//			    V.dump(stdout, "V");

			    Matrix Z(numMic, numMic - 1);
			    /* g0 は rank=1 なので、S は (+, 0, 0) の形
			       -> U(:, S==0) -> U の 1 列目以外を採用。 */
			    for(int k = 0; k < Z.getColumnCount(); k++){
				Z.setColumn(k, V.getColumn(k + 1));
			    }
				
				Matrix Hn(numMic, sourcePositions.size() - 1) ;//Non-target TFs 
			
				int m = 0;
				for(int k = 0; k < sourcePositions.size(); k++){
					if(k!=j){
						Hn.setColumn(m, H.getColumn(k));
						m++;
					}
			    }

//			    Hn.dump(stdout, "Hn");
			    Matrix Hu = Hn.transposed(false);
			    
//			    Hu.dump(stdout, "Hu");
//				g0.dump(stdout, "g0");
//			    (Hu * Z).dump(stdout, "(Hu * Z)");
//			    (Hu * Z).pseudoInverse().dump(stdout, "pinv(Hu * Z)");
//			    ((Hu * Z).pseudoInverse() * Hu).dump(stdout, "pinv(Hu * Z) * Hu");
//			    ((Hu * Z).pseudoInverse() * Hu * g0).dump(stdout, "pinv(Hu * Z) * Hu * g0");

			    Matrix y = -1 * (Hu * Z).pseudoInverse() * Hu * g0;	   
			    Matrix g = g0 + Z * y;

			    G.setRow(j, g.transposed(false));
			}
			else
			{		
					G.setRow(j, g0.transposed(false));
			}
		}
		break;
	    }

//	    G.dump(stdout, "G[%d]", i);
	    newGs.push_back(G);
	}

	/*== 結果の格納 ==*/
	this->sourcePositions = sourcePositions;
	newHs.swap(this->transformFunctions);
	newGs.swap(this->separationMatrices);
    }

    void
    Fixed::adaptSeparationMatrices()
    {
	/* do nothing */
    }
}
