/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include "BF_MSNR.h"

#include <stdio.h>
#include <iostream>

namespace BF
{
    CMSNR::CMSNR(int _numFreq, int _lowerfreq, int _upperfreq, int _samplingRate, double _soundSpeed, std::string _decomp_alg, double _alpha) :
	BF(_numFreq, _lowerfreq, _upperfreq, _samplingRate, _soundSpeed, UtID),
	decomp_alg(_decomp_alg), alpha(_alpha)
    {
    }

    CMSNR::~CMSNR()
    {
    }

    void
    CMSNR::updateSources(const std::vector<SourceLocation> &sourcePositions)
    {
/*== source from adaptBF.m

  % This algorithm requires VAD data (Adata.vad(NSrc), = 0(noise) or 1(signal))
  % This algorithm uses 2 intermediate variables Rss, Rnn, mean 
  % correlation matrixes of signal(incl. noise) and noise, respectively.
  % The freedom of scaling is fixed by |g| = 1. (g is BF coef. vector)
  % This algorithm requires to set initial G for getting size information.
  % To set zero matrix for G is OK.
      if isempty(Adata.Rss) % Initialization process
        Adata.Rss = zeros(NMic, NMic, NSrc, NFreq); % Signal correlation matrix
        Adata.Rnn = zeros(NMic, NMic, NSrc, NFreq); % Noise correlation matrix
        for nf = 1:NFreq
          for ns = 1:NSrc % for each source
            Adata.Rss(:, :, ns, nf) = eye(NMic);
            Adata.Rnn(:, :, ns, nf) = eye(NMic);
            G(ns, :, nf) = ones(NMic, 1)./sqrt(NMic); % Initial BF coef.
          end
        end
        return
      else % update process
        for ns = 1:NSrc % for each source
          vad = Adata.vad(ns);
          for nf = 1:NFreq
              x = Adata.x(:, nf);
              Rxx = x*x';
              alp = Adata.alp;
              Rss = Adata.Rss(:, :, ns, nf);
              Rnn = Adata.Rnn(:, :, ns, nf);
              if vad == 1 % Signal part
                  Rss = alp * Rss + (1-alp) * Rxx;
                  Adata.Rss(:, :, ns, nf) = Rss;
              else % Noise part
                  Rnn = alp * Rnn + (1-alp) * Rxx;
                  Adata.Rnn(:, :, ns, nf) = Rnn;
              end
              [V, D] = eig(Rss, Rnn);
              [msnr, mi] = max(diag(D));
              g = V(:, mi);
              g = g ./ sqrt(real(g'*g));
              G(ns, :, nf) = g'; % Store total BF coef.
          end
        end
        return
      end % For if initialization
      
*/
		const int numSrc = this->numSignalSources; // = sourcePositions.size();
		const int numMic = this->numMicChannels;

		/*== 直近フレームでの音源と、このフレームでの音源とのadaptSeparationMatrices対応をとる ==*/
		std::vector<int> lastSourceIndices; // i 番めの要素は、このフレームでの i 番目の音源（配列の i 番目の要素という意味）が、直近フレームでの何番目の音源（同）に相当するかを表す。
		for (int i = 0; i < numSrc; i++) {
#if 0
			const SourceLocation &sourcePosition = sourcePositions[i];

			/* 直近フレームの音源から、同じ ID を持つものを探す */
			int j;
			for (j = 0; j < this->sourcePositions.size(); j++) {
				if (sourcePosition.sourceID == this->sourcePositions[j].sourceID) {
					break;
				}
			}
			
			if (j < this->sourcePositions.size()) {
				/* 見つかった */
				lastSourceIndices.push_back(j);
			} else {
				/* 見つからなかった */
				lastSourceIndices.push_back(-1);
			}
#endif
			// 音源ID検索
			lastSourceIndices.push_back(getSourceIndexUsingID(this->sourcePositions, sourcePositions[i]));
		}

		/*== 伝達関数・分離行列を再計算する ==*/
		std::vector<Matrix> newRss, newGs;

		for (int i = 0; i < this->numFreq; i++) {
			// 対象周波数のみ計算
			if (i == 0 || i == this->numFreq-1 || i < lowerfreq || i > upperfreq) {
			    for (int j = 0; j < numSrc; j++) {
					newRss.push_back(Matrix::I(numMic));
				}
				newGs.push_back(Matrix::O(numSrc, numMic));
				continue;
			}

			Matrix newG(numSrc, numMic);
			for (int j = 0; j < numSrc; j++) {
				/* 入力の j 番目の音源と同じ ID の音源が、直近フレームにあるか、見当たらないか？ */
				const int lastSourceIndex = lastSourceIndices[j];
				if (lastSourceIndex < 0) { // 見当たらない
					// 初期値を入力
					newRss.push_back(Matrix::I(numMic));
//if (i == 0 && j == 0) {std::cout << "Rss initialized." << std::endl;}
					for (int k = 0; k < numMic; k++){
						newG.setElementAt(j, k, 1.0 / sqrt(numMic));
					}
				}
				else { // 直近フレームにある
					// データの引き継ぎ
					newRss.push_back(this->Rss[i * this->sourcePositions.size() + lastSourceIndex]);
					newG.setRow(j, this->separationMatrices[i].getRow(lastSourceIndex));
				}
			}
			newGs.push_back(newG);
		}

		/*== 結果の格納 ==*/
		this->sourcePositions = sourcePositions;
		newRss.swap(this->Rss);
		newGs.swap(this->separationMatrices);
    }

    void
    CMSNR::adaptSeparationMatrices()
    {
		const int numSrc = this->numSignalSources;
		const int numMic = this->numMicChannels;	

		// 最初の一回だけ、ここで、Rnnを初期化する
		if (this->Rnn.size() == 0) {
//std::cout << "Rnn initialized." << std::endl;
			std::vector<Matrix> newRnn;
			for (int i = 0; i < this->numFreq; i++) {
				newRnn.push_back(Matrix::I(numMic));
			}
			newRnn.swap(this->Rnn);
		}

		std::vector<Matrix> newRss, newGs, newRnn;

		for (int i = 0; i < this->numFreq; i++) {
			// 対象周波数のみを更新
//			if (i < lowerfreq || i > upperfreq) {
			if (i == 0 || i == this->numFreq-1 || i < lowerfreq || i > upperfreq) {
				if (numSrc > 0) {
					for (int j = 0; j < numSrc; j++) {
						newRss.push_back(Matrix::I(numMic));
					}
					newGs.push_back(Matrix::O(numSrc, numMic));
				}
				else {
					newRnn.push_back(Matrix::I(numMic));
				}
				continue;
			}

			Matrix x  = this->inputSpectrum.getColumn(i);/* / 32768.0;//8388608.0*/
			Matrix rxx = x * x.transposed();
		
			if (numSrc > 0) {
				Matrix rnn = this->Rnn[i];

				Matrix newG(numSrc, numMic);
				for (int j = 0; j < numSrc; j++) {
					Matrix rss = this->alpha * this->Rss[i * numSrc + j] + (1.0 - this->alpha) * rxx;
					newRss.push_back(rss);

					Matrix evec, eval, right_evec;
					/* Rnn が I でない場合、octave の結果とこの一般固有値の計算結果は一致しないようである。
					 * ただし、octave の結果は AV-BVD が 0 ではないので、octave 実装に誤差があると予想する。
					 */
					if(decomp_alg == "GEVD")
						rss.generalizedEigenValues(rnn, eval, evec);
					else if(decomp_alg == "GSVD")
						rss.generalizedSingularValues(rnn, evec, eval, right_evec);

					int m = eval.getMaximumValueIndexInColumn(0);
					Matrix g = evec.getColumn(m);
					/*g = 1.0 / g.norm().real() * g;*/
					//g = g * (1.0 / sqrt(g.norm().real()));
					/*g = g / g.norm(); // update*/
					if (g.norm() == 0.0)
						g = g * 0.0;
					else
						g = g / g.norm();
//g.dump(stdout, "g");
//std::cout << "[" << i << "," << j << "]" << g.norm() << std::endl;
					newG.setRow(j, g.transposed());
//if (i == 0) std::cout << "Rss[" << j << "] adapted." << std::endl;
				}
				newGs.push_back(newG);
			}
			else {
				Matrix rnn = this->alpha * this->Rnn[i] + (1.0 - this->alpha) * rxx;
				newRnn.push_back(rnn);
//if (i == 0) std::cout << "Rnn adapted." << std::endl;
			}
		}

		/*== 結果の格納 ==*/
		if (numSrc > 0) {
			newRss.swap(this->Rss);
			newGs.swap(this->separationMatrices);
		}
		else {
			newRnn.swap(this->Rnn);
		}
    }
}
