/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <iostream>
#include <memory>
#include <deque> 

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>
#include <Source.h>

#include "LocalizationCSP.hpp"

using namespace HARK;

class CSP;

DECLARE_NODE(CSP);
/*Node
 *
 * @name CSP
 * @category HARK:Localization
 * @description Sound Source Localization by CSP
 *
 * @input_name INPUT
 * @input_type Matrix<complex<float> >
 * @input_description 2-channel spectrum. In this matrix, a row is channel(2), and a column is spectrum for the corresponding channel.
 *
 * @output_name OUTPUT
 * @output_type Vector<ObjectRef>
 * @output_description Source location. A type of an element of the vector is Source.
 * 
 * @parameter_name DISTANCE_BETWEEN_MICS
 * @parameter_type float
 * @parameter_value 0.3
 * @parameter_description Distance between microphones (m).
 * 
 * @parameter_name SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_description Sampling rate (Hz).
 * 
 * @parameter_name SPEED_OF_SOUND
 * @parameter_type float
 * @parameter_value 340.0
 * @parameter_description Speed of sound (m/sec).
 * 
 * @parameter_name LENGTH
 * @parameter_type int
 * @parameter_value 512
 * @parameter_description The length of a frame (per channel).
 * 
 * @parameter_name LOWER_BOUND_FREQUENCY
 * @parameter_type int
 * @parameter_value 500
 * @parameter_description Lower bound of frequency (Hz) used for correlation function calculation.
 * 
 * @parameter_name UPPER_BOUND_FREQUENCY
 * @parameter_type int
 * @parameter_value 2800
 * @parameter_description Upper bound of frequency (Hz) used for correlation function calculation.
 * 
 * @parameter_name MANUAL_WEIGHT_SQUARE 
 * @parameter_type Vector<float> 
 * @parameter_value <Vector<float> 0.0 2000.0 4000.0 6000.0 8000.0> 
 * @parameter_valid SPECTRUM_WEIGHT_TYPE=Manual_Square 
 * @parameter_description Cross spectrum weight for each frequency bin. This is a M order vector. The element represents the frequency points for the square wave. "M" represents the number of key points for the square wave weight. The format is "<Vector<float> 1 2 3 4 5>". 
 * 
 * @parameter_name MIN_DEG
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description Lower source direction.
 * 
 * @parameter_name MAX_DEG
 * @parameter_type int
 * @parameter_value 180
 * @parameter_description Higher source direction.
 * 
 * @parameter_name WINDOW
 * @parameter_type int
 * @parameter_value 50
 * @parameter_description The number of frames used for calculating a correlation function.
 * 
 * @parameter_name WINDOW_TYPE
 * @parameter_type string
 * @parameter_list PAST:MIDDLE:FUTURE
 * @parameter_value FUTURE
 * @parameter_description Window selection to accumulate a correlation function. If PAST, the past WINDOW frames from the current frame are used for the accumulation. If MIDDLE, the current frame will be the middle of the accumulated frames. If FUTURE, the future WINDOW frames from the current frame are used for the accumulation. FUTURE is the default from version 1.0, but this makes a delay since we have to wait for the future information. PAST generates a internal buffers for the accumulation, which realizes no delay for localization. 
 * 
 * @parameter_name PERIOD
 * @parameter_type int
 * @parameter_value 50
 * @parameter_description The period in which the source localization is processed.
 * 
 * @parameter_name CSP_THRESHOLD
 * @parameter_type float
 * @parameter_value 0.0
 * @parameter_description Threshold of CSP coefficient.
 * 
 * @parameter_name MAXNUM_OUT_PEAKS
 * @parameter_type int
 * @parameter_value -1
 * @parameter_description Maximum number of output peaks. If MAXNUM_OUT_PEAKS > 0, number of output peaks is limited to MAXNUM_OUT_PEAKS.
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_description Debug option. If the parameter is true, this node outputs sound localization results to a standard output.
 * 
END*/

class CSP : public BufferedNode {
public:
	CSP(string nodeName, ParameterSet params)
	  : BufferedNode(nodeName, params), cspID(-1)
	{
		// Get and Check terminal

		this->inputID  = addInput("INPUT");
		this->outputID = addOutput("OUTPUT");

		// Get parameters

		float distanceMics = dereference_cast<float>(parameters.get("DISTANCE_BETWEEN_MICS"));
		if (distanceMics <= 0.0f) {
			throw new NodeException(NULL, "CSP:DISTANCE_BETWEEN_MICS is invalid value.", __FILE__, __LINE__);
		}

		int samplingRate = dereference_cast<int>(parameters.get("SAMPLING_RATE"));
		if (samplingRate <= 0) {
			throw (new NodeException(NULL, "CSP:SAMPLING_RATE is invalid value.", __FILE__, __LINE__));
		}

		float speedSound = dereference_cast<float>(parameters.get("SPEED_OF_SOUND"));
		if (speedSound <= 0) {
			throw new NodeException(NULL, "CSP:SPEED_OF_SOUND is invalid value.", __FILE__, __LINE__);
		}

		this->length = dereference_cast<int>(parameters.get("LENGTH"));
		if (this->length <= 0) {
			throw new NodeException(NULL, "CSP:LENGTH is invalid value.", __FILE__, __LINE__);
		}

		int lower_bound_freq = dereference_cast<int>(parameters.get("LOWER_BOUND_FREQUENCY"));
		int upper_bound_freq = dereference_cast<int>(parameters.get("UPPER_BOUND_FREQUENCY"));
		if (lower_bound_freq < 0 || lower_bound_freq > samplingRate / 2) {
			throw new NodeException(NULL, "CSP:LOWER_BOUND_FREQUENCY is invalid value.", __FILE__, __LINE__);
		}
		if (upper_bound_freq < 0 || upper_bound_freq > samplingRate / 2) {
			throw new NodeException(NULL, "CSP:UPPER_BOUND_FREQUENCY is invalid value.", __FILE__, __LINE__);
		}
		if (lower_bound_freq > upper_bound_freq) {
			throw new NodeException(NULL, 
			      "CSP:LOWER_BOUND_FREQUENCY is larger than UPPER_BOUND_FREQUENCY.", __FILE__, __LINE__);
		}

		Vector<float> SSW_SQ = object_cast<Vector<float> >(parameters.get("MANUAL_WEIGHT_SQUARE"));       

		int minDeg = dereference_cast<int>(parameters.get("MIN_DEG"));
		int maxDeg = dereference_cast<int>(parameters.get("MAX_DEG"));
		if (minDeg > maxDeg) {
			throw new NodeException(NULL, "CSP:MIN_DEG is larger than MAX_DEG.", __FILE__, __LINE__);
		}

		this->window = dereference_cast<int>(parameters.get("WINDOW"));
		if (this->window <= 0) {
			throw new NodeException(NULL, "CSP:WINDOW is invalid value.", __FILE__, __LINE__);
		}

		this->window_type = object_cast<String>(parameters.get("WINDOW_TYPE"));
    
		this->period = dereference_cast<int>(parameters.get("PERIOD"));
		if (this->period <= 0) {
			throw new NodeException(NULL, "CSP:PERIOD is invalid value.", __FILE__, __LINE__);
		}

		float csp_threshold = dereference_cast<float>(parameters.get("CSP_THRESHOLD"));

		int max_num_out_peaks = dereference_cast<int>(parameters.get("MAXNUM_OUT_PEAKS"));

		this->is_debug = dereference_cast<bool>(parameters.get("DEBUG"));

		// Set variable

		this->pslength = this->length / 2 + 1;

		this->MSSW.resize(this->pslength);
		vector<float> mws;
		mws.resize(this->pslength);
		for (unsigned int i = 0; i < this->pslength; i++) {
			int SSW_j = -1;
			for (unsigned int j = 0; j < SSW_SQ.size(); j++) {
				if (SSW_SQ[j] <= samplingRate * (float)i / (float)this->length) {
					SSW_j = j;
				}
			}
			if (SSW_j % 2 == 0) { mws[i] = 1.0f; }
			else                { mws[i] = 0.0f; }
		}
		int lower = floor((float)lower_bound_freq * this->length / samplingRate);
		int upper = ceil((float)upper_bound_freq * this->length / samplingRate);
		for (unsigned int i = lower; i < upper; i++) {
			this->MSSW[i] = mws[i];
		}

		this->unitTau = (float)samplingRate * distanceMics / speedSound;
		int minTau = floor(this->unitTau * cos((float)maxDeg * M_PI / 180.0f));
		int maxTau = ceil(this->unitTau * cos((float)minDeg * M_PI / 180.0f));

		this->xbuf.resize(0);

		if (this->window_type == "FUTURE") {
			this->look_ahead = this->window - 1;
			this->look_back = 0;
		} else if (window_type == "MIDDLE") {
			this->look_ahead = this->window / 2;
			this->look_back = this->window / 2 + (this->window % 2) - 1;
		} else if (window_type == "PAST") {
			this->look_ahead = 0;
			this->look_back  = 0;
		} else {
			throw new NodeException(NULL, "CSP: Unexpected WINDOW_TYPE is set.",__FILE__,__LINE__);
		}

		inputsCache[this->inputID].lookAhead = this->look_ahead;
		inputsCache[this->inputID].lookBack = this->look_back;

		this->sum_count = 0;

		this->Rx.resize(this->pslength);
		for (int k = 0; k < this->pslength; k++) { this->Rx[k] = BF::Complex(0.0f, 0.0f); }
 
		this->localizer.reset(new LocalizationCSP(this->length, this->MSSW, minTau, maxTau, 
		                                          csp_threshold, max_num_out_peaks));

		inOrder = true;
	}

	virtual void initialize()
	{
		int max_lookAhead = outputs[this->outputID].lookAhead; 
		int max_lookBack  = outputs[this->outputID].lookBack; 
    
		if (this->cspID != -1) { 
			max_lookAhead = max(max_lookAhead, outputs[this->cspID].lookAhead); 
			max_lookBack  = max(max_lookBack,  outputs[this->cspID].lookBack);       
		}

		outputs[this->outputID].lookAhead = 1 + max_lookAhead; 
		outputs[this->outputID].lookBack  = 1 + max_lookBack;     
    
		if (this->cspID != -1) {
			outputs[this->cspID].lookAhead = 1 + max_lookAhead; 
			outputs[this->cspID].lookBack  = 1 + max_lookBack; 
		}

		this->BufferedNode::initialize();
	} 

	virtual int translateOutput (const string &outputName)
	{
		for (size_t i = 0; i < outputNames.size(); i++) {	
			if (outputNames[i] == outputName) {
				return i;
			}
		}

		if (outputName == "CSPVALUE") {
			return this->cspID = addOutput(outputName);
		} else {
			//throw new NodeException(this, "CSP:" + outputName+ " is not supported.", __FILE__, __LINE__);
			return addOutput(outputName);
		}
	}

	void calculate(int output_id, int count, Buffer &out)
	{
		// input spectrum
		RCPtr<Matrix<complex<float> > > inputp = getInput(this->inputID, count);
		if (count == 0) { 
			if (2 != (*inputp).nrows()) {
				throw new NodeException(this, "CSP:Input channel size is not 2 channels.", __FILE__, __LINE__); 
			}
			if (this->pslength != (*inputp).ncols())  {
				throw new NodeException(this, "CSP:Input frame size is not parameter LENGTH.", __FILE__, __LINE__);
			}
		}

		// output source
		RCPtr<Vector<ObjectRef> > sources(Vector<ObjectRef>::alloc(0));
		(*(outputs[this->outputID].buffer))[count] = sources;

		// output BF spectrum power
		RCPtr<Vector<float > > coef(new Vector<float >(length));    
		if (this->cspID != -1) {    
			(*(outputs[this->cspID].buffer))[count] = coef;
			for (unsigned int i = 0; i < length; i++) {
				(*coef)[i] = 0.0f; 
			}
		}

		// get spectrum data and normalize crossCorrelation

		BF::Matrix x(2, this->pslength);
		BF::Complex xval;
		if (this->window_type == "PAST") {
			// input cashe 
			inputp = getInput(this->inputID, count);
			for (int chan = 0; chan < 2; chan++) {
				for (int k = 0; k < this->pslength; k++) {
					xval = (*inputp)(chan, k) * this->MSSW[k];
					x.setElementAt(chan, k, xval);
				}
			}
			if (this->xbuf.size() < this->window) { 
				this->xbuf.push_back(x); 
			} else {
				this->xbuf.push_back(x);       
				this->xbuf.pop_front();       
			}
      
			if ((count % this->period == 0) && ((count != 0) || (this->period == 1))) {
				for (int i = 0; i < this->xbuf.size(); i++) { 
					for (int chan = 0; chan < 2; chan++) {
						for (int k = 0; k < pslength; k++) {
							x.setElementAt(chan, k, this->xbuf[i].getElementAt(chan, k));
						}
					}
					this->localizer->AddCrossCorrelation(this->Rx, x);
					this->sum_count++;
				}
				this->localizer->NormalizeCrossCorrelation(this->Rx, this->sum_count);
				this->sum_count = 0;
			} else {
				return;
			}
		} else {
			if (count % this->period == 0 && count >= this->look_back) { 
				for (int i = -this->look_back; i <= this->look_ahead; i++) { 
					inputp = getInput(this->inputID, count + i);

					for (int chan = 0; chan < 2; chan++) {
						for (int k = 0; k < pslength; k++) {
							xval = (*inputp)(chan, k) * this->MSSW[k];
							x.setElementAt(chan, k, xval);
						}
					}
					this->localizer->AddCrossCorrelation(this->Rx, x);
					this->sum_count++;
				}
				this->localizer->NormalizeCrossCorrelation(this->Rx, this->sum_count);
				this->sum_count = 0;
			} else {
				return;
			}
		}

		// Localize

		vector<float> Csp;
		Csp.resize(this->length);
		vector<int> FoundId;
		FoundId.resize(this->length);
		vector<float> FoundCsp;
		FoundCsp.resize(this->length);

		int n_result = this->localizer->Localize(this->Rx, Csp, FoundId, FoundCsp);

		// Output Source

		if (this->is_debug) { printf("Frame %d\n", count); }
		for (int i = 0; i < n_result; i++) {
			float theta = (float)FoundId[i] / this->unitTau;
			if (theta > 1)       { theta = 1; }
			else if (theta < -1) { theta = -1; }
			theta = acos(theta);

			RCPtr<Source> src(new Source);

			src->x[0] = cos(theta);
			src->x[1] = sin(theta);
			src->x[2] = 0.0f;
			src->power = FoundCsp[i];

			src->id = 0;
			src->tfindex[0] = FoundId[i];
			src->tfindex[1] = 0;
			src->tfindex[2] = 0;
			src->tfindex_max[0] = 0;
			src->tfindex_max[1] = 0;
			src->tfindex_max[2] = 0;
			src->tfindex_min[0] = 0;
			src->tfindex_min[1] = 0;
			src->tfindex_min[2] = 0;
			src->compare_mode = Source::TFINDEX;

			if (this->is_debug) { printf("Source: %d %f %f %f\n", FoundId[i], src->x[0], src->x[1], src->power); }

			sources->push_back(src);
		}

		// Output CSP coefficient

		if (this->cspID != -1) {
			for (unsigned int i = 0; i < this->length; i++) {
				(*coef)[i] = Csp[i];
			}
		}

		if (this->is_debug) {
			printf("CSP coefficent:\n");
			for (unsigned int i = 0; i < this->length; i++) {
				printf("%f ", Csp[i]);
			}
			printf("\n");
		}

		this->localizer->ClearCrossCorrelation(this->Rx);
	}

private:
	// input/output

	int inputID;
	int outputID;
	int cspID;

	// parameter

	int length;
	int window;
	string window_type;
	int period;
	bool is_debug;

	// variable

	int pslength;
	vector<float> MSSW;
	double unitTau;
	deque<BF::Matrix> xbuf;
	int look_ahead;
	int look_back;
	int sum_count;
	vector<BF::Complex> Rx;

	std::unique_ptr<LocalizationCSP> localizer;

	//IN_ORDER_NODE_SPEEDUP(CSP);
};

