/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <iostream>
#include <BufferedNode.h>
#include <Buffer.h>
#include <Matrix.h>
#include <Vector.h>
#include <Stream.h>
#include <Source.h>
#include <Map.h>
#include <Time.h>

using namespace HARK;

class DataLogger;

DECLARE_NODE(DataLogger)
/*Node
 *
 * @name DataLogger
 * @category HARK:MISC
 * @description Log the input to stdout.
 *
 * @input_name INPUT
 * @input_type any
 * @input_description The input is what you want to log.
 *
 * @output_name OUTPUT
 * @output_type any
 * @output_description The output is the same as the input.
 *
 * @parameter_name LABEL
 * @parameter_type string
 * @parameter_description add a label to logged data
 *
 END*/

class DataLogger: public BufferedNode {
    int inputID;
    int outputID;

    std::string label;

public:
    DataLogger(std::string nodeName, ParameterSet params) :
            BufferedNode(nodeName, params) {
        inputID = addInput("INPUT");
        outputID = addOutput("OUTPUT");

        if(parameters.exist("LABEL")){
            label = object_cast<std::string>(parameters.get("LABEL"));
        }

        inOrder = true;
    }

    void dump(std::ostream &os, ObjectRef input) {
        if(typeid(*input) == typeid(Map<int, ObjectRef>)){
            std::cout << "{"
                      << "\"type\": \"Map<int, ObjectRef>\"";
            RCPtr<Map<int, ObjectRef> > in = input;
            std::cout << ", \"size\": " << in->size();

            std::cout << ", \"values\": {";
            Map<int, ObjectRef>::iterator it = in->begin();
            while(it != in->end()){
                std::cout << "\"" << it->first << "\": ";
                this->dump(std::cout, it->second);
                ++it;
                if(it != in->end()){
                    std::cout << ", ";
                }
            }
            std::cout << "}"
                      << "}";
        }
        else if(typeid(*input) == typeid(Vector<ObjectRef>)){
            std::cout << "{"
                      << "\"type\": \"Vector<ObjectRef>\"";
            RCPtr<Vector<ObjectRef> > in = input;
            std::cout << ", \"size\": " << in->size();

            std::cout << ", \"values\": [";
            Vector<ObjectRef>::iterator it = in->begin();
            while(it != in->end()){
                this->dump(std::cout, *it);
                ++it;
                if(it != in->end()){
                    std::cout << ", ";
                }
            }
            std::cout << "]"
                      << "}";
        }
        else if(typeid(*input) == typeid(Vector<float>)){
            std::cout << "{"
                      << "\"type\": \"Vector<float>\"";
            RCPtr<Vector<float> > vec = input;
            std::cout << ", \"size\": " << vec->size();

            std::cout << ", \"values\": [";
            int i = 0;
            while(i < vec->size()){
                std::cout << (*vec)[i];
                ++i;
                if(i < vec->size()){
                    std::cout << ", ";
                }
            }
            std::cout << "]"
                      << "}";
        }
        else if(typeid(*input) == typeid(Vector<int>)){
            std::cout << "{"
                      << "\"type\": \"Vector<int>\"";
            RCPtr<Vector<int> > vec = input;
            std::cout << ", \"size\": " << vec->size();

            std::cout << ", \"values\": [";
            int i = 0;
            while(i < vec->size()){
                std::cout << (*vec)[i];
                ++i;
                if(i < vec->size()){
                    std::cout << ", ";
                }
            }
            std::cout << "]"
                      << "}";
        }
        else if(typeid(*input) == typeid(Vector<bool>)){
            std::cout << "{"
                      << "\"type\": \"Vector<bool>\"";
            RCPtr<Vector<bool> > vec = input;
            std::cout << ", \"size\": " << vec->size();

            std::cout << ", \"values\": [";
            int i = 0;
            while(i < vec->size()){
                std::cout << (*vec)[i];
                ++i;
                if(i < vec->size()){
                    std::cout << ", ";
                }
            }
            std::cout << "]"
                      << "}";
        }
        else if(typeid(*input) == typeid(Vector<std::complex<float> >)){
            std::cout << "{"
                      << "\"type\": \"Vector<std::complex<float> >\"";
            RCPtr<Vector<std::complex<float> > > vec = input;
            std::cout << ", \"size\": " << vec->size();

            std::cout << ", \"values\": [";
            int i = 0;
            while(i < vec->size()){
                std::cout << (*vec)[i];
                ++i;
                if(i < vec->size()){
                    std::cout << ", ";
                }
            }
            std::cout << "]"
                      << "}";
        }
        else if(typeid(*input) == typeid(Matrix<ObjectRef>)){
            std::cout << "{"
                      << "\"type\": \"Matrix<ObjectRef>\"";
            RCPtr<Matrix<ObjectRef> > in = input;
            std::cout << ", \"size\": (" << in->nrows() << ", " << in->ncols() << ")";

            std::cout << ", \"values\": [";
            int i = 0, j = 0;
            while(i < in->nrows()){
                std::cout << "[";
                while(j < in->ncols()){
                    this->dump(std::cout, (*in)(i, j));
                    ++j;
                    if(j < in->ncols()){
                        std::cout << ", ";
                    }
                }
                j = 0;
                std::cout << "]";
                ++i;
                if(i < in->nrows()){
                    std::cout << ", ";
                }
            }
            std::cout << "]"
                      << "}";
        }
        else if(typeid(*input) == typeid(Matrix<float>)){
            std::cout << "{"
                      << "\"type\": \"Matrix<float>\"";
            RCPtr<Matrix<float> > mat = input;
            std::cout << ", \"size\": (" << mat->nrows() << ", " << mat->ncols() << ")";

            std::cout << ", \"values\": [";
            int i = 0, j = 0;
            while(i < mat->nrows()){
                std::cout << "[";
                while(j < mat->ncols()){
                    std::cout << (*mat)(i, j);
                    ++j;
                    if(j < mat->ncols()){
                        std::cout << ", ";
                    }
                }
                j = 0;
                std::cout << "]";
                ++i;
                if(i < mat->nrows()){
                    std::cout << ", ";
                }
            }
            std::cout << "]"
                      << "}";
        }
        else if(typeid(*input) == typeid(Matrix<int>)){
            std::cout << "{"
                      << "\"type\": \"Matrix<int>\"";
            RCPtr<Matrix<int> > mat = input;
            std::cout << ", \"size\": (" << mat->nrows() << ", " << mat->ncols() << ")";

            std::cout << ", \"values\": [";
            int i = 0, j = 0;
            while(i < mat->nrows()){
                std::cout << "[";
                while(j < mat->ncols()){
                    std::cout << (*mat)(i, j);
                    ++j;
                    if(j < mat->ncols()){
                        std::cout << ", ";
                    }
                }
                j = 0;
                std::cout << "]";
                ++i;
                if(i < mat->nrows()){
                    std::cout << ", ";
                }
            }
            std::cout << "]"
                      << "}";
        }
        else if(typeid(*input) == typeid(Matrix<bool>)){
            std::cout << "{"
                      << "\"type\": \"Matrix<bool>\"";
            RCPtr<Matrix<bool> > mat = input;
            std::cout << ", \"size\": (" << mat->nrows() << ", " << mat->ncols() << ")";

            std::cout << ", \"values\": [";
            int i = 0, j = 0;
            while(i < mat->nrows()){
                std::cout << "[";
                while(j < mat->ncols()){
                    std::cout << (*mat)(i, j);
                    ++j;
                    if(j < mat->ncols()){
                        std::cout << ", ";
                    }
                }
                j = 0;
                std::cout << "]";
                ++i;
                if(i < mat->nrows()){
                    std::cout << ", ";
                }
            }
            std::cout << "]"
                      << "}";
        }
        else if(typeid(*input) == typeid(Matrix<std::complex<float> >)){
            std::cout << "{"
                      << "\"type\": \"Matrix<std::complex<float> >\"";
            RCPtr<Matrix<std::complex<float> > > mat = input;
            std::cout << ", \"size\": (" << mat->nrows() << ", " << mat->ncols() << ")";

            std::cout << ", \"values\": [";
            int i = 0, j = 0;
            while(i < mat->nrows()){
                std::cout << "[";
                while(j < mat->ncols()){
                    std::cout << (*mat)(i, j);
                    ++j;
                    if(j < mat->ncols()){
                        std::cout << ", ";
                    }
                }
                j = 0;
                std::cout << "]";
                ++i;
                if(i < mat->nrows()){
                    std::cout << ", ";
                }
            }
            std::cout << "]"
                      << "}";
        }
        else if(typeid(*input) == typeid(Map<int, float>)){
            std::cout << "{"
                      << "\"type\": \"Map<int, float>\"";
            RCPtr<Map<int, float> > in = input;
            std::cout << ", \"size\": " << in->size();

            std::cout << ", \"values\": {";
            Map<int, float>::iterator it = in->begin();
            while(it != in->end()){
                std::cout << "\"" << it->first << "\": " << it->second;
                ++it;
                if(it != in->end()){
                    std::cout << ", ";
                }
            }
            std::cout << "}"
                      << "}";
        }
        else if(typeid(*input) == typeid(Map<int, int>)){
            std::cout << "{"
                      << "\"type\": \"Map<int, int>\"";
            RCPtr<Map<int, int> > in = input;
            std::cout << ", \"size\": " << in->size();

            std::cout << ", \"values\": {";
            Map<int, int>::iterator it = in->begin();
            while(it != in->end()){
                std::cout << "\"" << it->first << "\": " << it->second;
                ++it;
                if(it != in->end()){
                    std::cout << ", ";
                }
            }
            std::cout << "}"
                      << "}";
        }
        else if(typeid(*input) == typeid(Map<int, bool>)){
            std::cout << "{"
                      << "\"type\": \"Map<int, bool>\"";
            RCPtr<Map<int, bool> > in = input;
            std::cout << ", \"size\": " << in->size();

            std::cout << ", \"values\": {";
            Map<int, bool>::iterator it = in->begin();
            while(it != in->end()){
                std::cout << "\"" << it->first << "\": " << it->second;
                ++it;
                if(it != in->end()){
                    std::cout << ", ";
                }
            }
            std::cout << "}"
                      << "}";
        }
        else if(typeid(*input) == typeid(Map<int, std::complex<float> >)){
            std::cout << "{"
                      << "\"type\": \"Map<int, std::complex<float> >\"";
            RCPtr<Map<int, std::complex<float> > > in = input;
            std::cout << ", \"size\": " << in->size();

            std::cout << ", \"values\": {";
            Map<int, std::complex<float> >::iterator it = in->begin();
            while(it != in->end()){
                std::cout << "\"" << it->first << "\": " << it->second;
                ++it;
                if(it != in->end()){
                    std::cout << ", ";
                }
            }
            std::cout << "}"
                      << "}";
        }
        else if(typeid(*input) == typeid(Time)){
            std::cout << "{"
                      << "\"type\": \"Time\"";
            RCPtr<Time> in = input;
            std::cout << ", \"values\": " << _TimeLog(in->time);
            std::cout << "}";
        }
        else if (typeid(*input) == typeid(Source)) {
            std::cout << "{"
                      << "\"type\": \"Source\"";
            RCPtr<Source> in = input;
            std::cout << ", \"values\": {"
                      << "\"id\": " << in->id << ", "
                      << "\"x\": [" << in->x[0] << ", " << in->x[1] << ", " << in->x[2] << "], "
                      << "\"power\": " << in->power << ", "
                      << "\"remaining\": " << in->remaining << ", "
                      << "\"count\": " << in->count << ", "
                      << "\"last_found_count\": " << in->last_found_count << ", "
                      << "\"tfindex\": [" << in->tfindex[0] << ", " << in->tfindex[1] << ", " << in->tfindex[2] << "], "
                      << "\"tfindex_max\": [" << in->tfindex_max[0] << ", " << in->tfindex_max[1] << ", " << in->tfindex_max[2] << "], "
                      << "\"tfindex_min\": [" << in->tfindex_min[0] << ", " << in->tfindex_min[1] << ", " << in->tfindex_min[2] << "], "
                      << "\"compare_mode\": " << ((in->compare_mode == Source::DEG) ? "\"DEG\"" : "\"TFINDEX\"");
            std::cout << "}"
                      << "}";
        }
        else {
            std::cout << "{"
                      << "\"type\": \"unknown\""
                      << "}";
        }
    }

    void calculate(int output_id, int count, Buffer &out) {
        ObjectRef input(getInput(inputID, count));
        out[count] = input;

        std::cout << label << "[" << count << "]: " << typeid(*input).name() << ": ";
        this->dump(std::cout, input);
        std::cout << std::endl;
    }

    IN_ORDER_NODE_SPEEDUP(DataLogger);
};
