/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <iostream>
#include <memory>

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>
#include <Source.h>

#include "EstimationTransferFunction.hpp"

using namespace HARK;

class EstimateTF;

DECLARE_NODE(EstimateTF);
/*Node
 *
 * @name EstimateTF
 * @category HARK:TF
 * @description Sound Source Localization by BFDS
 *
 * @input_name INPUT
 * @input_type Matrix<complex<float> >
 * @input_description Input multi-channel spectrum. A row is a channel, and a column is a spectrum for the corresponding channel. 
 *
 * @input_name SOURCES
 * @input_type Vector<ObjectRef>
 * @input_description Source locations with ID. Each element of the vector is a source location with ID specified by "Source".
 *
 * @input_name INPUT_INPROGRESS
 * @input_type any
 * @input_description Flag in progress. If 0/false, archives to transfer function file.
 *
 * @output_name OUTPUT_INPROGRESS
 * @output_type any
 * @output_description Flag in progress.
 *
 * @parameter_name EST_TF_FILENAME
 * @parameter_type string
 * @parameter_description Name of a estimated transfer function file.
 * 
 * @parameter_name LENGTH
 * @parameter_type int
 * @parameter_value 512
 * @parameter_description The length of a frame (per channel).
 * 
 * @parameter_name SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_description Sampling rate (Hz).
 * 
 * @parameter_name INPUT_TYPE
 * @parameter_type string
 * @parameter_value TF
 * @parameter_list TF:INTERMEDIATE
 * @parameter_description The input method for based transfer function file. "TF" uses a transfer function file, and "INTERMEDIATE" adaptively estimates an optimal step size as LCMV_LC_MYU.
 *
 * @parameter_name BASE_TF_FILENAME
 * @parameter_type string
 * @parameter_valid INPUT_TYPE=TF
 * @parameter_description Name of based transfer function file. This is used only when INPUT_TYPE=TF.
 * 
 * @parameter_name IMPORT_INTERMEDIATE_DATA
 * @parameter_type string
 * @parameter_valid INPUT_TYPE=INTERMEDIATE
 * @parameter_description Directory of based intermediate files. This is used only when INPUT_TYPE=INTERMEDIATE.
 * 
 * @parameter_name EXPORT_INTERMEDIATE_DATA
 * @parameter_type string
 * @parameter_description Directory of intermediate files.
 * 
 * @parameter_name EXPORT_HISTORY_FILE 
 * @parameter_type bool
 * @parameter_value false 
 * @parameter_list true:false 
 * @parameter_description If true, outputs history file in the directory containing intermediate files. 
 * 
 * @parameter_name HISTORY_FILE_FRAMES
 * @parameter_type int
 * @parameter_value 10000
 * @parameter_valid EXPORT_HISTORY_FILE=true
 * @parameter_description The number of frame (per file). This is used only when EXPORT_HISTORY_FILE=true.
 * 
 * @parameter_name HISTORY_FILE_PERIOD
 * @parameter_type int
 * @parameter_value 1
 * @parameter_valid EXPORT_HISTORY_FILE=true
 * @parameter_description The number of frame (per file). This is used only when EXPORT_HISTORY_FILE=true.
 * 
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_description Debug option. If true, outputs debug to a standard output.
 * 
END*/

enum INPUT_TYPE {
  INPUT_TF, INPUT_INTERMEDIATE
};

class EstimateTF : public BufferedNode {
public:
	EstimateTF(string nodeName, ParameterSet params)
	  : BufferedNode(nodeName, params)
	{
		/*== Assign input/output ID ==*/

		this->inputID  = addInput("INPUT");
		this->inputSourceID = addInput("SOURCES");
		this->inputProgressID = addInput("INPUT_INPROGRESS");
		this->outputProgressID = addOutput("OUTPUT_INPROGRESS");

    	/*== Get parameters ==*/

		this->filenameEstTF = "";
		if (parameters.exist("EST_TF_FILENAME")) {
			this->filenameEstTF = object_cast<string>(parameters.get("EST_TF_FILENAME"));
		}
		if (this->filenameEstTF.empty()) {
			throw new NodeException(NULL, "EstimateTF: EST_TF_FILENAME is not set.", __FILE__, __LINE__);
		}

		int length = dereference_cast<int>(parameters.get("LENGTH"));
		if (length <= 0) {
			throw new NodeException(NULL, "EstimateTF: LENGTH is invalid value.", __FILE__, __LINE__);
		}

		int samplingRate = dereference_cast<int>(parameters.get("SAMPLING_RATE"));
		if (samplingRate <= 0) {
			throw (new NodeException(NULL, "EstimateTF: SAMPLING_RATE is invalid value.", __FILE__, __LINE__));
		}

		string inputType = object_cast<string>(parameters.get("INPUT_TYPE"));
		int input_type;
		if (inputType == "TF") { input_type = INPUT_TF; }
		else if (inputType == "INTERMEDIATE") { input_type = INPUT_INTERMEDIATE; }
		else { throw (new NodeException(NULL,"EstimateTF: Unexpected INPUT_TYPE is set.",__FILE__,__LINE__)); }

		string filenameBaseTF = "";
		if (parameters.exist("BASE_TF_FILENAME")) {
			filenameBaseTF = object_cast<string>(parameters.get("BASE_TF_FILENAME"));
		}
		if (input_type == INPUT_TF && filenameBaseTF.empty()) {
			throw new NodeException(NULL, "EstimateTF: BASE_TF_FILENAME is not set.", __FILE__, __LINE__);
		}

		string dirnameImport = "";
		if (parameters.exist("IMPORT_INTERMEDIATE_DATA")) {
			dirnameImport = object_cast<string>(parameters.get("IMPORT_INTERMEDIATE_DATA"));
		}
		if (input_type == INPUT_INTERMEDIATE && dirnameImport.empty()) {
			throw new NodeException(NULL, "EstimateTF: IMPORT_INTERMEDIATE_DATA is not set.", __FILE__, __LINE__);
		}

		string dirnameExport = "";
		if (parameters.exist("EXPORT_INTERMEDIATE_DATA")) {
			dirnameExport = object_cast<string>(parameters.get("EXPORT_INTERMEDIATE_DATA"));
		}

		bool isExport = dereference_cast<bool>(parameters.get("EXPORT_HISTORY_FILE"));

		if (isExport) {
			if (dirnameExport.empty()) {
				throw new NodeException(NULL, 
				      "EstimateTF: If export history data, set param EXPORT_INTERMEDIATE_DATA.", __FILE__, __LINE__);
			}
		}

		int historyFrames;
		if (parameters.exist("HISTORY_FILE_FRAMES")) {
			historyFrames = dereference_cast<int>(parameters.get("HISTORY_FILE_FRAMES"));
			if (historyFrames <= 0) {
				throw (new NodeException(NULL, 
				       "EstimateTF: HISTORY_FILE_FRAMES is invalid value.", __FILE__, __LINE__));
			}
		}

		int historyPeriod;
		if (parameters.exist("HISTORY_FILE_PERIOD")) {
			historyPeriod = dereference_cast<int>(parameters.get("HISTORY_FILE_PERIOD"));
			if (historyPeriod <= 0) {
				throw (new NodeException(NULL, 
				       "EstimateTF: HISTORY_FILE_PERIOD is invalid value.", __FILE__, __LINE__));
			}
		}

		bool isDebug = dereference_cast<bool>(parameters.get("DEBUG"));

		/*== Initialize ==*/

		this->estimater.reset(new EstimationTransferFunction(samplingRate, length, isDebug, 
		                                                     dirnameExport, isExport, historyFrames, historyPeriod));

		// get base data
		if (input_type == INPUT_TF) {
			// read TF from file and initialize intermediate data
			if (!this->estimater->getBasefromFile(filenameBaseTF, 
			                                      this->numFrequency, this->numChannel, this->numSource)) {
				throw (new NodeException(NULL, 
				       "EstimateTF: Failed to get base data from the file.", __FILE__, __LINE__));
			}
		} else { //input_type == INPUT_INTERMEDIATE
			// read TF and initermediate data in directory
			if (!this->estimater->getBaseinDirectory(dirnameImport, 
			                                         this->numFrequency, this->numChannel, this->numSource)) {
				throw (new NodeException(NULL, 
				       "EstimateTF: Failed to get base data in the directory.", __FILE__, __LINE__));
			}
		}
//cout << endl << "ch: " << numChannel << ", freq: " << numFrequency << ", source: " << numSource << endl << endl;

		// initialize history data
		if (isExport) { // !!
			this->estimater->initHistoryData();
		}

		// export
		if (!dirnameExport.empty()) {
			// Export TF file
			if (!this->estimater->exportTransferFunction(dirnameExport, true)) {
				throw new NodeException(NULL, "EstimateTF: Failed to export estimated TF file.", __FILE__, __LINE__);
			}

			// Export intermediate data
			if (!this->estimater->exportIntermData(dirnameExport, -1)) {
				throw (new NodeException(NULL, 
				       "EstimateTF: Failed to export intermediate data.", __FILE__, __LINE__));
			}

			if (isExport) {
				// Export history data
				if (!this->estimater->exportHistoryData(dirnameExport, -1)) {
					throw (new NodeException(NULL, "EstimateTF: Failed to export history file.", __FILE__, __LINE__));
				}
			}
		}

		// TFを転置してコピー
		this->estimater->copyKeepTF();

		inOrder = true;
	}

	virtual void initialize()
	{
		this->BufferedNode::initialize();
	} 

	void calculate(int output_id, int count, Buffer &out)
	{
//cout << count << endl;

		/*== set input/output ==*/

		// input spectrum
		RCPtr<Matrix<complex<float> > > input_p = getInput(this->inputID, count);

		// input source
		RCPtr<Vector<ObjectRef> > sources_p = getInput(this->inputSourceID, count);

		// input in-progress
		ObjectRef in_progress;
//		if (this->inputProgressID >= 0) {
			in_progress = getInput(this->inputProgressID, count);
//		}

		// output in-progress
		Buffer &outPbuffer = *(outputs[this->outputProgressID].buffer);
//		outPbuffer[count] = (this->outProgress ? TrueObject : FalseObject);
		outPbuffer[count] = (this->inProgress ? TrueObject : FalseObject);

		/*== input data ==*/

		// spectrum data
		if (count == 0) {
			if (this->numChannel != (*input_p).nrows()) {
				throw new NodeException(this, 
				      "EstimateTF: Input channel size is different from one in TF.", __FILE__, __LINE__);
			}
			if (this->numFrequency != (*input_p).ncols()) {
				throw new NodeException(this, 
				      "EstimateTF: Input frame size is different from one in TF.", __FILE__, __LINE__);
			}
		}
		BF::Matrix x(this->numChannel, this->numFrequency);
		for (int j = 0; j < this->numChannel; j++) {
			for (int k = 0; k < this->numFrequency; k++) {
				x.setElementAt(j, k, (*input_p)(j, k));
			}
		}

		// localized data
		const Vector<ObjectRef> *sources = sources_p.get();
		int numLocSource = sources->size();
		vector<BF::SourceLocation> sourceLocalized;
		for (int i = 0 ; i < numLocSource; i++) {
			RCPtr<Source> src = sources->at(i);
			if (src->compare_mode == Source::TFINDEX) {
				sourceLocalized.push_back(BF::SourceLocation(src->x[0], src->x[1], src->x[2], 
				                                             src->tfindex[0], src->id, true));
			} else { // Source::DEG
				sourceLocalized.push_back(BF::SourceLocation(src->x[0], src->x[1], src->x[2], 
				                                             -1, src->id, true));
			}
		}

		// in-progress data
		if (typeid(*in_progress) == typeid(Bool)) { 
			bool bInput = dereference_cast<bool>(in_progress); 
			this->inProgress = (bInput ? 1 : 0); 
		} else if (typeid(*in_progress) == typeid(Int)) {
			this->inProgress = dereference_cast<int> (in_progress);
		} else {
			throw new NodeException(this, 
			      "EstimateTF: OUTPUT_INPROGRESS input type is not appropriate.", __FILE__, __LINE__);
		} 

		if (this->inProgress == 0) {
			// Save TF file
			if (!this->estimater->exportTransferFunction(this->filenameEstTF, false)) {
				throw new NodeException(this, "EstimateTF: Failed to export estimated TF file.", __FILE__, __LINE__);
			}
			return;
		}

		// No localized data
		if (numLocSource == 0) { 
cout << "skip(1)" << endl;
			return; 
		}
		// input power is 0
		if (x.absoluteValues().getMaximumValue() == 0.0) {
cout << "skip(2)" << endl;
			return;		
		}

		/*== estimate ==*/

		this->estimater->estimateTansferFunction(count, x, sourceLocalized);
	}

private:
	// input/output

	int inputID;
	int inputSourceID;
	int inputProgressID;
	int outputProgressID;

	// parameter

	string filenameEstTF;

	// variable

	int numFrequency;
	int numChannel;
	int numSource;
	int inProgress = 0;
//	int outProgress = 0;

	std::unique_ptr<EstimationTransferFunction> estimater;

	//IN_ORDER_NODE_UP(EstimateTF);
};

