/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <string.h>

#ifdef ENABLE_HARKIO3
  #include "HarkFileIO_NH.h"
#elif  ENABLE_HARKIO2
  #include "HarkFileIO.h"
  using namespace harkio2;
#elif ENABLE_HARKIO1
  #include "HarkFileIO.h"
  using namespace harkio;
#endif

#include "HarkFileIOByMatrix.h"

#include "EstimationTransferFunction.hpp"

EstimationTransferFunction::EstimationTransferFunction(int sampling_rate, int fft_length, bool is_debug, 
  string dir_export, bool is_export, int history_frames, int history_period) :
  SamplingRate(sampling_rate), numFFT(fft_length), IsDebug(is_debug), 
  IsExport(is_export), DirExport(dir_export), HistoryFrames(history_frames), HistoryPeriod(history_period)
{
	this->numFreq = this->numFFT / 2 + 1;
}

//EstimationTransferFunction::~EstimationTransferFunction()
//{
//}

bool EstimationTransferFunction::getBasefromFile(const string &file_name, int &num_frq, int &num_mic, int &num_src)
{
	num_frq = 0;
	num_mic = 0;
	num_src = 0;

	CHarkFileIOByMatrix	cHarkFile;

	// read header
	string fileType = string("transfer function");
	long int hFs, hNFFT, hNSrc, hNMic;
	if (!cHarkFile.ReadHeader(file_name.c_str(), fileType.c_str(), &hFs, &hNFFT, &hNSrc, &hNMic)) {
		cerr << "EstimationTransferFunction: Failed to get header from the file." << endl;
		return false;
	}
	// check sampling rate
	if (hFs != this->SamplingRate) {
		cerr << "EstimationTransferFunction: Sampling rate is different between parameter and TF header." << endl;
		return false;
	}

	// read TF
	if (!cHarkFile.ReadTFData(file_name.c_str(), -1.0f, this->TransFuncDatabase)) {
		cerr << "EstimationTransferFunction: Failed to get transfer function from the file." << endl;
		return false;
	}

	// TF size is nfreq, element is Matrix(nmic, nsrc)

	int nfreq = this->TransFuncDatabase.Data.size();
	
	if (nfreq > 0) {
		// check number of freq. bin
		if (nfreq != this->numFreq) {
			cerr << "EstimationTransferFunction: Number of freq. bin is different between parameter and TF." << endl;
			return false;
		}

		this->numMic = this->TransFuncDatabase.Data[0].getRowCount();
		this->numSrc = this->TransFuncDatabase.Data[0].getColumnCount();
	} else {
		this->numMic = 0;
		this->numSrc = 0;
	}
	// check size of TF
	if (this->numMic == 0 || this->numSrc == 0) {
		cerr << "EstimationTransferFunction: Size of transfer function is invalid." << endl;
		return false;
	}

	num_frq = this->numFreq;
	num_mic = this->numMic;
	num_src = this->numSrc;
	
	// initialize intermediate data
	this->initIntermData();

	return true;
}

bool EstimationTransferFunction::getBaseinDirectory(const string &dir_name, int &num_frq, int &num_mic, int &num_src)
{
	num_frq = 0;
	num_mic = 0;
	num_src = 0;

	string spathname = dir_name + string("/source.xml");
 	char *spath = (char *)malloc((spathname.length() + 1) * sizeof(char));
	strcpy(spath, spathname.c_str());
	harkio_XML* xml = harkio_XML_newFromXML(spath, NULL);
	free(spath);
	if (!xml) {
		cerr << "EstimationTransferFunction: Failed to read source position file." << endl;
		return false;
	}
	// check sampling rate
	if (xml->cfg){
		if (xml->cfg->samplingRate != this->SamplingRate) {
			cerr << "EstimationTransferFunction: Sampling rate is different between parameter and xml." << endl;
			harkio_XML_delete(&xml);
			return false;
		}
	}
	harkio_XML_delete(&xml);

	// read TF
	CHarkFileIOByMatrix	cHarkFile;
	if (!cHarkFile.ReadTFDataFromDir(dir_name.c_str(), -1.0f, this->TransFuncDatabase)) {
		cerr << "EstimationTransferFunction: Failed to get transfer function in the directory." << endl;
		return false;
	}

	this->numMic = this->TransFuncDatabase.micLocations.size();
	this->numSrc = this->TransFuncDatabase.sourceLocations.size();

	// check number of microphone and source of xml
	if (this->numMic == 0 || this->numSrc == 0) {
		cerr << "EstimationTransferFunction: Number of mic./source positions is invalid." << endl;
		return false;
	}

	int nfreq = this->TransFuncDatabase.Data.size();
cout << "nfreq:" << nfreq << endl;

	if (nfreq > 0) {
		if (nfreq != this->numFreq) {
			cerr << "EstimationTransferFunction: Number of freq. bin is different between parameter and TF." << endl;
			return false;
		}

		// check number of microphne and source of TF
		if ((this->numMic != this->TransFuncDatabase.Data[0].getRowCount()) || 
		    (this->numSrc != this->TransFuncDatabase.Data[0].getColumnCount())) {
			cerr << "EstimationTransferFunction: Number of mic./source positions is different from TF." << endl;
			return false;
		}
	}
	else {
		this->TransFuncDatabase.Data.resize(this->numFreq, BF::Matrix::O(this->numMic, this->numSrc));
	}

	// read intermediate data
	string ipathname = dir_name + string("/intermediate.mat"); 
 	char *ipath = (char *)malloc((ipathname.length() + 1) * sizeof(char));
	strcpy(ipath, ipathname.c_str());
	CHarkMatFileIOByFMatrix cHarkMat;
	bool rtn = cHarkMat.ReadMatData(ipath, this->IntermData);
	free(ipath);
	if (rtn) {
		int nmic = (this->IntermData.getColumnCount() - 5) / (2 * 2 * this->numFreq);
		int nsrc = this->IntermData.getRowCount();
		// check size of intermediate data
		if (nmic != this->numMic || nsrc != this->numSrc) {
			cerr << "EstimationTransferFunction: Number of mic./source of intermediate data is invalid." << endl;
			return false;		
		}
	} else {
		cout << "EstimationTransferFunction: Initialize intermediate data." << endl;

		// initialize intermediate data
		this->initIntermData();
	}

	num_frq = this->numFreq;
	num_mic = this->numMic;
	num_src = this->numSrc;

	return true;
}

void EstimationTransferFunction::initIntermData()
{
	int nrow = this->numSrc;
	int ncol = 1 + 3 + 1 + this->numFreq * 2 * this->numMic + this->numFreq * 2 * this->numMic;
	this->IntermData = BF::FMatrix(nrow, ncol);
	
	for (int i = 0; i < nrow; i++) {
		int j = 0;
		this->IntermData.setElementAt(i, j++, this->TransFuncDatabase.sourceLocations[i].id);
		this->IntermData.setElementAt(i, j++, this->TransFuncDatabase.sourceLocations[i].x);
		this->IntermData.setElementAt(i, j++, this->TransFuncDatabase.sourceLocations[i].y);
		this->IntermData.setElementAt(i, j++, this->TransFuncDatabase.sourceLocations[i].z);
		for(; j < ncol; j++) {
			this->IntermData.setElementAt(i, j, 0.0f);
		}
	}
}

void EstimationTransferFunction::initHistoryData()
{
	int nrow = this->HistoryFrames;
	int ncol = 1 + 1 + 3 + 1 + this->numFreq * 2 * this->numMic + this->numFreq * 2 * this->numMic;
	this->HistoryData = BF::FMatrix(nrow, ncol);

	for (int i = 0; i < nrow; i++) {
		for(int j = 0; j < ncol; j++) {
			this->HistoryData.setElementAt(i, j, -1.0f);
		}
	}
	
	this->idxHist = -1; // index for ring
}

bool EstimationTransferFunction::exportIntermData(const string &dir_name, int i_row)
{
	string pathname = dir_name + string("/intermediate.mat");
 	char *path = (char *)malloc((pathname.length() + 1) * sizeof(char));
	strcpy(path, pathname.c_str());

	CHarkMatFileIOByFMatrix cHarkMat;
	bool rtn;
	if (i_row < 0) {
		rtn = cHarkMat.WriteMatData(path, this->IntermData);
	} else {
		rtn = cHarkMat.ReplaceMatData(path, this->IntermData, i_row);
	}
	
	free(path);	

	return rtn;
}

bool EstimationTransferFunction::exportHistoryData(const string &dir_name, int i_row)
{
	string pathname = dir_name + string("/history.mat");
 	char *path = (char *)malloc((pathname.length() + 1) * sizeof(char));
	strcpy(path, pathname.c_str());

	CHarkMatFileIOByFMatrix cHarkMat;
	bool rtn;
	if (i_row < 0) {
		rtn = cHarkMat.WriteMatData(path, this->HistoryData);
	} else {
		rtn = cHarkMat.ReplaceMatData(path, this->HistoryData, i_row);
	}

	free(path);
		
	return rtn;
}

bool EstimationTransferFunction::exportTransferFunction(const string &path, bool isDir)
{
	string comment = string("transferFunction");

	bool rtn;
	CHarkFileIOByMatrix	cHarkFile;
	if (isDir) {
		rtn = cHarkFile.WriteTFDataToDir(path.c_str(), this->SamplingRate, this->numFFT, comment.c_str(),
	                                     this->TransFuncDatabase);
	} else {
		rtn = cHarkFile.WriteTFData(path.c_str(), this->SamplingRate, this->numFFT, comment.c_str(),
	                                this->TransFuncDatabase);
	}

	return rtn;
}

void EstimationTransferFunction::copyKeepTF()
{
	// size(numFreq), element(numMic,numSrc) -> size(numSrc), element(numMic,numFreq)

	this->keepTFdata.resize(this->numSrc, BF::Matrix(this->numMic, this->numFreq));

	for(int j = 0; j < this->numFreq; j++) {
		for(int i = 0; i < this->numSrc; i++) {
			this->keepTFdata[i].setColumn(j, this->TransFuncDatabase.Data[j].getColumn(i));
		}
	}
}

BF::Matrix EstimationTransferFunction::_RelaTF(BF::Matrix X)
{	// 相対TFの算出
	
	int nrow = X.getRowCount();
	int ncol = X.getColumnCount();

	BF::Matrix Y0n(nrow, ncol);
	BF::Matrix Yr = BF::Matrix::O(nrow, ncol);
	BF::Complex Y0m, Y0p;

	for (int j = 0; j < ncol; j++) {
		Y0n.setColumn(j, X.getColumn(j) / X.getColumn(j).norm()); // 振幅正規化
		Y0m = Y0n.getColumn(j).getElementSum();
		Y0m /= nrow; // 平均位相
		Y0p = std::conj(Y0m / abs(Y0m)); // 振幅1にする
		Yr.setColumn(j, Y0n.getColumn(j) * Y0p); // 位相正規化
	}
	
	return Yr;
}

void EstimationTransferFunction::_getIntermediate(int tfId, float &id, float &x, float &y, float &z, float &frames, 
                                                  BF::Matrix &myu, BF::Matrix &sig)
{	// 中間データの取得
	
	int ncol1 = 5 + this->numFreq * 2 * this->numMic;
	int ncol2 = ncol1 + this->numFreq * 2 * this->numMic;
	int ii, jj;
	int c = 0;
	id     = this->IntermData.getElementAt(tfId, c++);
	x      = this->IntermData.getElementAt(tfId, c++);
	y      = this->IntermData.getElementAt(tfId, c++);
	z      = this->IntermData.getElementAt(tfId, c++);
	frames = this->IntermData.getElementAt(tfId, c++);
	ii = 0; jj = 0;
	for ( ; c < ncol1; ) {
		double re = this->IntermData.getElementAt(tfId, c++);
		double im = this->IntermData.getElementAt(tfId, c++);
		myu.setElementAt(ii, jj++, BF::Complex(re, im));
		if (jj == this->numFreq) { ii++; jj = 0; }
	}
	ii = 0; jj = 0;
	for ( ; c < ncol2; ) {
		double re = this->IntermData.getElementAt(tfId, c++);
		double im = this->IntermData.getElementAt(tfId, c++);
		sig.setElementAt(ii, jj++, BF::Complex(re, im));
		if (jj == this->numFreq) { ii++; jj = 0; }
	}
}

void EstimationTransferFunction::_setIntermediate(int tfId, float frames, BF::Matrix myu, BF::Matrix sig)
{	// 中間データの更新

	int c = 4;
	this->IntermData.setElementAt(tfId, c++, frames);
	for (int i = 0; i < this->numMic; i++) {
		for (int j = 0; j < this->numFreq; j++) {
			this->IntermData.setElementAt(tfId, c++, myu.getElementAt(i, j).real());
			this->IntermData.setElementAt(tfId, c++, myu.getElementAt(i, j).imag());
		}
	}
	for (int i = 0; i < this->numMic; i++) {
		for (int j = 0; j < this->numFreq; j++) {
			this->IntermData.setElementAt(tfId, c++, sig.getElementAt(i, j).real());
			this->IntermData.setElementAt(tfId, c++, sig.getElementAt(i, j).imag());
		}
	}
}

void EstimationTransferFunction::estimateTansferFunction(int count, BF::Matrix &x, 
                                                         vector<BF::SourceLocation> sources)
{
	BF::Matrix xx(x);	

	// 入力スペクトルから相対TFを算出
	BF::Matrix YR = _RelaTF(xx);//

	// TF更新処理

	BF::Matrix TF1(this->numMic, this->numFreq); // 更新TF
	
	float i_id;                                    // 当該座標の中間データ:TFID
	float i_x, i_y, i_z;                           //                   :x,y,z座標
	float i_frames;                                //                   :処理フレーム数
	BF::Matrix i_myu(this->numMic, this->numFreq); //                   :平均(Mic x Freq)
	BF::Matrix i_sig(this->numMic, this->numFreq); //                   :分散(Mic x Freq)
	
	for (int k = 0; k < sources.size(); k++) {
		int tfid = sources[k].id; // 更新対象tfid

		if (tfid > this->numSrc) { continue; };

		// 当該座標の中間データを取得
		_getIntermediate(tfid, i_id, i_x, i_y, i_z, i_frames, i_myu, i_sig);

		float inc_frames = i_frames + 1.0f; // 次のフレーム数
		
		if (i_frames > 0) {
			// 伝達関数の更新
			TF1 = this->keepTFdata[tfid]; // ノーマライズ前の伝達関数
			BF::Matrix TFa = YR;
			BF::Matrix TFp = TF1.elementwiseFMultiplied(TFa);
			BF::Complex Y0m, Y0p;
			for (int j = 0; j < this->numFreq; j++) {
				Y0m = TFp.getColumn(j).getElementSum(); // 更新前と本フレームの位相差
				Y0p = std::conj(Y0m / abs(Y0m)); // 振幅1にする
				TFa.setColumn(j, TFa.getColumn(j) * Y0p); // 更新量
			}
			TF1 = TF1 * i_frames;
			TF1 = TF1 + TFa;
			TF1 = TF1 / inc_frames; // フレームの加重をかける
			this->keepTFdata[tfid] = TF1; // ノーマライズ前のデータをとっておく			
			TF1 = _RelaTF(TF1); // 中間データに保存するのはノーマライズ後のもの

			// 分散の更新
			BF::Complex myuAt, sigAt, TFat;
			double sqr_old, sqr_new, myu_r, sig_r, myu_i, sig_i;
			for (int i = 0; i < this->numMic; i++) {
				for (int j = 0; j < this->numFreq; j++) {
					myuAt = i_myu.getElementAt(i, j);
					sigAt = i_sig.getElementAt(i, j);
					TFat = TF1.getElementAt(i, j);
					
					sqr_old = (sigAt.real() + myuAt.real() * myuAt.real()) * i_frames;
					sqr_new = sqr_old + TFat.real() * TFat.real();
					myu_r = (myuAt.real() * i_frames + TFat.real()) / inc_frames;
					sig_r = sqr_new / inc_frames - myu_r * myu_r;
					
					sqr_old = (sigAt.imag() + myuAt.imag() * myuAt.imag()) * i_frames;
					sqr_new = sqr_old + TFat.imag() * TFat.imag();
					myu_i = (myuAt.imag() * i_frames + TFat.imag()) / inc_frames;
					sig_i = sqr_new / inc_frames - myu_i * myu_i;
					
					i_myu.setElementAt(i, j, BF::Complex(myu_r, myu_i));
					i_sig.setElementAt(i, j, BF::Complex(sig_r, sig_i));
				}
			}
		}
		else {
			// 伝達関数
			TF1 = YR; // 最初はコピー
			this->keepTFdata[tfid] = TF1; // ノーマライズ前のデータをとっておく

			// 分散
			i_myu = TF1;
			i_sig = BF::Matrix::O(this->numMic, this->numFreq);
		}

//for (int i = 0; i < this->numMic; i++) {
//  for (int j = 0; j < this->numFreq; j++) {
//    cout << TF1.getElementAt(i, j) << ", ";
//    cout << i_myu.getElementAt(i, j) << ", ";
//    cout << i_sig.getElementAt(i, j) << ", ";
//  }
//  cout << endl;
//}
//cout << endl;

		// 伝達関数更新
		for(int j = 0; j < this->numFreq; j++) {
			this->TransFuncDatabase.Data[j].setColumn(tfid, TF1.getColumn(j));
		}

		// 中間データの更新
		_setIntermediate(tfid, inc_frames, i_myu, i_sig);

		if (!this->DirExport.empty()) {
			// 伝達関数のエクスポート
			CHarkFileIOByMatrix	cHarkFile;
			if (!cHarkFile.UpdateTFDataToDir(this->DirExport.c_str(), tfid, this->TransFuncDatabase)) {
				cerr << "EstimationTransferFunction: Failed to update TF." << endl;
			}

			// 中間データのエクスポート
			if (!exportIntermData(this->DirExport.c_str(), tfid)) {
				cerr << "EstimationTransferFunction: Failed to export intermediate data." << endl;
			}
		}
		
		// ヒストリデータ
		if (this->IsExport && (count % this->HistoryPeriod) == 0) {
			// ヒストリ追加			
			if (++this->idxHist == this->HistoryFrames) { this->idxHist = 0; }; // リングバッファ			
			this->HistoryData.setElementAt(this->idxHist, 0, count); // フレームインデクス			
			BF::FMatrix hDmat = this->IntermData.getRow(tfid);       // 更新した中間データを
			this->HistoryData.setSubmatrix(this->idxHist, 1, hDmat); //                   追加

			// エクスポート
			if (!exportHistoryData(this->DirExport, this->idxHist)) {
				cerr << "EstimationTransferFunction: Failed to export history data." << endl;
			}
		}
		
		if (this->IsDebug) {
			printf("(Frame %d) Estimate transfer function:\n", count);
			// 更新音源座標
			printf("Source: %d %f %f %f\n", (int)i_id, i_x, i_y, i_z);
			// 更新後の伝達関数
			BF::Complex TFat;
			printf("Updated TF:\n");
			for (int j = 0; j < this->numFreq; j++) {
				for (int i = 0; i < this->numMic; i++) {
					TFat = TF1.getElementAt(i, j);
					printf("%f+%fi ", TFat.real(), TFat.imag());
				}
			}
			printf("\n");
		}

	}
}
