/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include "HarkFileIOByMatrix.h"

#include <vector>
#include <string.h>
#include <iostream>
#include <sys/stat.h>
// --------------------------------------------------------------------

#ifdef ENABLE_HARKIO3
char HARKFILEIOBYMAT_SIGNATURE_TFMAT[] = "transfer function";
char HARKFILEIOBYMAT_SIGNATURE_SPMAT[] = "separation matrix";
char HARKFILEIOBYMAT_STRTYPE_COMPLEX[] = "complex";
char HARKFILEIOBYMAT_STRTYPE_FLOAT32[] = "float32";
#endif

// --------------------------------------------------------------------
CHarkFileIOByMatrix::CHarkFileIOByMatrix()
{
	m_pcTFfilename = NULL;	///< TFファイル名
	m_pcFileType = NULL;
#ifdef ENABLE_HARKIO3
	m_ptf = NULL;	///< TFファイルポインタ
#else
#endif
}
// --------------------------------------------------------------------
CHarkFileIOByMatrix::~CHarkFileIOByMatrix()
{
	CloseTFfile();
}
// --------------------------------------------------------------------
// tfファイルのオープン(cFilenameが既に開いている場合はスルーする)
bool CHarkFileIOByMatrix::OpenTFfile(const char* cFilename, const char* cFileType)
{
	if (m_pcTFfilename){
		if (strcmp(m_pcTFfilename, cFilename) == 0){
			// すでに開いているので何もしない
			if (strcmp(m_pcFileType, cFileType) != 0){
				return false;	// 読込みファイルタイプが異なる
			}
			return true;
		}
	}
	
	return (ReopenTFfile(cFilename, cFileType));
}
// --------------------------------------------------------------------
// tfディレクトリのオープン(ロード)(cDirnameが既に開いている場合はスルーする)
bool CHarkFileIOByMatrix::OpenTFfileFromDir(const char* cDirname, const char* cFileType)
{
#ifdef ENABLE_HARKIO3
	if (m_pcTFfilename){
		if (strcmp(m_pcTFfilename, cDirname) == 0){
			// すでに開いているので何もしない
			if (strcmp(m_pcFileType, cFileType) != 0){
				return false;	// 読込みファイルタイプが異なる
			}
			return true;
		}
	}
	
	return (ReopenTFfileFromDir(cDirname, cFileType));
#else
	return false;
#endif
}
// --------------------------------------------------------------------
// tfファイルの再オープン
bool CHarkFileIOByMatrix::ReopenTFfile(const char* cFilename, const char* cFileType)
{
	int i;
	
	if (m_pcTFfilename){
		CloseTFfile();
	}

#ifdef ENABLE_HARKIO3
#else
	if (strcmp(m_pcFileType, HARKFILEIOBYMAT_SIGNATURE_TFMAT) != 0 &&
		strcmp(m_pcFileType, HARKFILEIOBYMAT_SIGNATURE_SPMAT) != 0){
		return false;
	}
#endif
	
	m_pcTFfilename = (char *)malloc(strlen(cFilename)+1);
	strcpy(m_pcTFfilename, cFilename);
	m_pcFileType = (char *)malloc(strlen(cFileType)+1);
	strcpy(m_pcFileType, cFileType);
	
#ifdef ENABLE_HARKIO3
	m_ptf = harkio_TransferFunction_fromFile(m_pcTFfilename, m_pcFileType);
	if (!m_ptf){
		std::cerr << "Failed to open " << m_pcTFfilename << "@CHarkFileIOByMatrix::ReopenTFfile By" << m_pcFileType << std::endl;
		free(m_pcTFfilename);
		free(m_pcFileType);
		m_pcTFfilename = NULL;
		m_pcFileType = NULL;
		return false;
	}
#else
	if (!m_ctf.OpenTFfile(m_pcTFfilename)){
		std::cerr << "Failed to open " << m_pcTFfilename << "@CHarkFileIOByMatrix::ReopenTFfile By" << m_pcFileType << std::endl;
		free(m_pcTFfilename);
		free(m_pcFileType);
		m_pcTFfilename = NULL;
		m_pcFileType = NULL;
		return false;
	}
	if (strcmp(m_pcFileType, HARKFILEIOBYMAT_SIGNATURE_TFMAT) == 0){
		if (m_ctf.m_iFileType != FORMAT_TYPE_TF){
			std::cerr << "Failed to open " << m_pcTFfilename << "@CHarkFileIOByMatrix::ReopenTFfile By" << m_pcFileType << "(Format type error)" << std::endl;
			free(m_pcTFfilename);
			free(m_pcFileType);
			m_pcTFfilename = NULL;
			m_pcFileType = NULL;
			m_ctf.CloseTFfile();
			return false;
		}
	}
	else if (strcmp(m_pcFileType, HARKFILEIOBYMAT_SIGNATURE_SPMAT) == 0){
		if (m_ctf.m_iFileType != FORMAT_TYPE_W){
			std::cerr << "Failed to open " << m_pcTFfilename << "@CHarkFileIOByMatrix::ReopenTFfile By" << m_pcFileType << "(Format type error)" << std::endl;
			free(m_pcTFfilename);
			free(m_pcFileType);
			m_pcTFfilename = NULL;
			m_pcFileType = NULL;
			m_ctf.CloseTFfile();
			return false;
		}
	}
#endif
	
	return true;
}
// --------------------------------------------------------------------
// tfディレクトリの再オープン
bool CHarkFileIOByMatrix::ReopenTFfileFromDir(const char* cDirname, const char* cFileType)
{
	int i;
	
#ifdef ENABLE_HARKIO3
	if (m_pcTFfilename){
		CloseTFfile();
	}

	m_pcTFfilename = (char *)malloc(strlen(cDirname)+1);
	strcpy(m_pcTFfilename, cDirname);
	m_pcFileType = (char *)malloc(strlen(cFileType)+1);
	strcpy(m_pcFileType, cFileType);
	
try{
	m_ptf = harkio_TransferFunction_fromDirectory(m_pcTFfilename, m_pcFileType);
	if (!m_ptf){
		std::cerr << "Failed to open " << m_pcTFfilename << "@CHarkFileIOByMatrix::ReopenTFfile By" << m_pcFileType << std::endl;
		free(m_pcTFfilename);
		free(m_pcFileType);
		m_pcTFfilename = NULL;
		m_pcFileType = NULL;
		return false;
	}
}
catch(...){
	std::cerr << "Failed to open " << m_pcTFfilename << "@CHarkFileIOByMatrix::ReopenTFfile By" << m_pcFileType << std::endl;
	free(m_pcTFfilename);
	free(m_pcFileType);
	m_pcTFfilename = NULL;
	m_pcFileType = NULL;
	return false;
}
#else
	return false;
#endif
	return true;

}
// --------------------------------------------------------------------
// tfファイルのクローズ
bool CHarkFileIOByMatrix::CloseTFfile()
{
	if (m_pcTFfilename){
		free(m_pcTFfilename);
	}
	if (m_pcFileType){
		free(m_pcFileType);
	}
#ifdef ENABLE_HARKIO3
	if (m_ptf){
		harkio_TransferFunction_delete(&m_ptf);
	}	
#else
	m_ctf.CloseTFfile;
#endif

	// 初期化
	m_pcTFfilename = NULL;
	m_pcFileType = NULL;
#ifdef ENABLE_HARKIO3
	m_ptf = NULL;
#else
#endif
	return true;
}
// --------------------------------------------------------------------
// ファイル読み込み
bool CHarkFileIOByMatrix::ReadHeader(const char* cFilename, const char* cFileType,
				long *plFs,
				long *plNFFT,
				long *plNSrc,
				long *plNMic)
{
	try {
		int i;

		if (!OpenTFfile(cFilename, cFileType)){
			return false;
		}
		
#ifdef ENABLE_HARKIO3
		harkio_TransferFunction* tf = m_ptf;
		
		if (plFs){ *plFs = tf->cfg->samplingRate; }
		if (plNFFT){ *plNFFT = tf->cfg->nfft; }
		if (plNSrc){ *plNSrc = tf->poses->size; }
		if (plNMic){ *plNMic = tf->mics->size; }
#else
		MICARY_TFDBFileHeader sHeader;
		m_ctf.ReadHeader(cFilename, &sHeader);
		if (plFs){ *plFs = sHeader.lFs; }
		if (plNFFT){ *plNFFT = sHeader.lNFFT; }
		if (plNSrc){ *plNSrc = sHeader.lNSrc; }
		if (plNMic){ *plNMic = sHeader.lNMic; }
#endif
		
	}
	catch(std::exception &ex){
	
		return false ;
	
	}

	return true ;
}
// --------------------------------------------------------------------
// TFデータを読込み
bool CHarkFileIOByMatrix::ReadTFData(const char* cFilename, 
				float thresh,
				BF::TransformFunctionDatabase &sDatabase)
{
	if (!OpenTFfile(cFilename, HARKFILEIOBYMAT_SIGNATURE_TFMAT)){
		return false;
	}
	
	return(_ReadTFData(thresh, sDatabase));
}	
// --------------------------------------------------------------------
// TFデータをディレクトリから読込み
bool CHarkFileIOByMatrix::ReadTFDataFromDir(const char* cDirname, 
				float thresh,
				BF::TransformFunctionDatabase &sDatabase)
{
	if (!OpenTFfileFromDir(cDirname, HARKFILEIOBYMAT_SIGNATURE_TFMAT)){
		return false;
	}
	
	return(_ReadTFData(thresh, sDatabase));
}	
// --------------------------------------------------------------------
// TFデータを読込み
bool CHarkFileIOByMatrix::_ReadTFData(float thresh,
				BF::TransformFunctionDatabase &sDatabase)
{
	if (!m_pcTFfilename){
		return false;
	}

#ifdef ENABLE_HARKIO3
	try {
		harkio_TransferFunction* tf = m_ptf;
		int i, j, k;
		float fcoord[3];
		float *pfc;
		COMPLEX_TYPE value;
		BF::Complex	inputvalue;

		// Mic情報の取得
		sDatabase.micLocations.clear();
		if (tf->mics->size > 0){
		sDatabase.micLocations.resize(tf->mics->size, BF::MicLocation(0,0,0,0));
		for(i = 0; i < tf->mics->size ; i++){
			if (tf->mics->pos[i]->sys != Cartesian){
				harkio_Position_Convert(tf->mics->pos[i], fcoord, Cartesian);
				pfc = fcoord;
			}
			else {
				pfc = tf->mics->pos[i]->coord;
			}
			BF::MicLocation mictmp(pfc[0], pfc[1], pfc[2], tf->mics->pos[i]->id);
//			sDatabase.micLocations.push_back(mictmp);
			sDatabase.micLocations[i] = mictmp;
		}
		}
		
		// Src情報の取得
		sDatabase.sourceLocations.clear();
		if (tf->poses->size > 0){
		sDatabase.sourceLocations.resize(tf->poses->size, BF::SourceLocation(0,0,0,0,0,false));
		for(i = 0; i < tf->poses->size ; i++){
			if (tf->poses->pos[i]->sys != Cartesian){
				harkio_Position_Convert(tf->poses->pos[i], fcoord, Cartesian);
				pfc = fcoord;
			}
			else {
				pfc = tf->poses->pos[i]->coord;
			}
			BF::SourceLocation srctmp(pfc[0], pfc[1], pfc[2], tf->poses->pos[i]->id, tf->poses->pos[i]->id, true);
//			sDatabase.sourceLocations.push_back(srctmp);
			sDatabase.sourceLocations[i] = srctmp;
		}
		}
		// TFデータの取得
		// size_is(numFreq), elemnent_is(Matrix(numMic, numSrc)) 
		sDatabase.Data.clear();
		if (tf->cfg->nfft > 0){
		sDatabase.Data.resize(tf->cfg->nfft/2+1);
		for (k = 0 ; k < tf->cfg->nfft/2+1 ; k++){
			BF::Matrix m = BF::Matrix::O(tf->mics->size, tf->poses->size);
			if (tf->loctfs){
			for (j = 0 ; j < tf->poses->size ; j++){
				if (tf->loctfs[j]){
				for (i = 0 ; i < tf->mics->size ; i++){
//					harkio_Matrix_getValueComplex(tf->septfs[j], i, k, &value);
////					inputvalue = value;	// H=conj(TFCONJ)
////					inputvalue = conj(value);	// H=conj(TFCONJ)
//					inputvalue = std::conj(std::complex<double>(value));	// H=conj(TFCONJ)
//					m.setElementAt(i, j, inputvalue);
					harkio_Matrix_getValueComplex(tf->loctfs[j], i, k, &value);
					m.setElementAt(i, j, value);
				}
				}
			}
			}
//			sDatabase.Data.push_back(m);
			sDatabase.Data[k] = m;
		}
		}
		
		// Neighbors取得
		if (thresh < 0){
			sDatabase.sourceNeighbors.clear();
			if (tf->nbrs->size > 0){
			sDatabase.sourceNeighbors.resize(tf->nbrs->size);
			for (i=0; i<tf->nbrs->size; i++){
				sDatabase.sourceNeighbors[i].resize(tf->nbrs->numNeighbors[i]);
				for (j=0; j<tf->nbrs->numNeighbors[i]; j++){
					sDatabase.sourceNeighbors[i][j] = tf->nbrs->neighbors[i][j];
				}
			}
			}
			else {
			_CalcNeighbors(sDatabase.sourceLocations, thresh, sDatabase.sourceNeighbors);
			}
		}
		else {
			_CalcNeighbors(sDatabase.sourceLocations, thresh, sDatabase.sourceNeighbors);
		}
	}
	catch(std::exception &ex){
		return false ;
	}
#else
    /* HARK 形式のデータベース読み込み */
    MICARY_TFDBFileHeader tfdbh;
    CHarkFileIO harkFileIO;
    if(!harkFileIO.ReadHeader(this->tfFileName.c_str(), &tfdbh)){
		std::cerr << "Failed to open " << m_pcTFfilename << "@CHarkFileIOByMatrix::ReadTFData By" << m_pcFileType << "(ReadHeader error)" << std::endl;
		return false;
	}
	

    MICARY_TFDatabase tfdb;
    TFDatabaseAlloc(&tfdb, tfdbh.lNFreq, tfdbh.lNSrc, tfdbh.lNMic);
    if(!harkFileIO.ReadInfo(this->tfFileName.c_str(), &tfdbh, &tfdb.micary_locaSource, &tfdb.micary_locaMic, &tfdb.micary_iaFreqLine)
       || !harkFileIO.ReadData(this->tfFileName.c_str(), &tfdb.micary_cmpaTF)){
		std::cerr << "Failed to open " << m_pcTFfilename << "@CHarkFileIOByMatrix::ReadTFData By" << m_pcFileType << "(ReadInfo/Data error)" << std::endl;
		return false;
    }

    /* データベース形式を変換 */
    int numSrc  = tfdb.micary_locaSource.ha_numLocation;
    int numMic  = tfdb.micary_locaMic.ha_numLocation;
    int numFreq = tfdb.micary_iaFreqLine.ha_numValue;
    sDatabase.sourceLocations.clear();
    for(int i = 0; i < numSrc; i++){
		MICARY_Location &l = tfdb.micary_locaSource.micary_Location[i];
		BF::SourceLocation sl(l.ha_fX, l.ha_fY, l.ha_fZ, tfdb.micary_locaSource.ha_idLocation[i], 
							  tfdb.micary_locaSource.ha_idLocation[i], true);
		sDatabase.sourceLocations.push_back(sl);
    }
    sDatabase.micLocations.clear();
    for(int i = 0; i < numMic; i++){
		MICARY_Location &l = tfdb.micary_locaMic.micary_Location[i];
		BF::MicLocation ml(l.ha_fX, l.ha_fY, l.ha_fZ, tfdb.micary_locaMic.ha_idLocation[i]);
		sDatabase.micLocations.push_back(ml);
    }
    sDatabase.Data.clear();
    for(int i = 0; i < numFreq; i++){
		BF::Matrix tf(numMic, numSrc);

		for(int j = 0; j < numMic; j++){
		    for(int k = 0; k < numSrc; k++){
				// micary_cmpaData->ha_cmpValueは(周波数(音源(マイク)))の順に入力
				HA_Complex &v =tfdb.micary_cmpaTF.ha_cmpValue[(i * numSrc + k) * numMic + j];
				// H=conj(TFCONJ)
				tf.setElementAt(j, k, v.re, -v.im);
		    }
		}
		sDatabase.Data.push_back(tf);
    }
	TFDatabaseFree(&tfdb);

	// Neighbors取得
	_CalcNeighbors(sDatabase.sourceLocations, thresh, sDatabase.sourceNeighbors);
#endif
	return true;
}
// --------------------------------------------------------------------
// TFデータを書き出し(2020/03/16new!)
bool CHarkFileIOByMatrix::WriteTFData(const char* cFilename, 
				long lFs,
				long lNFFT,
				const char* comment,
				BF::TransformFunctionDatabase sDatabase)
{
#ifdef ENABLE_HARKIO3
	try {
		int iRtn;
		harkio_TransferFunction *tf = NULL;
		if (!_WriteTFData(sDatabase, lFs, lNFFT, comment, &tf)){
			std::cerr << "Failed to convert BF::TransformFunctionDatabase to harkio_Transferfunction @CHarkFileIOByMatrix::WriteTFData" << std::endl;
			return false;
		}
		if (!tf){
			std::cerr << "Failed to create harkio_Transferfunction @CHarkFileIOByMatrix::WriteTFData" << std::endl;
			return false;
		}
	
		char *cName = (char*)malloc((strlen(cFilename)+1)*sizeof(char));
		strcpy(cName, cFilename);
		iRtn = harkio_TransferFunction_write(tf, cName, HARKFILEIOBYMAT_SIGNATURE_TFMAT);
		free(cName);
		
		harkio_TransferFunction_delete(&tf);
		
		if (iRtn == EXIT_FAILURE){
			return false;
		}
	}
	catch (std::exception &ex){
		return false;
	}
#else
	return false;
#endif
	return true;

}
// --------------------------------------------------------------------
// TFデータをディレクトリに書き出し(2020/03/16new!)
bool CHarkFileIOByMatrix::WriteTFDataToDir(const char* cDirname, 
				long lFs,
				long lNFFT,
				const char* comment,
				BF::TransformFunctionDatabase sDatabase)
{
#ifdef ENABLE_HARKIO3
	try {
		int iRtn;
		harkio_TransferFunction *tf = NULL;
		if (!_WriteTFData(sDatabase, lFs, lNFFT, comment, &tf)){
			std::cerr << "Failed to convert BF::TransformFunctionDatabase to harkio_Transferfunction @CHarkFileIOByMatrix::WriteTFDataToDir" << std::endl;
			return false;
		}
		if (!tf){
			std::cerr << "Failed to create harkio_Transferfunction @CHarkFileIOByMatrix::WriteTFDataToDir" << std::endl;
			return false;
		}
	
		char *cName = (char*)malloc((strlen(cDirname)+1)*sizeof(char));
		strcpy(cName, cDirname);
		iRtn = harkio_TransferFunction_write_toDirectory(tf, cName, HARKFILEIOBYMAT_SIGNATURE_TFMAT);
		free(cName);
		
		harkio_TransferFunction_delete(&tf);
		
		if (iRtn == EXIT_FAILURE){
			return false;
		}
	}
	catch (std::exception &ex){
		return false;
	}
#else
	return false;
#endif
	return true;
}
// --------------------------------------------------------------------
// TFデータを書き出し
bool CHarkFileIOByMatrix::_WriteTFData(BF::TransformFunctionDatabase sDatabase,
				long lFs,
				long lNFFT,
				const char* comment,
				harkio_TransferFunction **tf)
{
#ifdef ENABLE_HARKIO3
	try{
		int i, j, k;
		float  coord[3];
		COMPLEX_TYPE inputvalue;
		
		// TFとして保存する
		*tf = harkio_TransferFunction_new();
		if (!(*tf)){
			std::cerr << "Failed to create harkio_Transferfunction @CHarkFileIOByMatrix::_WriteTFData" << std::endl;
			return false;
		}

		// config
		harkio_Config_setSynchronousAverage((*tf)->cfg, 1);
		harkio_Config_setTSPpath((*tf)->cfg, "");
		harkio_Config_setTSPoffset((*tf)->cfg, 0);
		harkio_Config_setPeakSearchFrom((*tf)->cfg, 0);
		harkio_Config_setPeakSearchTo((*tf)->cfg, 0);
		harkio_Config_setNfft((*tf)->cfg, lNFFT);
		harkio_Config_setComment((*tf)->cfg, comment);
		harkio_Config_setSamplingRate((*tf)->cfg, lFs);
		harkio_Config_setTSPlength((*tf)->cfg, 1);
		harkio_Config_setSignalMax((*tf)->cfg, 1);

		harkio_Position *pos;

		// 音源位置の保存
	    (*tf)->poses->type = (char*) malloc(sizeof(char) * (strlen("soundsource") + 1));
		strcpy((*tf)->poses->type, "soundsource");
		for (i = 0 ; i < sDatabase.sourceLocations.size() ; i++){
			coord[0] = sDatabase.sourceLocations[i].x;
			coord[1] = sDatabase.sourceLocations[i].y;
			coord[2] = sDatabase.sourceLocations[i].z;
			pos = harkio_Position_new(sDatabase.sourceLocations[i].id, Cartesian, coord, NULL);
			harkio_Positions_append((*tf)->poses, pos);
		}

		// Neighborsの設定
		int *neighbors;
		if (sDatabase.sourceNeighbors.size() == sDatabase.sourceLocations.size()){
			for (i = 0; i < sDatabase.sourceNeighbors.size(); i++){
				neighbors = NULL;
				if (sDatabase.sourceNeighbors[i].size() > 0){
					neighbors = (int*)malloc(sizeof(int) * sDatabase.sourceNeighbors[i].size());
					for (k = 0; k<sDatabase.sourceNeighbors[i].size() ; k++){
						neighbors[k] = sDatabase.sourceNeighbors[i][k];
					}
				}
				harkio_Neighbors_append((*tf)->nbrs, sDatabase.sourceLocations[i].id, sDatabase.sourceNeighbors[i].size(), neighbors);
				(*tf)->nbrs->alg = NearestNeighbor;
				if (neighbors){
					free(neighbors);
				}
			}
		}
		else {
			if ((*tf)->nbrs){
				harkio_Neighbors_delete(&(*tf)->nbrs);
				(*tf)->nbrs = NULL;
			}
			(*tf)->nbrs = harkio_Neighbors_new((*tf)->poses);
			harkio_Neighbors_calcNeighbors((*tf)->nbrs, NearestNeighbor, 1, 300.0);
		}

		// マイク位置の保存
	    (*tf)->mics->type = (char*) malloc(sizeof(char) * (strlen("microphone") + 1));
		strcpy((*tf)->mics->type, "microphone");
		for (i = 0 ; i < sDatabase.micLocations.size() ; i++){
			coord[0] = sDatabase.micLocations[i].x;
			coord[1] = sDatabase.micLocations[i].y;
			coord[2] = sDatabase.micLocations[i].z;
			pos = harkio_Position_new(sDatabase.micLocations[i].id, Cartesian, coord, NULL);
			harkio_Positions_append((*tf)->mics, pos);
		}

		// loctfsに保存
		(*tf)->loctfs = (harkio_Matrix **) malloc(sizeof(harkio_Matrix *) * (*tf)->poses->size);
		for (i = 0 ; i < (*tf)->poses->size ; i++){
			(*tf)->loctfs[i] = harkio_Matrix_new(HARKFILEIOBYMAT_STRTYPE_COMPLEX, (*tf)->mics->size, lNFFT/2+1);
		}
		// TF:sDatabase.Data: size_is(numFreq), elemnent_is(Matrix(numMic, numSrc))
		for (k=0 ; k < lNFFT/2+1 ; k++){
			for (i = 0 ; i < (*tf)->mics->size ; i++){
				for (j = 0 ; j < (*tf)->poses->size ; j++){
					BF::Complex cvalue = sDatabase.Data[k].getElementAt(i, j);
					inputvalue =  cvalue.__rep();
					harkio_Matrix_setValueComplex((*tf)->loctfs[j], i, k, inputvalue);
				}
			}
		}

		// septfsに保存(TFデータの共役複素数)
		(*tf)->septfs = (harkio_Matrix **) malloc(sizeof(harkio_Matrix *) * (*tf)->poses->size);
		for (i = 0 ; i < (*tf)->poses->size ; i++){
			(*tf)->septfs[i] = harkio_Matrix_new(HARKFILEIOBYMAT_STRTYPE_COMPLEX, (*tf)->mics->size, lNFFT/2+1);
		}
		// TF:sDatabase.Data: size_is(numFreq), elemnent_is(Matrix(numMic, numSrc))
		for (k=0 ; k < lNFFT/2+1 ; k++){
			for (i = 0 ; i < (*tf)->mics->size ; i++){
				for (j = 0 ; j < (*tf)->poses->size ; j++){
					BF::Complex cvalue = sDatabase.Data[k].getElementAt(i, j);
					inputvalue =  std::conj(cvalue).__rep();
					harkio_Matrix_setValueComplex((*tf)->septfs[j], i, k, inputvalue);
				}
			}
		}
	}
	catch (std::exception &ex){
		std::cerr << "Failed to convert BF::TransformFunctionDatabase to harkio_Transferfunction @CHarkFileIOByMatrix::_WriteTFData(0)" << std::endl;
		return false;
	}
#else
	return false;
#endif
	return true;
}
// --------------------------------------------------------------------
// TFデータをディレクトリに書き出し(2020/03/16new!)
// 指定した音源位置の伝達関数を更新
bool CHarkFileIOByMatrix::UpdateTFDataToDir(const char* cDirname, 
				int isrcindex,
				BF::TransformFunctionDatabase sDatabase)
{
#ifdef ENABLE_HARKIO3
	try{
	    struct stat statbuffer;
		char loc[4192];
		char sep[4192];
		char tmppath[4192];
		harkio_Matrix *mat;
		int i, k;
		COMPLEX_TYPE inputvalue;
		int iRtn;
		int iMics, iFreqs;
		
		if (sDatabase.Data.size() == 0){
			std::cerr << "sDatabase have no data @CHarkFileIOByMatrix::UpdateTFDataToDir" << std::endl;
			return false;
		}
		if (isrcindex < 0){
			std::cerr << "isrcindex is smaller than 0 @CHarkFileIOByMatrix::UpdateTFDataToDir" << std::endl;
			return false;
		}
		if (sDatabase.Data[0].getColumnCount() <= isrcindex){
			std::cerr << "isrcindex is larger than matrix size @CHarkFileIOByMatrix::UpdateTFDataToDir" << std::endl;
			return false;
		}

    	sprintf(loc, "%s/localization", cDirname);
    	sprintf(sep, "%s/separation", cDirname);

    	// make directory
    	if (stat(cDirname, &statbuffer) != 0){
			mkdir(cDirname, 0775);
    	}
    	if (stat(loc, &statbuffer) != 0){
			mkdir(loc, 0775);
    	}
    	if (stat(sep, &statbuffer) != 0){
			mkdir(sep, 0775);
    	}

		iMics= sDatabase.micLocations.size();
		iFreqs = sDatabase.Data.size();
		mat = harkio_Matrix_new(HARKFILEIOBYMAT_STRTYPE_COMPLEX, iMics, iFreqs);
		if (!mat){
			std::cerr << "Failed to create harkio_Matrix @CHarkFileIOByMatrix::UpdateTFDataToDir" << std::endl;
			return false;
		}
		
		// 伝達関数の保存
		// sDatabase.Data: size_is(numFreq), elemnent_is(Matrix(numMic, numSrc))
		for (k=0 ; k < iFreqs ; k++){
			for (i = 0 ; i < iMics ; i++){
				BF::Complex cvalue = sDatabase.Data[k].getElementAt(i, isrcindex);
				inputvalue =  cvalue.__rep();
				harkio_Matrix_setValueComplex(mat, i, k, inputvalue);
			}
		}
		sprintf(tmppath, "%s/tf%05d.mat", loc, isrcindex);
		iRtn = harkio_Matrix_write(mat, tmppath);
		if (iRtn == EXIT_FAILURE){
			std::cerr << "Failed to write harkio_Matrix(loctf) @CHarkFileIOByMatrix::UpdateTFDataToDir" << std::endl;
			return false;
		}

		// 分離用伝達関数に保存(TFデータの共役複素数)
		// sDatabase.Data: size_is(numFreq), elemnent_is(Matrix(numMic, numSrc))
		for (k=0 ; k < iFreqs ; k++){
			for (i = 0 ; i < iMics ; i++){
				BF::Complex cvalue = sDatabase.Data[k].getElementAt(i, isrcindex);
				inputvalue =  std::conj(cvalue).__rep();
				harkio_Matrix_setValueComplex(mat, i, k, inputvalue);
			}
		}
		sprintf(tmppath, "%s/tf%05d.mat", sep, isrcindex);
		iRtn = harkio_Matrix_write(mat, tmppath);
		if (iRtn == EXIT_FAILURE){
			std::cerr << "Failed to write harkio_Matrix(septf) @CHarkFileIOByMatrix::UpdateTFDataToDir" << std::endl;
			return false;
		}

	}
	catch (std::exception &ex){
		return false;
	}
#else
	return false;
#endif
	return true;
}
// --------------------------------------------------------------------
// Wデータを読込み
bool CHarkFileIOByMatrix::ReadWData(const char* cFilename, 
				BF::Database &sDatabase)
{
	if (!OpenTFfile(cFilename, HARKFILEIOBYMAT_SIGNATURE_SPMAT)){
		return false;
	}
	
#ifdef ENABLE_HARKIO3
	try {
		harkio_TransferFunction* tf = m_ptf;
		int i, j, k;
		float fcoord[3];
		float *pfc;
		COMPLEX_TYPE value;
		BF::Complex	inputvalue;
		// Mic情報の取得
		sDatabase.micLocations.clear();
		for(i = 0; i < tf->mics->size ; i++){
			if (tf->mics->pos[i]->sys != Cartesian){
				harkio_Position_Convert(tf->mics->pos[i], fcoord, Cartesian);
				pfc = fcoord;
			}
			else {
				pfc = tf->mics->pos[i]->coord;
			}
			BF::MicLocation mictmp(pfc[0], pfc[1], pfc[2], tf->mics->pos[i]->id);
			sDatabase.micLocations.push_back(mictmp);
		}
		
		// Src情報の取得
		sDatabase.sourceLocations.clear();
		for(i = 0; i < tf->poses->size ; i++){
			if (tf->poses->pos[i]->sys != Cartesian){
				harkio_Position_Convert(tf->poses->pos[i], fcoord, Cartesian);
				pfc = fcoord;
			}
			else {
				pfc = tf->poses->pos[i]->coord;
			}
			BF::SourceLocation srctmp(pfc[0], pfc[1], pfc[2], tf->poses->pos[i]->id, tf->poses->pos[i]->id, true);
			sDatabase.sourceLocations.push_back(srctmp);
		}
		
		// Wデータの取得
		// size_is(numSrc), elemnent_is(Matrix(numMic, numFreq))
		sDatabase.Data.clear();
		for (j = 0 ; j < tf->poses->size ; j++){
			BF::Matrix m = BF::Matrix::O(tf->mics->size, tf->cfg->nfft/2+1);
			for (k = 0 ; k < tf->cfg->nfft/2+1 ; k++){
				for (i = 0 ; i < tf->mics->size ; i++){
					harkio_Matrix_getValueComplex(tf->septfs[j], i, k, &value);
					inputvalue = value;
					m.setElementAt(i, k, inputvalue);
				}
			}
			sDatabase.Data.push_back(m);
		}	
	}
	catch(std::exception &ex){
		return false ;
	}
#else
    /* HARK 形式のデータベース読み込み */
    MICARY_TFDBFileHeader tfdbh;
    CHarkFileIO harkFileIO;
    if(!harkFileIO.ReadHeader(this->tfFileName.c_str(), &tfdbh)){
		std::cerr << "Failed to open " << m_pcTFfilename << "@CHarkFileIOByMatrix::ReadTFData By" << m_pcFileType << "(ReadHeader error)" << std::endl;
		return false;
	}
	
    MICARY_TFDatabase tfdb;
    TFDatabaseAlloc(&tfdb, tfdbh.lNFreq, tfdbh.lNSrc, tfdbh.lNMic);
    if(!harkFileIO.ReadInfo(this->tfFileName.c_str(), &tfdbh, &tfdb.micary_locaSource, &tfdb.micary_locaMic, &tfdb.micary_iaFreqLine)
       || !harkFileIO.ReadData(this->tfFileName.c_str(), &tfdb.micary_cmpaTF)){
		std::cerr << "Failed to open " << m_pcTFfilename << "@CHarkFileIOByMatrix::ReadTFData By" << m_pcFileType << "(ReadInfo/Data error)" << std::endl;
		return false;
    }

    /* データベース形式を変換 */
    int numSrc  = tfdb.micary_locaSource.ha_numLocation;
    int numMic  = tfdb.micary_locaMic.ha_numLocation;
    int numFreq = tfdb.micary_iaFreqLine.ha_numValue;
    sDatabase.sourceLocations.clear();
    for(int i = 0; i < numSrc; i++){
		MICARY_Location &l = tfdb.micary_locaSource.micary_Location[i];
		BF::SourceLocation sl(l.ha_fX, l.ha_fY, l.ha_fZ, tfdb.micary_locaSource.ha_idLocation[i], 
							  tfdb.micary_locaSource.ha_idLocation[i], true);
		sDatabase.sourceLocations.push_back(sl);
    }
    sDatabase.micLocations.clear();
    for(int i = 0; i < numMic; i++){
		MICARY_Location &l = tfdb.micary_locaMic.micary_Location[i];
		BF::MicLocation ml(l.ha_fX, l.ha_fY, l.ha_fZ, tfdb.micary_locaMic.ha_idLocation[i]);
		sDatabase.micLocations.push_back(ml);
    }
    sDatabase.Data.clear();
    // size_is(numSrc), elemnent_is(Matrix(numMic, numFreq))
    for(int k = 0; k < numSrc; k++){
		BF::Matrix tf = Matrix::O(numMic, numFreq);
	    for(int i = 0; i < numFreq; i++){
			for(int j = 0; j < numMic; j++){
				// micary_cmpaData->ha_cmpValueは(周波数(マイク(音源)))の順に入力
				HA_Complex &v =tfdb.micary_cmpaTF.ha_cmpValue[(i * numMic + j) * numSrc + k];
				tf.setElementAt(k, j, v.re, v.im);
		    }
		}
		sDatabase.Data.push_back(tf);
    }
	TFDatabaseFree(&tfdb);
#endif
	return true;
}
// --------------------------------------------------------------------
// Neighborsリスト取得
void CHarkFileIOByMatrix::_CalcNeighbors(std::vector<BF::SourceLocation> sourceLocations, float thresh, std::vector<std::vector<int> > &viNbrs)
{
	std::vector<std::pair<float, int> > vifclosestIDs;
	int i, j;

	viNbrs.clear();
	viNbrs.resize(sourceLocations.size());
	
	for (i = 0 ; i < sourceLocations.size() ; i++){
		// 初期化
		vifclosestIDs.clear();
		
		// threshould以内を検索
		for(j = 0 ; j < sourceLocations.size() ; j++)
		{
			float tmpdist = sourceLocations[i].euclidDistance(sourceLocations[j]);
			if (tmpdist < thresh){
				vifclosestIDs.push_back(std::pair<float, int>(tmpdist, sourceLocations[j].id));
			}
   	    }
		
		// sort
		sort(vifclosestIDs.begin(), vifclosestIDs.end());

		// Neighbor構造体内に保存
		viNbrs[i].resize(vifclosestIDs.size());
		for (j = 0 ; j < vifclosestIDs.size() ; j++){
			viNbrs[i][j] = vifclosestIDs[j].second;
		}
		
	}
	
}
// --------------------------------------------------------------------
// データ書き込み
bool CHarkFileIOByMatrix::WriteWData(const char* cFilename, 
				long lFs,
				long lNFFT,
				BF::Database sDatabase)
{
#ifdef ENABLE_HARKIO3
	try{
		int i, j, k;
		float  coord[3];
		COMPLEX_TYPE inputvalue;

		// Wとして保存する
		harkio_TransferFunction *tf = harkio_TransferFunction_new();
		if (!tf){
			std::cerr << "Failed to create harkio_Transferfunction @CHarkFileIOByMatrix::WriteWData" << std::endl;
			return false;
		}

		// config
		tf->cfg->samplingRate = lFs;
		tf->cfg->nfft = lNFFT;

		harkio_Position *pos;

		// 音源位置の保存
	        tf->poses->type = (char*) malloc(sizeof(char) * (strlen("soundsource") + 1));
		strcpy(tf->poses->type, "soundsource");
		for (i = 0 ; i < sDatabase.sourceLocations.size() ; i++){
			coord[0] = sDatabase.sourceLocations[i].x;
			coord[1] = sDatabase.sourceLocations[i].y;
			coord[2] = sDatabase.sourceLocations[i].z;
			pos = harkio_Position_new(sDatabase.sourceLocations[i].id, Cartesian, coord, NULL);
			harkio_Positions_append(tf->poses, pos);
		}

		// Neighborsの設定
		if (tf->nbrs){
			harkio_Neighbors_delete(&tf->nbrs);
			tf->nbrs = NULL;
		}
		tf->nbrs = harkio_Neighbors_new(tf->poses);
		harkio_Neighbors_calcNeighbors(tf->nbrs, NearestNeighbor, 1, 300.0);

		// マイク位置の保存
	        tf->mics->type = (char*) malloc(sizeof(char) * (strlen("microphone") + 1));
		strcpy(tf->mics->type, "microphone");
		for (i = 0 ; i < sDatabase.micLocations.size() ; i++){
			coord[0] = sDatabase.micLocations[i].x;
			coord[1] = sDatabase.micLocations[i].y;
			coord[2] = sDatabase.micLocations[i].z;
			pos = harkio_Position_new(sDatabase.micLocations[i].id, Cartesian, coord, NULL);
			harkio_Positions_append(tf->mics, pos);
		}


		// septfsに保存
		tf->loctfs = NULL;
		tf->septfs = (harkio_Matrix **) malloc(sizeof(harkio_Matrix *) * tf->poses->size);
		for (i = 0 ; i < tf->poses->size ; i++){
			tf->septfs[i] = harkio_Matrix_new(HARKFILEIOBYMAT_STRTYPE_COMPLEX, tf->mics->size, lNFFT/2+1);
		}
		// sDatabase.Data: size_is(numSrc), elemnent_is(Matrix(numMic, numFreq))
		for (k=0 ; k < lNFFT/2+1 ; k++){
			for (i = 0 ; i < tf->mics->size ; i++){
				for (j = 0 ; j < tf->poses->size ; j++){
					BF::Complex cvalue = sDatabase.Data[j].getElementAt(i, k);
					inputvalue =  cvalue.__rep();
					harkio_Matrix_setValueComplex(tf->septfs[j], i, k, inputvalue);
				}
			}
		}

		char *cName = (char*)malloc((strlen(cFilename)+1)*sizeof(char));
		strcpy(cName, cFilename);
		harkio_TransferFunction_write(tf, cName, HARKFILEIOBYMAT_SIGNATURE_SPMAT);
		free(cName);
		
		harkio_TransferFunction_delete(&tf);
	}
	catch(std::exception &ex){
	
		return false ;
	
	}
#else
	int numMic = sDatabase.micLocations.size();
	int numSrc = sDatabase.sourceLocations.size();
	int numFreq = lNFFT/2+1;

	MICARY_SMDatabase smDatabase = {0};
		
	SMDatabaseAlloc(&smDatabase, 0, 0, 0);
	SMDatabaseAlloc(&smDatabase, numFreq, numSrc, numMic);

	for(int i = 0; i < numMic; i++){
	    const BF::MicLocation &ml = sDatabase.micLocations[i];
	    smDatabase.micary_locaMic.ha_idLocation[i] = ml.id;
	    smDatabase.micary_locaMic.micary_Location[i].ha_fX = ml.x;
	    smDatabase.micary_locaMic.micary_Location[i].ha_fY = ml.y;
	    smDatabase.micary_locaMic.micary_Location[i].ha_fZ = ml.z;
	    smDatabase.micary_locaMic.micary_Location[i].ha_idPos = ml.id;
	}

	for(int i = 0; i < numSrc; i++){
	    const BF::SourceLocation &sl = sDatabase.sourceLocations[i];
	    smDatabase.micary_locaSource.ha_idLocation[i] = sl.sourceID;
	    smDatabase.micary_locaSource.micary_Location[i].ha_fX = sl.x;
	    smDatabase.micary_locaSource.micary_Location[i].ha_fY = sl.y;
	    smDatabase.micary_locaSource.micary_Location[i].ha_fZ = sl.z;
	    smDatabase.micary_locaSource.micary_Location[i].ha_idPos = sl.id;
	}

	for(int i = 0 ; i < numFreq; i++){
	    smDatabase.micary_smSeparation.micary_iaFreqLinex.ha_iValue[i] = i;
	}

	int c = 0;
	// sDatabase.Data: size_is(numSrc), elemnent_is(Matrix(numMic, numFreq))
    for(int j = 0 ; j < numSrc; j++){
	    const BF::Matrix &mat = sDatabase.Data[i];
		for(int i = 0 ; i < numFreq; i++){
			for(int k = 0 ; k < numMic; k++){
				// micary_cmpaData->ha_cmpValueは(周波数(マイク(音源)))の順に入力
				c = (i * numMic + j) * numSrc + k;
		    	smDatabase.micary_smSeparation.micary_cmpaSeparation.ha_cmpValue[c].re = mat.getElementAt(j, k).real();
			    smDatabase.micary_smSeparation.micary_cmpaSeparation.ha_cmpValue[c].im = mat.getElementAt(j, k).imag();
			}
	    }
	}

	MICARY_TFDBFileHeader tfdbfh;
	tfdbfh.lFs       = lFs;
	tfdbfh.lNFFT     = lNFFT;
	tfdbfh.lNMic     = numMic;
	tfdbfh.lNSrc     = numSrc;
	tfdbfh.lNFreq    = numFreq;
	tfdbfh.iVersion  = 1;
	tfdbfh.iFileType = FORMAT_TYPE_W;

	CHarkFileIO cHark;
	if(!cHark.WriteData(cFilename.c_str(), &tfdbfh,
			    smDatabase.micary_smSeparation.micary_cmpaSeparation,
			    smDatabase.micary_smSeparation.micary_iaFreqLinex,
			    smDatabase.micary_locaSource,
			    smDatabase.micary_locaMic,
			    "",
			    "",
			    "",
			    "",
			    "",
			    "W_EXPORT")){
		std::cerr << "Failed to WriteData @CHarkFileIOByMatrix::WriteWData" << std::endl;
	    SMDatabaseFree(&smDatabase);
	    return false;
	}

	SMDatabaseFree(&smDatabase);
#endif

	return true;
}
// --------------------------------------------------------------------
#ifdef ENABLE_HARKIO3
CHarkMatFileIOByFMatrix::CHarkMatFileIOByFMatrix()
{
	m_pcMatfilename = NULL;	///< TFファイル名
	m_pmat = NULL;	///< TFファイルポインタ
}
// --------------------------------------------------------------------
CHarkMatFileIOByFMatrix::~CHarkMatFileIOByFMatrix()
{
	CloseMatfile();
}
// --------------------------------------------------------------------
// matファイルのオープン(cFilenameが既に開いている場合はスルーする)
bool CHarkMatFileIOByFMatrix::OpenMatfile(const char* cFilename)
{
	if (m_pcMatfilename){
		if (strcmp(m_pcMatfilename, cFilename) == 0){
			return true;
		}
	}
	
	return (ReopenMatfile(cFilename));
}
// --------------------------------------------------------------------
// matファイルの再オープン
bool CHarkMatFileIOByFMatrix::ReopenMatfile(const char* cFilename)
{
	int i;
	
	if (m_pcMatfilename){
		CloseMatfile();
	}
	
	m_pcMatfilename = (char *)malloc(strlen(cFilename)+1);
	strcpy(m_pcMatfilename, cFilename);
	
	m_pmat = harkio_Matrix_fromFile(m_pcMatfilename);
	if (!m_pmat){
		std::cerr << "Failed to open " << m_pcMatfilename << "@CHarkMatFileIOByFMatrix::ReopenMatfile" << std::endl;
		free(m_pcMatfilename);
		m_pcMatfilename = NULL;
		return false;
	}
	
	return true;
}
// --------------------------------------------------------------------
// matファイルのクローズ
bool CHarkMatFileIOByFMatrix::CloseMatfile()
{
	if (m_pcMatfilename){
		free(m_pcMatfilename);
	}
	if (m_pmat){
		harkio_Matrix_delete(&m_pmat);
	}	

	// 初期化
	m_pcMatfilename = NULL;
	m_pmat = NULL;
}
// --------------------------------------------------------------------
// Matデータを読込み
bool CHarkMatFileIOByFMatrix::ReadMatData(const char* cFilename, 
				BF::FMatrix &sMat)
{
	float value;
	int i,j;
	
	if (!OpenMatfile(cFilename)){
		return false;
	}
	
	if (strcmp(m_pmat->type, HARKFILEIOBYMAT_STRTYPE_FLOAT32) != 0){
		std::cerr << "Failed invalid type @CHarkMatFileIOByFMatrix::ReadMatData, " << m_pmat->type << std::endl;
		return false;
	}
	if (m_pmat->dim != 2){
		std::cerr << "Failed invalid dim @CHarkMatFileIOByFMatrix::ReadMatData, " << m_pmat->dim << std::endl;
		return false;
	}
	
	sMat = BF::FMatrix(m_pmat->rows, m_pmat->cols);
	for (i = 0 ; i < m_pmat->rows ; i++){
		for (j = 0 ; j < m_pmat->cols ; j++){
			harkio_Matrix_getValueFloat32(m_pmat, i, j, &value);
			sMat.setElementAt(i, j, value);
		}
	}
	
	return true;
}
// --------------------------------------------------------------------
// Matデータを書き込み
bool CHarkMatFileIOByFMatrix::WriteMatData(const char* cFilename, 
				BF::FMatrix sMat)
{
	int i, j;
	float value;
	int iRtn;

	harkio_Matrix *mat=harkio_Matrix_new(HARKFILEIOBYMAT_STRTYPE_FLOAT32, sMat.getRowCount(), sMat.getColumnCount());
	if (!mat){
		std::cerr << "Failed to create harkio_Matrix @CHarkMatFileIOByFMatrix::WriteMatData" << std::endl;
		return false;
	}
	
	for (i=0; i<mat->rows; i++){
		for (j=0; j<mat->cols; j++){
			value = sMat.getElementAt(i,j);
			harkio_Matrix_setValueFloat32(mat, i,j,value);
		}
	}
	
	char *cName = (char*)malloc((strlen(cFilename)+1)*sizeof(char));
	strcpy(cName, cFilename);
	iRtn = harkio_Matrix_write(mat, cName);
	if (iRtn == EXIT_FAILURE){
		return false;
	}
	
	harkio_Matrix_delete(&mat);
	
	return true;
}
// --------------------------------------------------------------------
// Matデータを行書き換え
bool CHarkMatFileIOByFMatrix::ReplaceMatData(const char* cFilename, 
				BF::FMatrix sMat,
				int iRowindex)
{
	int i, j;
	float value;
	int iRtn;

	harkio_Matrix *mat=harkio_Matrix_new(HARKFILEIOBYMAT_STRTYPE_FLOAT32, sMat.getRowCount(), sMat.getColumnCount());
	if (!mat){
		std::cerr << "Failed to create harkio_Matrix @CHarkMatFileIOByFMatrix::ReplaceMatData" << std::endl;
		return false;
	}
	
	for (j=0; j<mat->cols; j++){
		value = sMat.getElementAt(iRowindex,j);
		harkio_Matrix_setValueFloat32(mat,iRowindex,j,value);
	}
	
	char *cName = (char*)malloc((strlen(cFilename)+1)*sizeof(char));
	strcpy(cName, cFilename);
	iRtn = harkio_Matrix_replaceC(mat, cName, iRowindex);
	if (iRtn == EXIT_FAILURE){
		harkio_Matrix_delete(&mat);
		return false;
	}
	
	harkio_Matrix_delete(&mat);
	
	return true;
}
// --------------------------------------------------------------------
#endif
