/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <iostream>
#include <fstream>
#include <list>

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <format.h>
#include <Map.h>

#include <string>
#include <sstream>

//#include <boost/tokenizer.hpp>

using namespace std;
using namespace FD;
//using namespace boost;

class LoadMapFrames;

DECLARE_NODE(LoadMapFrames);
/*Node
 *
 * @name LoadMapFrames
 * @category HARK:MISC
 * @description Load map data of frames from files
 *
 * @output_name OUTPUT
 * @output_type any
 * @output_description Data Stream
 *
 * @output_name NOT_EOF
 * @output_type bool
 * @output_description True if the current frame does not reach the End-of-File.
 *
 * @parameter_name FILENAME
 * @parameter_type string
 * @parameter_description The filename to load.
 *
 * @parameter_name INPUT_TYPE
 * @parameter_type string
 * @parameter_list TEXT_float:RAW_float:TEXT_complex_float:RAW_complex_float
 * @parameter_value TEXT_float
 * @parameter_description Input file format type [default: TEXT_float]
 *
 * @parameter_name COL_SIZE
 * @parameter_type int
 * @parameter_value 512
 * @parameter_description COL size of Data
 *
 END*/

class FileIndexTracker {
    public:
    int index;
    string filename;

    public: 
    FileIndexTracker() : index(0) {filename.clear();}

    void set_fname(const std::string fname) { filename.clear();}
    void set_idx(const int idx){ index = idx;}
    const int get_idx(){ return index;}

};


class LoadMapFrames : public BufferedNode {
    #define ENABLE_LOG 0
    int outputID;
    int noteofID;

    typedef enum {TEXT_float, RAW_float, TEXT_complex_float, RAW_complex_float} InputMode;
    typedef enum {MAIN, INTERNAL} FileType;


    InputMode mode;
    string inputType;
    string filename;
    int col;
    ifstream fin;
    ifstream fin_internal;
    bool isObject;
    list<FileIndexTracker> fidx_tracker;

    ofstream file;

    public:
    LoadMapFrames(string nodeName, ParameterSet params) : BufferedNode(nodeName, params), mode(TEXT_float), col(512), isObject(true)
    {
        outputID = addOutput("OUTPUT");
        noteofID = addOutput("NOT_EOF");
        inputType = object_cast<String> (parameters.get("INPUT_TYPE"));
        filename = object_cast<string>(parameters.get("FILENAME"));
        col = dereference_cast<int>(parameters.get("COL_SIZE"));

        if (inputType.compare("RAW_float") == 0) {
            mode = RAW_float;
        }
        else if(inputType.compare("TEXT_complex_float") == 0) {
            mode = TEXT_complex_float;
        }
        else if(inputType.compare("RAW_complex_float") == 0) {
            mode = RAW_complex_float;
        }
        else if(inputType.compare("TEXT_float") == 0) {
            mode = TEXT_float;
        }
        else {
            cout << "Invalid Input Type " << inputType << endl;
        }

        fin.open(filename.c_str());
        if(fin.fail()) {
            throw new NodeException(NULL, string("Can't open '")+filename+string("'."), __FILE__, __LINE__);
        }

        // This file is for debugging.  Will remove lines that does operation for this file once deployed.
        if(ENABLE_LOG) file.open("load.txt");
    }

    ~LoadMapFrames()
    {
        fin.close();
    }

    virtual void initialize()
    {
        outputs[outputID].lookAhead =
        outputs[noteofID].lookAhead = 1 + max(outputs[outputID].lookAhead,outputs[noteofID].lookAhead);

        outputs[outputID].lookBack =
        outputs[noteofID].lookBack =  1 + max(outputs[outputID].lookBack,outputs[noteofID].lookBack);

        this->BufferedNode::initialize();
    }

    int get_file_index(const std::string fname)
    {
        bool isExist = false;
        int index = 0;
        for (std::list<FileIndexTracker>::iterator it=fidx_tracker.begin(); it != fidx_tracker.end(); ++it) {
            if(it->filename == fname) {
                isExist = true;
                //if file already listed, get last index and increment it by 1
                index = it->get_idx() + 1;
                it->set_idx(index);
                break;
            }
        }

        if(isExist == false) {
            // push data
            FileIndexTracker f_idx;
            f_idx.filename = fname;
            f_idx.index = 0;
            index = 0;
            fidx_tracker.push_back(f_idx);
        }
        return index;
    }

    bool transferTextToContainer(vector<string>& tokens, const int file, const int index) {
        string buffer;
        ifstream::pos_type cp;

        switch(file)
        {
            case MAIN:
            {
                int new_frame = 0;
                bool frame_flag = true;
                bool rc = false;
                while(frame_flag) {
                    buffer.clear();
                    getline(fin, buffer);

                    // Get the Frame value of the line retrieved
                    std::size_t pos = buffer.find(" ");
                    std::string frame_str = buffer.substr(0,pos);
                    new_frame = atoi(frame_str.c_str());

                    // Compare the Current Frame value to Count/Iteration value
                    if(new_frame == index) {
                        if(ENABLE_LOG) cout << "Retrieved Frame value str["<< frame_str.c_str() << "] [" << new_frame << "] is equal to Count value [" << index << "]" << endl;
                        frame_flag = false;
                        rc = true;
                    } 
                    // Check values, we might be looking fo a non existing Frame Value
                    // This condition should never be true
                    else if (new_frame > index) {
                        frame_flag = false;
                    } 

                    if(fin.eof()) {
                        isObject = false;
                        frame_flag = false;
                    }
                }
                if(rc == false) { return rc; }
            }
                break;
            case INTERNAL:
            {
                for(int i=0; i<=index; ++i) {
                    buffer.clear();
                    getline(fin_internal, buffer);
                }

               if(fin_internal.eof()) {
                    return false;
                }
            }
            break;
            default:
                return false;
                break;

        }

        //typedef boost::tokenizer<boost::char_separator<char> > tokenizer_space;
        //boost::char_separator<char> sep(" ");
        //tokenizer_space tokenizer(buffer, sep);
        char sep = ' ';
        stringstream bufferStream(buffer);
        string token;
        //for (tokenizer_space::iterator it = tokenizer.begin(); it != tokenizer.end(); ++it) {
            //tokens.push_back(*it);
        while (getline(bufferStream, token, sep)) {
            tokens.push_back(move(token));
        }
        return true;
    }

    void calculate_load_map_data(RCPtr<Map<int, ObjectRef> > &paramout, const std::string fname, const int id,  const int index) {
        //Open Internal files as listed in the main file
        fin_internal.open(fname.c_str(), ios::binary);
        if(fin_internal.fail()) {
            throw new NodeException(NULL, string("Can't open '")+fname+string("'."), __FILE__, __LINE__);
        }

        switch(mode)
        {
            case TEXT_float: {
                vector<string> tokens;
                if (transferTextToContainer(tokens, INTERNAL, index)) {
                    if(ENABLE_LOG) cout << "TEXT_float Actual Size [" << tokens.size() << "] vs Expected Size [" << col << "] @ Index [" << index << "]" << endl;
                    if (tokens.size() != col) {
                        throw new NodeException(this, "TEXT_float : Read data is not equal to expected size.  Check COL value if set correctly.", __FILE__, __LINE__);
                    }

                    RCPtr <Vector<float> > maptmp(new Vector<float>(col));
                    for (unsigned int cnt = 0; cnt < col; cnt++) {
                        char *endptr;
                        float val = strtof(tokens[cnt].c_str(), &endptr);
                        if (endptr == "\0")
                            cout << "The string that was not converted is " << endptr << "." << endl;
                        (*maptmp)[cnt] = val;
                        // For debug
                        if(ENABLE_LOG) file << val << " ";
                    }
                    (*paramout).insert(make_pair(id, maptmp));
                } 
            }
                break;
            case RAW_float:
            {
                for(int i =0; i<=index; ++i) {
                    std::vector<float> buf(col);
                    fin_internal.read((char *)&buf[0], buf.size() * sizeof(float));
                    // Check if COL parameter is set correctly
                    int actual_size = fin_internal.gcount();
                    if(!fin_internal) {
                        if(ENABLE_LOG) cout << "RAW_float Actual Size [" << actual_size << "] vs Expected Size [" << col << "] @ Index [" << index << "]" << endl;
                        isObject = false; } 
                    else {
                        if(i == index) {
                            RCPtr<Vector<float> >maptmp(new Vector<float>(col));
                            for (unsigned int cnt = 0; cnt < col; cnt++) {
                                (*maptmp)[cnt] = buf[cnt];
                                if(ENABLE_LOG) file << buf[cnt] << " ";
                            }
                            (*paramout).insert(make_pair(id, maptmp));
                        }
                        buf.clear();
                    }
                    if((actual_size != (col * sizeof(float))) && actual_size !=0 ){
                        throw new NodeException(this, "RAW_float : Read data is not equal to expected size.  Check COL value if set correctly.", __FILE__, __LINE__);
                    }
                }
            }
                break;
            case TEXT_complex_float:
            {
                vector<string> tokens;
                if(transferTextToContainer(tokens, INTERNAL, index)) {
                    if(ENABLE_LOG) cout << "TEXT_complex_float Actual Size [" << (tokens.size()/2) << "] vs Expected Size [" << col << "] @ Index [" << index << "]" << endl;
                    if (tokens.size() != col * 2) {
                        throw new NodeException(this, "TEXT_complex_float : Read data is not equal to expected size.  Check COL value if set correctly.", __FILE__, __LINE__);
                    }
                    RCPtr <Vector<complex<float> > > maptmp(new Vector<complex<float> >(col));
                    for (unsigned int cnt = 0; cnt < col; cnt++) {
                        char *endptr1, *endptr2;
                        string str_real = tokens[2 * cnt + 0];
                        string str_imag = tokens[2 * cnt + 1];
                        float tmp_r = strtof(str_real.c_str(), &endptr1);
                        if (endptr1 == "\0")
                            cout << "The string that was not converted is " << endptr1 << "." << endl;
                        float tmp_i = strtof(str_imag.c_str(), &endptr2);
                        if (endptr2 == "\0")
                            cout << "The string that was not converted is " << endptr2 << "." << endl;
                        (*maptmp)[cnt] = complex<float>(tmp_r, tmp_i);
                        if(ENABLE_LOG) file << tmp_r << " " << tmp_i << " ";
                    }
                    (*paramout).insert(make_pair(id, maptmp));
                }
            }
                break;
            case RAW_complex_float:
            {
                for(int i =0; i<=index; ++i) {
                    std::vector<float> buf(col * 2);
                    fin_internal.read((char *)(buf.data()), buf.size()*sizeof(float));
                    // Check if COL parameter is set correctly
                    int actual_size = fin_internal.gcount();
                    if(!fin_internal) {
                        if(ENABLE_LOG) cout << "RAW_complex_float Actual Size [" << actual_size << "] vs Expected Size [" << col << "] @ Index [" << index << "]" << endl;
                        isObject=false;
                    } else {
                        if(i == index) {
                            RCPtr<Vector<complex<float> > >maptmp(new Vector<complex<float> >(col));
                            for (unsigned int cnt = 0; cnt < (int)col; cnt++) {
                                float tmp_r = buf[2 * cnt + 0];
                                float tmp_i = buf[2 * cnt + 1];
                                (*maptmp)[cnt] = complex<float>(tmp_r, tmp_i);
                                if(ENABLE_LOG) file << complex<float>(tmp_r, tmp_i) << " ";
                            }
                            (*paramout).insert(make_pair(id, maptmp));
                        }
                        buf.clear();
                    }
                    if((actual_size != (col * sizeof(float) * 2)) && (actual_size != 0)){
                        throw new NodeException(this, "RAW_complex_float : Read data is not equal to expected size.  Check COL value if set correctly.", __FILE__, __LINE__);
                    }
                }
            }
                break;
            default:
                break;
        }
        if(fin_internal.is_open())
            fin_internal.close();
    }

    void calculate(int output_id, int count, Buffer &out) {
        RCPtr<Map<int, ObjectRef> > paramout(new Map<int, ObjectRef>);
        (*(outputs[outputID].buffer))[count] = paramout;

        // Get all filenames for frame[count]
        vector<string> fnames;
        if (transferTextToContainer(fnames, MAIN, count)) {
            int size = fnames.size();
            // Start with index 1 since index 0 is for the frame value
            for (unsigned int i = 1; i < size; i++) {
            string fname(fnames[i]);

            // Get index value which frame of the file will be processed
            int index = get_file_index(fname);

            // Get the Map Key as indicated in the filename.  Default is 4-digit Map Id.
            std::size_t pos = fname.find("_Vector_");
            int digits =  4;
            pos = pos - digits;
            // This is extra checking if in case map id is more than 4 digits
            while(fname.substr(pos-1, 1) != "_") {
                pos--;
                digits++;
            }
            string id_str = fname.substr(pos,digits);
            int map_id = atoi(id_str.c_str());
            if(ENABLE_LOG) cout << fname << " with id_str " << id_str  << " and id is " << map_id << " Pos is @ " <<  pos << " with digits " << digits << endl;

            // Get Load Map data
            calculate_load_map_data(paramout, fname, map_id, index);
            fin_internal.close();
            }
        } 

        (*(outputs[noteofID].buffer))[count] = ((fin.eof() || (isObject == false)) ? FalseObject : TrueObject);
        if(ENABLE_LOG) { 
            if(isObject == false) { cout << "This is the end Frame @ Count [" << count  << "]"<< endl; }
            else { file << endl; }
        }
    }
};

