/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <iostream>
#include <fstream>

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>

#include "format.h"
#include <string>
#include <sstream>

//#include <boost/tokenizer.hpp>

using namespace std;
using namespace FD;
//using namespace boost;

class LoadVectorFrames;

DECLARE_NODE(LoadVectorFrames);
/*Node
 *
 * @name LoadVectorFrames
 * @category HARK:MISC
 * @description Load vector data of frames from file
 *
 * @output_name OUTPUT
 * @output_type any
 * @output_description Data Stream
 *
 * @output_name NOT_EOF
 * @output_type bool
 * @output_description True if the current frame does not reach the End-of-File.
 *
 * @parameter_name FILENAME
 * @parameter_type string
 * @parameter_description The filename to load.
 *
 * @parameter_name INPUT_TYPE
 * @parameter_type string
 * @parameter_list TEXT_float:RAW_float:TEXT_complex_float:RAW_complex_float
 * @parameter_value TEXT_float
 * @parameter_description Input file format type [default: TEXT_float]
 *
 * @parameter_name COL_SIZE
 * @parameter_type int
 * @parameter_value 512
 * @parameter_description COL size of Vector
 *
 END*/

class LoadVectorFrames : public BufferedNode {
    #define ENABLE_LOG 0
    int outputID;
    int noteofID;

    typedef enum {TEXT_float, RAW_float, TEXT_complex_float, RAW_complex_float} InputMode;

    InputMode mode;
    string inputType;
    string filename;
    int col;
    int length;
    int line;
    ifstream fin;
    bool isObject;

    ofstream file;

    public:
    LoadVectorFrames(string nodeName, ParameterSet params) : BufferedNode(nodeName, params), mode(TEXT_float), col(512), length(0), line(0), isObject(true)
    {
        outputID = addOutput("OUTPUT");
        noteofID = addOutput("NOT_EOF");
        inputType = object_cast<String> (parameters.get("INPUT_TYPE"));
        filename = object_cast<string>(parameters.get("FILENAME"));
        col = dereference_cast<int>(parameters.get("COL_SIZE"));

        fin.open(filename.c_str(), ios::binary);
        if(fin.fail())
        {
            throw new NodeException(NULL, string("Can't open '")+filename+string("'."), __FILE__, __LINE__);
        }

        fin.seekg (0, fin.end);
        int length = fin.tellg();
        fin.seekg (0, fin.beg);
        line = (length /col) / sizeof(float);

        if (inputType.compare("RAW_float") == 0) {
            mode = RAW_float;
        }
        else if(inputType.compare("TEXT_complex_float") == 0) {
            mode = TEXT_complex_float;
            line = line/2;
        }
        else if(inputType.compare("RAW_complex_float") == 0) {
            mode = RAW_complex_float;
            line = line/2;
        }
        else {
            mode = TEXT_float;
        }

        // This file is for debugging.  Will remove lines that does operation for this file once deployed.
        if(ENABLE_LOG) file.open("load.txt");
    }

    ~LoadVectorFrames()
    {
        fin.close();
    }

    virtual void initialize()
    {
        outputs[outputID].lookAhead =
        outputs[noteofID].lookAhead = 1 + max(outputs[outputID].lookAhead,outputs[noteofID].lookAhead);

        outputs[outputID].lookBack =
        outputs[noteofID].lookBack =  1 + max(outputs[outputID].lookBack,outputs[noteofID].lookBack);

        this->BufferedNode::initialize();
    }

    bool transferTextToContainer(vector<string>& tokens, int count) {
        string buffer;
        ifstream::pos_type cp;
        cp = fin.tellg();
        getline(fin, buffer);

        if(buffer.empty())
            return false;

        //typedef boost::tokenizer<boost::char_separator<char> > tokenizer_space;
        //boost::char_separator<char> sep(" ");
        //tokenizer_space tokenizer(buffer, sep);
        char sep = ' ';
        stringstream bufferStream(buffer);
        string token;
        //for (tokenizer_space::iterator it = tokenizer.begin(); it != tokenizer.end(); ++it) {
        //    tokens.push_back(*it);
        while (getline(bufferStream, token, sep)) {
            tokens.push_back(move(token));
        }
        if (fin.eof()) {
            isObject = false;
        }
        return true;
    }

    void calculate(int output_id, int count, Buffer &out) {
        (*(outputs[noteofID].buffer))[count] = ((fin.eof() || (isObject == false)) ? FalseObject : TrueObject);
        
        switch(mode)
        {
            case TEXT_float:
            {
                RCPtr <Vector<float> > paramout(new Vector<float>(col));
                (*(outputs[outputID].buffer))[count] = paramout;

                vector<string> tokens;
                if(transferTextToContainer(tokens, count)) {
                    if(ENABLE_LOG) cout << "TEXT_float Actual Size [" << tokens.size() << "] vs Expected Size [" << col << "] @ Count [" << count << "]" << endl;
                    if (tokens.size() != col) {
                        throw new NodeException(this, "Read data is not equal to expected size.  Check COL value if set correctly.", __FILE__, __LINE__);
                    }
                    for (unsigned int cnt = 0; cnt < col; cnt++) {
                        char *endptr;
                        float val = strtof(tokens[cnt].c_str(), &endptr);
                        if (endptr == "\0")
                            cout << "The string that was not converted is " << endptr << "." << endl;
                        (*paramout)[cnt] = val;
                        if(ENABLE_LOG) file << (*paramout)[cnt] << " ";
                    }
                    if(ENABLE_LOG) file << endl;
                }
            }
                break;
            case RAW_float:
            {
                RCPtr <Vector<float> >paramout(new Vector<float>(col));
                (*(outputs[outputID].buffer))[count] = paramout;

                std::vector<float> buf(col);
                fin.read((char *)&buf[0], buf.size() * sizeof(float));
                int actual_size = fin.gcount();
                if(ENABLE_LOG) cout << "RAW_float : Actual Size [" << actual_size << "] vs Expected Size [" << col*sizeof(float) << "] @ Count [" << count << "]" << endl;

                if (fin.eof() || !(actual_size)) {
                    isObject = false;
                } else {
                    if(actual_size != col * sizeof(float)) {
                        throw new NodeException(this, "Read data is not equal to expected size.  Check COL value if set correctly.", __FILE__, __LINE__);
                    }
                    for (unsigned int cnt = 0; cnt < col; cnt++) {
                        (*paramout)[cnt] = buf[cnt];
                        if(ENABLE_LOG) file << buf[cnt] << " ";
                    }
                    if(ENABLE_LOG) file << endl;
                }
            }
                break;
            case TEXT_complex_float:
            {
                RCPtr <Vector<complex<float> > > paramout(new Vector<complex<float> >(col));
                (*(outputs[outputID].buffer))[count] = paramout;

                vector<string> tokens;
                if(transferTextToContainer(tokens, count)) {
                    if(ENABLE_LOG) cout << "TEXT_complex_float : Actual Size [" << tokens.size() << "] vs Expected Size [" << (col * 2) << "] @ Count [" << count << "]" << endl;
                    if ((tokens.size() / 2) != col) {
                        throw new NodeException(this, "TEXT_complex_float : Read data is not equal to expected size.  Check COL value if set correctly.", __FILE__, __LINE__);
                    }

                    for (unsigned int cnt = 0; cnt < (int) col; cnt++) {
                        char *endptr1, *endptr2;
                        string str_real = tokens[2 * cnt + 0];
                        string str_imag = tokens[2 * cnt + 1];
                        float tmp_r = strtof(str_real.c_str(), &endptr1);
                        if (endptr1 == "\0")
                            cout << "The string that was not converted is " << endptr1 << "." << endl;
                        float tmp_i = strtof(str_imag.c_str(), &endptr2);
                        if (endptr2 == "\0")
                            cout << "The string that was not converted is " << endptr2 << "." << endl;
                        (*paramout)[cnt] = complex<float>(tmp_r, tmp_i);
                        if(ENABLE_LOG) file <<  complex<float>(tmp_r, tmp_i) << " ";
                    }
                    if(ENABLE_LOG) file << endl;
                }
            }
                break;
            case RAW_complex_float:
            {
                RCPtr <Vector<complex<float> > > paramout(new Vector<complex<float> >(col));
                (*(outputs[outputID].buffer))[count] = paramout;

                std::vector<float> buf(col * 2);
                fin.read((char *)&buf[0], buf.size() * sizeof(float));
                int actual_size = fin.gcount();
                if(ENABLE_LOG) cout << "RAW_complex_float : Actual Size [" << actual_size << "] vs Expected Size [" << (col*sizeof(float) *2) << "] @ Count [" << count << "]" << endl;

                if (fin.eof() || !(actual_size)) {
                    isObject = false;
                } else {
                    if(actual_size != (col * 2 * sizeof(float))) {
                        throw new NodeException(this, "Read data is not equal to expected size.  Check COL value if set correctly.", __FILE__, __LINE__);
                    }
                    for (unsigned int cnt = 0; cnt < (int) col; cnt++) {
                        float tmp_r = buf[2 * cnt + 0];
                        float tmp_i = buf[2 * cnt + 1];
                        (*paramout)[cnt] = complex<float>(tmp_r, tmp_i);
                        if(ENABLE_LOG) file <<  complex<float>(tmp_r, tmp_i) << " ";
                    }
                    if(ENABLE_LOG) file << endl;
                }
            }
                break;
            default:
                break;
        }

        if(fin.eof() || isObject == false) { 
            (*(outputs[noteofID].buffer))[count] = FalseObject; }
    }
};

