/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <string.h>

#ifdef ENABLE_HARKIO3
  #include "HarkFileIO_NH.h"
#elif  ENABLE_HARKIO2
  #include "HarkFileIO.h"
  using namespace harkio2;
#elif ENABLE_HARKIO1
  #include "HarkFileIO.h"
  using namespace harkio;
#endif

#include "HarkFileIOByMatrix.h"

#include "LocalizationBFDS.hpp"

LocalizationBFDS::LocalizationBFDS(int sampling_rate, int fft_length, float sound_speed) :
  SamplingRate(sampling_rate), numFFT(fft_length), SoundSpeed(sound_speed)
{
	this->numFreq = this->numFFT / 2 + 1;
}

bool LocalizationBFDS::ReadTransferFunction(const string &file_name, vector<int> &ch_select, int &num_mic, int &num_src)
{
	num_mic = 0;
	num_src = 0;

	CHarkFileIOByMatrix	cHarkFile;

	// read header
	string fileType = string("transfer function");
	long int hFs, hNFFT, hNSrc, hNMic;
	if (!cHarkFile.ReadHeader(file_name.c_str(), fileType.c_str(), &hFs, &hNFFT, &hNSrc, &hNMic)) {
		cerr << "LocalizationBFDS: Failed to get header from the file." << endl;
		return false;
	}
	// check sampling rate
	if (hFs != this->SamplingRate) {
		cerr << "LocalizationBFDS: Sampling rate is different between parameter and TF header." << endl;
		return false;
	}

	BF::TransformFunctionDatabase TransFuncDatabase;

	// read TF
	if (!cHarkFile.ReadTFData(file_name.c_str(), -1.0f, TransFuncDatabase)) {
		cerr << "LocalizationBFDS: Failed to get transfer function." << endl;
		return false;
	}

	// TF: size is nfreq, element is Matrix(nmic, nsrc)

	int nfreq = TransFuncDatabase.Data.size();
	
	if (nfreq != this->numFreq) {
		cerr << "LocalizationBFDS: Number of freq bin is different between parameter and TF." << endl;
		return false;
	}

	int nmic = TransFuncDatabase.Data[0].getRowCount();
	this->numSrc = TransFuncDatabase.Data[0].getColumnCount();

	// check size of TF
	if (nmic == 0 || this->numSrc == 0) {
		cerr << "LocalizationBFDS: Size of transfer function is invalid." << endl;
		return false;
	}

	// if channel is not set, set default
	if (ch_select.size() == 0) {
		ch_select.resize(nmic);
		for (int c = 0; c < nmic; c++) {
			ch_select[c] = c;
		}
	}
	// channel check
	for (int c = 0; c < ch_select.size(); c++) {
		if (!(ch_select[c] >= 0 && ch_select[c] < nmic)) {
			cerr << "LocalizationBFDS: Nonexistent channel is selected." << endl;
			return false;
		}
	}
	this->numMic = ch_select.size();

	// TF: size(numFreq), element(numMic,numSrc) -> size(numSrc), element(numMic,numFreq)
	BF::Complex cval;
	this->Adata.resize(this->numSrc, BF::Matrix(this->numMic, this->numFreq));
	for (int i = 0; i < this->numSrc; i++) {
		for (int j = 0; j < this->numMic; j++) {
//cout << i << " - " << j << endl;
			for (int k = 0; k < this->numFreq; k++) {
				cval = TransFuncDatabase.Data[k].getElementAt(ch_select[j], i);
				Adata[i].setElementAt(j, k, cval);
//cout << cval << ", ";
			}
//cout << endl;
		}
	}

	// source
	TransFuncDatabase.sourceLocations.swap(this->SrcPositions);

//for (int i = 0; i < this->numSrc; i++) {
//	cout << "src[" << i << "] " << this->SrcPositions[i].id << " (" << this->SrcPositions[i].x << ", " << this->SrcPositions[i].y << ", " << this->SrcPositions[i].z << ")" << endl;
//}

	num_mic = this->numMic;
	num_src = this->numSrc;

	return true;
}

bool LocalizationBFDS::ReadXMLmicPos(const char *file_name, vector<int> &ch_select, int &num_mic)
{
	num_mic = 0;

	char *mic_file_name = (char *)malloc((strlen(file_name) + 1) * sizeof(char));
	strcpy(mic_file_name, file_name);
	harkio_XML* xml = harkio_XML_newFromXML(mic_file_name, NULL);
	free(mic_file_name);
	if (!xml) {
		cerr << "LocalizationBFDS: Failed to read microphone position file." << endl;
		return false;
	}

	int nmic = xml->poses->size;
	if (nmic == 0) {
		cerr << "LocalizationBFDS: Number of microphone positions is invalid." << endl;
		harkio_XML_delete(&xml);
		return false;
	}

	// if channel is not set, set default
	if (ch_select.size() == 0) {
		ch_select.resize(nmic);
		for (int c = 0; c < nmic; c++) {
			ch_select[c] = c;
		}
	}
	// channel check
	for (int c = 0; c < ch_select.size(); c++) {
		if (!(ch_select[c] >= 0 && ch_select[c] < nmic)) {
			cerr << "LocalizationBFDS: Nonexistent channel is selected." << endl;
			harkio_XML_delete(&xml);
			return false;
		}
	}
	this->numMic = ch_select.size();

	// get positions

	float fcoord[3];
	float *pfc;
		
	vector<float> fx, fy, fz;
	fx.resize(nmic); fy.resize(nmic); fz.resize(nmic);
	vector<int> id; id.resize(nmic);
	for (int j = 0 ; j < nmic; j++) {
		if (xml->poses->pos[j]->sys == Cartesian) {
			pfc = xml->poses->pos[j]->coord;
		} else { //if (xml->poses->pos[j]->sys == Polar) {
			harkio_Position_Convert(xml->poses->pos[j], fcoord, Cartesian);
			pfc = fcoord;
		}

		fx[j] = pfc[0];
		fy[j] = pfc[1];
		fz[j] = pfc[2];
		id[j] = xml->poses->pos[j]->id;

//cout << "mic[" << j << "]: " << id[j] << " (" << fx[j] << "," << fy[j] << "," << fz[j] << ")" << endl;
	}

	harkio_XML_delete(&xml);

	// get adjust positions

	vector<float>::iterator xmax = max_element(fx.begin(), fx.end());
	vector<float>::iterator xmin = min_element(fx.begin(), fx.end());
	this->MicXC = (*xmax + *xmin) / 2.0f;
	vector<float>::iterator ymax = max_element(fy.begin(), fy.end());
	vector<float>::iterator ymin = min_element(fy.begin(), fy.end());
	this->MicYC = (*ymax + *ymin) / 2.0f;
	vector<float>::iterator zmax = max_element(fz.begin(), fz.end());
	vector<float>::iterator zmin = min_element(fz.begin(), fz.end());
	this->MicZC = (*zmax + *zmin) / 2.0f;

//cout << "center: (" << MicXC << "," << MicYC << "," << MicZC << ")" << endl;

	this->MicPositions.resize(this->numMic, BF::MicLocation(0.0f, 0.0f, 0.0f, 0));
	for (int j = 0; j < this->numMic; j++) {
		int c = ch_select[j];

		this->MicPositions[j] = BF::MicLocation(fx[c], fy[c], fz[c], id[c]);

//cout << "mic[" << j << "]: id[c] (" << fx[c] << "," << fy[c] << "," << fz[c] << ")" << endl;
	}

	num_mic = this->numMic;
	
	return true;
}

bool LocalizationBFDS::ReadXMLsrcPos(const char *file_name, int &num_src)
{
	num_src = 0;

	char *src_file_name = (char *)malloc((strlen(file_name) + 1) * sizeof(char));
	strcpy(src_file_name, file_name);
	harkio_XML* xml = harkio_XML_newFromXML(src_file_name, NULL);
	free(src_file_name);
	if (!xml) {
		cerr << "LocalizationBFDS: Failed to read source position file." << endl;
		return false;
	}

	// check sampling rate
/*
	if (xml->cfg->samplingRate != this->SamplingRate) {
		cerr << "LocalizationBFDS: Sampling rate is different between parameter and xml." << endl;
		harkio_XML_delete(&xml);
		return false;
	}
*/
	this->numSrc = xml->poses->size;
	if (this->numSrc == 0) {
		cerr << "LocalizationBFDS: Number of source positions is invalid." << endl;
		harkio_XML_delete(&xml);
		return false;
	}

	// get positions

	float fcoord[3];
	float *pfc;

	this->SrcPositions.resize(this->numSrc, BF::SourceLocation(0.0f, 0.0f, 0.0f, 0, 0, true));

	for (int i = 0 ; i < this->numSrc; i++) {
		if (xml->poses->pos[i]->sys == Cartesian) {
			pfc = xml->poses->pos[i]->coord;
		} else { //if (xml->poses->pos[i]->sys == Polar) {
			harkio_Position_Convert(xml->poses->pos[i], fcoord, Cartesian);
			pfc = fcoord;
		}

		this->SrcPositions[i] = BF::SourceLocation(pfc[0], pfc[1], pfc[2], -1, xml->poses->pos[i]->id, true);

//cout << "src[" << i << "]: " << xml->poses->pos[i]->id << " (" << pfc[0] << ", " << pfc[1] << ", " << pfc[2] << ")" << endl;
	}

	harkio_XML_delete(&xml);

	num_src = this->numSrc;

	return true;
}

void LocalizationBFDS::SetBfDsFilterfromTF(vector<BF::Matrix> &bf_filter)
{
	vector<BF::Matrix> newF;
	newF.resize(this->numSrc, BF::Matrix(this->numMic, this->numFreq));

	for (int i = 0; i < this->numSrc; i++) {
		BF::Matrix &filter = newF[i];

		BF::Matrix Amat(Adata[i]);

		filter = Amat.conjugated().elementwiseDivided(Amat.elementwiseAbs());
	}

	newF.swap(bf_filter);
}

void LocalizationBFDS::SetBfDsFilterfromPOS(vector<BF::Matrix> &bf_filter)
{
	// 1サンプル遅延の位相スペクトル
	float pi = acos(-1.0);
	vector<float> p1v;
	p1v.resize(this->numFreq);
	for (int k = 0; k < this->numFreq; k++) {
		p1v[k] = (float)k / (float)this->numFFT * 2.0f * pi;
	}

	vector<BF::Matrix> newF;
	newF.resize(this->numSrc, BF::Matrix(this->numMic, this->numFreq));

	float dly;
	BF::Complex dsVal;
	for (int i = 0; i < this->numSrc; i++) {
		BF::Matrix &filter = newF[i];

		for (int j = 0; j < this->numMic; j++) {
			// 遅延時間の計算，時間原点は空間原点（マイク中心）にあわせる
			dly = -((this->SrcPositions[i].x - this->MicXC) * (this->MicPositions[j].x - this->MicXC) + 
			        (this->SrcPositions[i].y - this->MicYC) * (this->MicPositions[j].y - this->MicYC) + 
			        (this->SrcPositions[i].z - this->MicZC) * (this->MicPositions[j].z - this->MicZC)) 
			        / this->SoundSpeed * (float)this->SamplingRate;

			for (int k = 0; k < this->numFreq; k++) {
				// 位相補正用の位相スペクトル行列を複素スペクトルに
				dsVal = exp(BF::Complex(0.0f, dly * p1v[k]));

				filter.setElementAt(j, k, dsVal);
			}
		}
	}

	newF.swap(bf_filter);
}

void LocalizationBFDS::Localize(BF::Matrix &x, vector<BF::Matrix> &bf_filter, vector<float> &bf_power, int &found_id)
{
	// 入力スペクトル
	BF::Matrix xx(x);	

	BF::Matrix bx(this->numMic, this->numFreq);
	BF::Matrix bfspec(this->numSrc, this->numFreq);
	for (int i = 0; i < this->numSrc; i++) {
		const BF::Matrix &bf = bf_filter[i];
    
		// filtering
    	bx = bf.elementwiseMultiplied(xx);

		// channel summention power
		for (int k = 0; k < this->numFreq; k++) {
			BF::Complex bxs = bx.getColumn(k).getElementSum();
			bfspec.setElementAt(i, k, bxs.real() * bxs.real() + bxs.imag() * bxs.imag());
		}
	}

	// freq. summention
//	bf_power.resize(this->numSrc);//ここに入る前に確保されている
	for (int i = 0; i < this->numSrc; i++) {
		bf_power[i] = bfspec.getRow(i).getElementSum().real();
	}

	// search max
	vector<float>::iterator it = max_element(bf_power.begin(), bf_power.end());
	found_id = distance(bf_power.begin(), it);
}

