/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <iostream>
#include <functional>

#include "LocalizationCSP.hpp"

struct less_for_vec {
	bool operator()(vector<float>& v, int i, int j) { return v[i] < v[j]; }
};

LocalizationCSP::LocalizationCSP(int tlength, vector<float> mssw, int min_tau, int max_tau, 
                                 float csp_threshold, int maxn_out_peaks) :
  TimeLength(tlength), 
  MinTau(min_tau), MaxTau(max_tau), CspThreshold(csp_threshold), MaxNumOutputPeaks(maxn_out_peaks)
{
	this->FreqLength = this->TimeLength / 2 + 1;

	this->Mssw.resize(this->FreqLength);
	for (unsigned int i = 0; i < this->FreqLength; i++) {
		this->Mssw[i] = mssw[i];
	}

	if (this->MaxNumOutputPeaks < 0) { this->MaxNumOutputPeaks = 0; }

	this->rfft1.initialize(this->TimeLength);

//  this->win_conj.resize(this->TimeLength);
//  ConjWindow(this->win_conj);
}

void LocalizationCSP::AddCrossCorrelation(vector<BF::Complex> &rx, BF::Matrix &X)
{
	BF::Complex x_0, x_1;
	double abs_2;
	for (int k = 0; k < this->FreqLength; k++) {
		if (this->Mssw[k] == 0.0f) continue;

		x_0 = X.getElementAt(0, k);
		x_1 = X.getElementAt(1, k);

//		abs_2 = 1.0f / fabs(x_0) / fabs(x_1);
//		rx[k] = rx[k] + x_0 * conj(x_1) * abs_2;
		if (fabs(x_0) == 0.0f || fabs(x_1) == 0.0f) {
			abs_2 = 0.0;
		} else {
			abs_2 = 1.0f / fabs(x_0) / fabs(x_1);
		}
		rx[k] = rx[k] + x_0 * conj(x_1) * abs_2;
	}
}

void LocalizationCSP::NormalizeCrossCorrelation(vector<BF::Complex> &rx, int sum_count)
{
	double sum_count_1 = 1.0f / sum_count;
	for (int k = 0; k < this->FreqLength; k++) {
		if (this->Mssw[k] == 0.0f) continue;

    	rx[k] = rx[k] * sum_count_1;
	}
}

void LocalizationCSP::ClearCrossCorrelation(vector<BF::Complex> &rx)
{
	for (int k = 0; k < rx.size(); k++) {
		rx[k] = BF::Complex(0.0f, 0.0f);
	}
}

int LocalizationCSP::Localize(vector<BF::Complex> &rx, vector<float> &csp, vector<int>& found_id, vector<float> &found_csp)
{
	for (int k = 0; k < FreqLength; k++) {
		this->rfft1.A[k] = rx[k];
	}

	this->rfft1.synthesize();

	csp.resize(this->rfft1.tlength());
	for (int k = 0; k < rfft1.tlength(); k++) {
		csp[k] = this->rfft1.a[k] * (1.0f / this->rfft1.tlength());
	}

	return this->SearchLocation(csp, found_id, found_csp);
}

int LocalizationCSP::SearchLocation(vector<float> &csp, vector<int> &found_id, vector<float> &found_csp)
{
	vector<int> candidate_id;
	vector<float> candidate_val;

	int nh = csp.size();
	candidate_id.reserve(nh);
	candidate_id.resize(0);
	candidate_val.reserve(nh);
	candidate_val.resize(0);

	this->SearchPeak(csp, candidate_id, candidate_val);

	vector<int> index;
	for (int i = 0; i < candidate_id.size(); i++) {
		index.push_back(i);
	}
	sort(index.begin(), index.end(), 
	     std::bind<bool>(less_for_vec(), candidate_val, std::placeholders::_2, std::placeholders::_1));

	vector<int> res_id;
	int lag;
	for (int i = 0; i < index.size(); i++) {
		if (candidate_id[index[i]] <= this->FreqLength) {
			lag = candidate_id[index[i]];
		} else {
			lag = candidate_id[index[i]] - this->TimeLength;
		}
		res_id.push_back(lag);
	}
	vector<float> res_val;
	for (int i = 0; i < index.size(); i++) {
		res_val.push_back(candidate_val[index[i]]);
	}

	int n_result = 0;
	if (this->MaxNumOutputPeaks == 0) {
		n_result = res_id.size();
	} else if (res_id.size() < this->MaxNumOutputPeaks) {
		n_result = res_id.size();
	} else {
		n_result = this->MaxNumOutputPeaks;
	}

	found_id.resize(0);
	found_csp.resize(0);
	for (int i = 0; i < n_result; i++) {
		found_id.push_back(res_id[i]);
		found_csp.push_back(res_val[i]);
	}

	return n_result;
}

void LocalizationCSP::SearchPeak(vector<float> &csp, vector<int> &peak_cand_id, vector<float> &peak_cand_val)
{
	float c1;
	float c2;
	float c3;
	int nh = csp.size();
	int lag;
	for (int i = 0; i < nh; i++) {
		c1 = (i == 0 ? csp.back() : csp[i - 1]);
		c2 = csp[i];
		c3 = (i == csp.size() - 1 ? csp.front() : csp[i + 1]);

		if (i <= FreqLength) {
			lag = i;
		} else {
			lag = i - TimeLength;
		}

		if (c2 - c1 > 0 && c3 - c2 < 0 && c2 >= CspThreshold && lag >= MinTau && lag <= MaxTau) {
			peak_cand_id.push_back(i);
			peak_cand_val.push_back(c2);
		}
	}
}

