/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#define ENABLE_TIMER 0

#include <chrono>
#include "strDurlog.hpp"

#ifdef ENABLE_HARKIO_NH

#include "LocalizationMUSIC_NH.hpp"

#include <iostream>
#include <fstream>
#include <math.h>

#ifdef ENABLE_HARKIO3
extern "C" {
#include <libharkio3.h>
}
#endif

using namespace std;
#ifdef ENABLE_LAPACK
#include "HARK_Matrix.h"
#include <functional>

#else
// Define NDEBUG to guarantee accuracy in boost
#ifndef NDEBUG
#define NDEBUG
#endif

#include <boost/bind.hpp>
#include "evd.hpp"
#include "svd.hpp"
// #include "svdold.hpp" 
using namespace boost;
using namespace boost::numeric;

#endif
using namespace std::chrono;


struct less_for_vec {
  bool operator()(vector<float>& v, int i, int j) {
    return v[i] < v[j];
  }
};

LocalizationMUSIC::LocalizationMUSIC(int num_mic, int sampling_rate, int tlength, int shift, int maxn_out_peaks) :
    NumMicrophones(num_mic), TimeLength(tlength), Shift(shift), SamplingRate(sampling_rate), R(num_mic, num_mic), V(
        num_mic, num_mic), RN(num_mic, num_mic), VN(num_mic, num_mic), TempCalc1(num_mic, num_mic), TempCalc2(num_mic,
        num_mic), TempCalc3(num_mic, num_mic), TempCalc4(num_mic, num_mic), TempCalc5(num_mic, num_mic), TempCalc6(
        num_mic, num_mic), TempCalc7(num_mic, num_mic), TempCalc8(num_mic, num_mic), TempCalc9(num_mic, num_mic), NumSource(
        num_mic) {
  FreqLength = TimeLength / 2 + 1;

  UsesFreq.resize(FreqLength);

  maxnum_output_peaks = maxn_out_peaks;

  EigenValue.resize(FreqLength);
  CEigenValue.resize(NumMicrophones);
  REigenValue.resize(NumMicrophones);
  MaxRxx.resize(FreqLength);
  MaxRxxN.resize(FreqLength);
  E.resize(FreqLength);
  EN.resize(FreqLength);

#ifdef ENABLE_LAPACK
  for (int k = 0; k < FreqLength; k++) {
    EigenValue[k].resize(NumMicrophones);
	fcmatrix Emat(NumMicrophones, NumMicrophones);
	fcmatrix ENmat(NumMicrophones, NumMicrophones);

    E[k] = Emat;
	EN[k] = ENmat;
  }
#else
  for (int k = 0; k < FreqLength; k++) {
    EigenValue[k].resize(NumMicrophones);
    E[k].resize(NumMicrophones, NumMicrophones);
    EN[k].resize(NumMicrophones, NumMicrophones);
  }
#endif

  FoundId.resize(NumSource);

}

void LocalizationMUSIC::SetUsedFrequency(int lo, int up) {
  int lower = (int) ((float) lo * TimeLength / SamplingRate);
  int upper = (int) ((float) up * TimeLength / SamplingRate);
  for (int k = lower; k < upper; k++) {
    UsesFreq[k] = true;
  }
}

void LocalizationMUSIC::Localize(vector<fcmatrix>& Rxx, vector<fcmatrix>& RxxN, int mindeg, int maxdeg,
    string music_alg, fvector& MSSW, bool enable_eigenvalue_weight) {
#if ENABLE_TIMER
  system_clock::time_point start_time;
  system_clock::time_point end_time;
  start_time = std::chrono::system_clock::now();
#endif

  for (int k = 0; k < UsesFreq.size(); k++) {

    MaxRxx[k] = 1.0;
    MaxRxxN[k] = 1.0;

    if (!UsesFreq[k]) {
      continue;
    }
    float max = MaxOfAbsValue(Rxx[k]);
#ifdef ENABLE_LAPACK
    R = Rxx[k] / (max + 1.0f);
#else
    noalias(R) = Rxx[k] / (max + 1.0f);
#endif
    MaxRxx[k] = max + 1.0f;

    if (music_alg != "SEVD") {
      float maxN = MaxOfAbsValue(RxxN[k]);
#ifdef ENABLE_LAPACK
      RN = RxxN[k] / (maxN);
#else
      noalias(RN) = RxxN[k] / (maxN);
#endif
      MaxRxxN[k] = maxN;
    }
    //============= GEVD ==============
    //printf("%d ", k);

    if (music_alg == "SEVD") {
#ifdef ENABLE_LAPACK
      fcmatrix EV(NumMicrophones,1);
      R.eigenValues(EV,V);
      for (int ii = 0; ii < NumMicrophones; ii++){
          EigenValue[k][ii] = EV.getElementAt(ii,0).real();
      }
      sort_eig_fvec_ascending(EigenValue[k], V, EigenValue[k], V);
      //sort_eig_fvec_descending(EigenValue[k], V, EigenValue[k], V);
#else
      evd(R, EigenValue[k], V);
#endif
	

    } else if (music_alg == "GEVD") {
#ifdef ENABLE_LAPACK
      //RN.dump(stdout, "RN: %p"); ////////
      TempCalc1 = RN.Inverse();
      //TempCalc1.dump(stdout, "Tempcalc1: %p"); ////////
      fcmatrix EV(NumMicrophones,1);
      //EV.dump(stdout, "EV: %p"); ////////
      TempCalc1.eigenValues(EV,VN);
      //TempCalc1.dump(stdout, "Tempcalc1(2): %p"); ////////
      //EV.dump(stdout, "EV(2): %p"); ////////
      //VN.dump(stdout, "VN: %p"); ////////
      for (int ii = 0; ii < NumMicrophones; ii++){
          REigenValue[ii] = EV.getElementAt(ii,0).real();
      }
      sort_eig_fvec_ascending(REigenValue, VN, REigenValue, VN);//
      //sort_eig_fvec_descending(REigenValue, VN, REigenValue, VN);//
	  //evd.hでは昇順で並べ替えられている
      //VN.dump(stdout, "VN(2): %p"); ////////
#else
      invMAT(RN, TempCalc1);
      evd(TempCalc1, REigenValue, VN);
#endif
      for (int i = 0; i < REigenValue.size(); i++) {
        if (REigenValue[i] < 0)
          REigenValue[i] = -1.0f * pow(-1.0f * REigenValue[i], (float) 0.5);
        else
          REigenValue[i] = pow(REigenValue[i], (float) 0.5);
      }

#ifdef ENABLE_LAPACK
      TempCalc2 = VN.Inverse();
      //TempCalc2.dump(stdout, "TempCalc2: %p"); ////////
      for (int i = 0; i < TempCalc3.getRowCount(); i++) {
        for (int j = 0; j < TempCalc3.getColumnCount(); j++) {
          if (i == j)
            TempCalc3.setElementAt(i,j,REigenValue[i]);
          else
            TempCalc3.setElementAt(i,j,0.0);
        }
      }
      //TempCalc3.dump(stdout, "TempCalc3: %p"); ////////
      TempCalc6 = TempCalc3 * TempCalc2;
      //TempCalc6.dump(stdout, "TempCalc6: %p"); ////////
      TempCalc4 = VN * TempCalc6;
      //R.dump(stdout, "R: %p"); ////////
      //TempCalc4.dump(stdout, "TempCalc4: %p"); ////////
      TempCalc7 = R * TempCalc4;
      //TempCalc7.dump(stdout, "TempCalc7: %p"); ////////
      TempCalc5 = TempCalc4 * TempCalc7;
      //TempCalc5.dump(stdout, "TempCalc5: %p"); ////////

      fcmatrix EVtemp(NumMicrophones,1);
      //EVtemp.dump(stdout, "EVtemp: %p"); ////////
      TempCalc5.eigenValues(EVtemp,V);
      //TempCalc5.dump(stdout, "TempCalc5(2): %p"); ////////
      //EVtemp.dump(stdout, "EVtemp(2): %p"); ////////
      //V.dump(stdout, "V: %p"); ////////
      for (int ii = 0; ii < NumMicrophones; ii++){
          EigenValue[k][ii] = EVtemp.getElementAt(ii,0).real();
      }
      sort_eig_fvec_ascending(EigenValue[k], V, EigenValue[k], V);//
      //sort_eig_fvec_descending(EigenValue[k], V, EigenValue[k], V);//
#else
      invMAT(VN, TempCalc2);
      for (int i = 0; i < TempCalc3.size1(); i++) {
        for (int j = 0; j < TempCalc3.size2(); j++) {
          TempCalc3(i, j) = 0.0;
          if (i == j)
            TempCalc3(i, j) = REigenValue[i];
        }
      }
      noalias(TempCalc6) = prod(TempCalc3, TempCalc2);
      noalias(TempCalc4) = prod(VN, TempCalc6);
      noalias(TempCalc7) = prod(R, TempCalc4);
      noalias(TempCalc5) = prod(TempCalc4, TempCalc7);
      evd(TempCalc5, EigenValue[k], V);
#endif
    } else if (music_alg == "GSVD") {
#ifdef ENABLE_LAPACK
      TempCalc1 = RN.Inverse();
      TempCalc2= TempCalc1 * R;
      //TempCalc2.singularValueDecomposition(TempCalc3,TempCalc5,TempCalc4);//u,s,v
      TempCalc2.singularValueDecomposition(TempCalc4,TempCalc5,TempCalc3);//u,s,v
      for (int ii=0; ii<NumMicrophones; ii++){
          REigenValue[ii] = TempCalc5.getElementAt(ii,0).real();
      }
      sort_eig_fvec_ascending(REigenValue, TempCalc4, EigenValue[k], V);
      //sort_eig_fvec_descending(REigenValue, TempCalc4, EigenValue[k], V);
#else
      invMAT(RN, TempCalc1);
      noalias(TempCalc2) = prod(TempCalc1, R);
      svd(TempCalc2, REigenValue, TempCalc4, TempCalc3);//a,s,u,v
      sort_eig_fvec_ascending(REigenValue, TempCalc4, EigenValue[k], V);
#endif
    }

#ifdef ENABLE_LAPACK
    E[k] = V;
#else
    noalias(E[k]) = V;
#endif

  }

  //printf("\n");

  CalcAveragePower(EigenValue, E, A, UsesFreq, NumSource, Pavg, Pavg_bin, MSSW, enable_eigenvalue_weight);
  for (int ih = 0; ih < SearchId.size(); ih++) {
    Pavg_sum[ih] = 10 * log10(Pavg[ih] + 1.0f);
  }
  if (maxnum_output_peaks < 0)
    maxnum_output_peaks = NumSource;
  FoundSourceCount = SearchLocation(Pavg_sum, Pavg, maxnum_output_peaks, FoundId, mindeg, maxdeg);
#if ENABLE_TIMER
  end_time = std::chrono::system_clock::now();
  cout << DurationLog(end_time - start_time) << endl;
#endif

}

int LocalizationMUSIC::GetTFNumChannels(char* filename) {
  int nmic = -1;
#ifdef ENABLE_HARKIO3
  harkio_TransferFunction *tf = NULL;
  tf = harkio_TransferFunction_fromFile(filename, "transfer function");
  if(tf == NULL){
    cerr << "LocalizationMUSIC. Failed to read tf file." << endl;
    return -1;
  }
  if(tf->loctfs == NULL){
    cerr << "LocalizationMUSIC. No localization matrix in tf file." << endl;
    harkio_TransferFunction_delete(&tf);
    return -1;
  }
  nmic = tf->mics->size;
  if(tf->loctfs[0]->rows != nmic){
    cerr << "LocalizationMUSIC. The number of mics does not match to the size of tf file." << endl;
    harkio_TransferFunction_delete(&tf);
    return -1;
  }
  harkio_TransferFunction_delete(&tf);
#endif
  return nmic;
}

bool LocalizationMUSIC::ReadTransferFunction(char* filename, vector<int>& ch) {

#ifdef ENABLE_HARKIO3

  harkio_TransferFunction *tf = NULL;
  tf = harkio_TransferFunction_fromFile(filename, "transfer function");
  // tf = harkio_TransferFunction_new();
  // libharkio3_TransferFunction_readlibhakio2(&tf, filename, NULL);
  // ck_assert(tf != NULL);
  if (tf == NULL) {
    cerr << "LocalizationMUSIC. Failed to read tf file." << endl;
    return false;
  }

  if (tf->loctfs == NULL) {
    cerr << "LocalizationMUSIC. No localization matrix in tf file." << endl;
    harkio_TransferFunction_delete(&tf);
    return false;
  }
  int slength = tf->cfg->nfft / 2 + 1;
  if (tf->loctfs[0]->cols != slength) {
    cerr << "LocalizationMUSIC. The number of bins does not match to the size of tf file." << endl;
    harkio_TransferFunction_delete(&tf);
    return false;
  }

  int nsrc = tf->poses->size;
  // if ( tf->loctfs.size() != nsrc) {
  //   cerr << "LocalizationMUSIC. The number of loctfs does not match to the number of srcs." << endl;
  //   harkio_TransferFunction_delete(&tf);
  //   return false;
  // }
  if (tf->nbrs->size != nsrc) {
    cerr << "LocalizationMUSIC. The number of loctfs does not match to the size of nbrs." << endl;
    harkio_TransferFunction_delete(&tf);
    return false;
  }

  int nmic = tf->mics->size;
  if (tf->loctfs[0]->rows != nmic) {
    cerr << "LocalizationMUSIC. The number of mics does not match to the size of tf file." << endl;
    harkio_TransferFunction_delete(&tf);
    return false;
  }

  int nmic2 = ch.size();
  for (int c = 0; c < ch.size(); c++) {
    if (!(ch[c] >= 0 && ch[c] < nmic)) {
      cerr << "LocalizationMUSIC. Nonexistent channel is selected." << endl;
      harkio_TransferFunction_delete(&tf);
      return false;
    }
  }
#ifdef ENABLE_LAPACK
  A.resize(nsrc);
#else
  A.resize(extents[nsrc]);
#endif


  for (int iid = 0; iid < nsrc; iid++) {
#ifdef ENABLE_LAPACK
    fcmatrix Amat(slength, nmic2);
    A[iid] = Amat;
#else
    A[iid].resize(slength, nmic2);
#endif
  }
  COMPLEX_TYPE complexValue;
  for (int iid = 0; iid < nsrc; iid++) {
    for (int m = 0; m < nmic2; m++) {
      for (int k = 0; k < slength; k++) {
        harkio_Matrix_getValueComplex(tf->loctfs[iid], ch[m], k, &complexValue);
#ifdef ENABLE_LAPACK
        A[iid].setElementAt(k,m,complexValue);
#else
        A[iid](k, m) = complexValue;
#endif
      }
    }
  }

  AId.resize(nsrc);
  for (int iid = 0; iid < nsrc; iid++) {
    AId[iid] = tf->poses->pos[iid]->id;
  }

  SearchId.resize(nsrc);
  for (int iid = 0; iid < nsrc; iid++) {
    SearchId[iid].resize(3);
    if (tf->poses->pos[iid]->sys == Cartesian) {
      SearchId[iid][0] = tf->poses->pos[iid]->coord[0];
      SearchId[iid][1] = tf->poses->pos[iid]->coord[1];
      SearchId[iid][2] = tf->poses->pos[iid]->coord[2];
    }
    if (tf->poses->pos[iid]->sys == Polar) {
      float azimuth_radian = harkio_Position_convToRadian(tf->poses->pos[iid]->coord[0]);
      float elevation_radian = harkio_Position_convToRadian(tf->poses->pos[iid]->coord[1]);
      float coef = tf->poses->pos[iid]->coord[2] * cosf(elevation_radian);
      SearchId[iid][0] = coef * cosf(azimuth_radian);
      SearchId[iid][1] = coef * sinf(azimuth_radian);
      SearchId[iid][2] = tf->poses->pos[iid]->coord[2] * sinf(elevation_radian);
    }
  }

  nbrs_ids.resize(tf->nbrs->size);
  for (int i = 0; i < tf->nbrs->size; i++) {
    nbrs_ids[i] = tf->nbrs->ids[i];
  }
  nbrs_numNeighbors.resize(tf->nbrs->size);
  for (int i = 0; i < tf->nbrs->size; i++) {
    nbrs_numNeighbors[i] = tf->nbrs->numNeighbors[i];
  }
  nbrs_neighbors.resize(tf->nbrs->size);
  for (int i = 0; i < tf->nbrs->size; i++) {
    nbrs_neighbors[i].resize(nbrs_numNeighbors[i]);
    for (int j = 0; j < nbrs_numNeighbors[i]; j++) {
      nbrs_neighbors[i][j] = tf->nbrs->neighbors[i][j];
    }
  }

  harkio_TransferFunction_delete(&tf);

#endif  

  Pavg_sum.resize(SearchId.size());
  Pavg.resize(SearchId.size());
#ifdef ENABLE_LAPACK
  Pavg_bin.resize(A[0].getRowCount());
  for (int bins = 0; bins < A[0].getRowCount(); bins++) {
    Pavg_bin[bins].resize(SearchId.size());
  }

  cerr << SearchId.size() << " srcs, " << A[0].getColumnCount() << " microphones, " << (A[0].getRowCount() - 1) * 2 << " points"
      << endl;
#else
  Pavg_bin.resize(A[0].size1());
  for (int bins = 0; bins < A[0].size1(); bins++) {
    Pavg_bin[bins].resize(SearchId.size());
  }

  cerr << SearchId.size() << " srcs, " << A[0].size2() << " microphones, " << (A[0].size1() - 1) * 2 << " points"
      << endl;
#endif

  return true;
}

void LocalizationMUSIC::CalcAveragePower(vector<fvector>& w, vector<fcmatrix>& E, AMatrixArray1D& A,
    vector<bool>& UsesFreq, int num_source, vector<float>& Pavg, vector<vector<float> >& Pavg_bin,
    fvector& Weight, bool enable_eigenvalue_weight) {
  int nh = Pavg.size();
  int num_mic = w[0].size();

  for (int ih = 0; ih < nh; ih++) {
    Pavg[ih] = 0.0f;
  }

  for (int k = 0; k < Pavg_bin.size(); k++) {
    for (int ih = 0; ih < Pavg_bin[0].size(); ih++) {
      Pavg_bin[k][ih] = 0.0f;
    }
  }

  float G = 0.0f;
  float norm_a = 0.0f;

#ifdef ENABLE_LAPACK
  for (int k = 0; k < UsesFreq.size(); k++) {
    if (!UsesFreq[k]) {
      continue;
    }
    for (int ih = 0; ih < nh; ih++) {
      G = 0.0f;
      for (int m = num_mic - num_source - 1; m >= 0; m--) {
		G += norm(A[ih].getRow(k).transposed(true).elementwiseMultiplied(E[k].getColumn(m)).getElementSum());
        //G += norm(ublas::inner_prod(ublas::herm(ublas::row(A[ih], k)), ublas::column(E[k], m)));
      }
      //norm_a = ublas::inner_prod(ublas::herm(ublas::row(A[ih], k)), ublas::row(A[ih], k)).real();
	  norm_a = A[ih].getRow(k).conjugated().elementwiseMultiplied(A[ih].getRow(k)).getElementSum().real();
      float EV_Weight = (enable_eigenvalue_weight ? sqrt(w[k][num_mic - 1]) : 1.0);
      Pavg[ih] += (norm_a / G) * Weight[k] * EV_Weight;
      Pavg_bin[k][ih] = 10 * log10((norm_a / G * Weight[k]) + 1.0f);
      cerr << "";
    }
  }
#else

  for (int k = 0; k < UsesFreq.size(); k++) {
    if (!UsesFreq[k]) {
      continue;
    }
    for (int ih = 0; ih < nh; ih++) {
      G = 0.0f;
      for (int m = num_mic - num_source - 1; m >= 0; m--) {
        G += norm(ublas::inner_prod(ublas::herm(ublas::row(A[ih], k)), ublas::column(E[k], m)));
      }
      norm_a = ublas::inner_prod(ublas::herm(ublas::row(A[ih], k)), ublas::row(A[ih], k)).real();
      float EV_Weight = (enable_eigenvalue_weight ? sqrt(w[k][num_mic - 1]) : 1.0);
      Pavg[ih] += (norm_a / G) * Weight[k] * EV_Weight;
      Pavg_bin[k][ih] = 10 * log10((norm_a / G * Weight[k]) + 1.0f);
      cerr << "";
    }
  }
#endif
}

int LocalizationMUSIC::SearchLocation(vector<float>& Pavg_sum, vector<float>& Pavg, int n_source,
    vector<int>& direction, int mindeg, int maxdeg) {
  int n_result;
  int nh = Pavg.size();
  vector<int> candidate_id;
  vector<float> candidate_val;

  candidate_id.reserve(nh);
  candidate_id.resize(0);
  candidate_val.reserve(nh);
  candidate_val.resize(0);
  SearchPeak(Pavg_sum, candidate_id, candidate_val, mindeg, maxdeg);

  vector<int> index;
  for (int i = 0; i < candidate_id.size(); i++)
    index.push_back(i);

#ifdef ENABLE_LAPACK 
  sort(index.begin(), index.end(), std::bind<bool>(less_for_vec(), candidate_val, std::placeholders::_2, std::placeholders::_1));
#else
  sort(index.begin(), index.end(), boost::bind<bool>(less_for_vec(), candidate_val, _2, _1));
#endif


  vector<int> res_id;
  for (int i = 0; i < index.size(); i++)
    res_id.push_back(candidate_id[index[i]]);

  vector<float> res_val;
  for (int i = 0; i < index.size(); i++)
    res_val.push_back(candidate_val[index[i]]);

  if (n_source == 0) {
    n_result = res_id.size();
  } else if (res_id.size() < n_source) {
    n_result = res_id.size();
  } else {
    n_result = n_source;
  }

  direction.resize(0);
  for (int i = 0; i < n_result; i++) {
    direction.push_back(res_id[i]);
  }

  return n_result;
}

void LocalizationMUSIC::SearchPeak(vector<float>& power, vector<int>& peak_cand_id, vector<float>& peak_cand_val,
    int mindeg, int maxdeg) {

  while (maxdeg - mindeg > 360)
    maxdeg -= 360;
  while (maxdeg - mindeg <= 0)
    maxdeg += 360;

  int nh = power.size();

  for (int ih = 0; ih < nh; ih++) {

    double x = (SearchId[ih][0] == -0.0) ? 0.0 : SearchId[ih][0];
    double y = (SearchId[ih][1] == -0.0) ? 0.0 : SearchId[ih][1];
    int deg = (int) (180.0 / M_PI * atan2(y, x));
    while (deg - mindeg > 360)
      deg -= 360;
    while (deg - mindeg <= 0)
      deg += 360;
    if ((deg < mindeg) || (deg > maxdeg))
      continue;
    float maxc = power[ih];
    bool pf = true;

    int i;
    for (i = 0; i < nbrs_ids.size(); i++) {
      if (nbrs_ids[i] == AId[ih])
        break;
    }
    for (int j = 0; j < nbrs_numNeighbors[i]; j++) {
      int jh;
      for (jh = 0; jh < AId.size(); jh++) {
        if (nbrs_neighbors[i][j] == AId[jh])
          break;
      }
      if (maxc < power[jh])
        pf = false;
    }
    if (pf) {
      peak_cand_id.push_back(ih);
      peak_cand_val.push_back(power[ih]);
    }

  }

}

#endif
