/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <iostream>
#include <memory>

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>
#include <Source.h>

#include "LocalizationBFDS.hpp"

using namespace HARK;

class LocalizeBFDS;

DECLARE_NODE(LocalizeBFDS);
/*Node
 *
 * @name LocalizeBFDS
 * @category HARK:Localization
 * @description Sound Source Localization by BFDS
 *
 * @input_name INPUT
 * @input_type Matrix<complex<float> >
 * @input_description Input multi-channel spectrum. A row is a channel, and a column is a spectrum for the corresponding channel. 
 *
 * @output_name OUTPUT
 * @output_type Vector<ObjectRef>
 * @output_description Source location. A type of an element of the vector is Source.
 *
 * @parameter_name TF_CHANNEL_SELECTION
 * @parameter_type object
 * @parameter_value <Vector<int> >
 * @parameter_description Microphone channels for localization. If vacant, all channels will be used.
 * 
 * @parameter_name LENGTH
 * @parameter_type int
 * @parameter_value 512
 * @parameter_description The length of a frame (per channel).
 * 
 * @parameter_name SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_description Sampling rate (Hz).
 * 
 * @parameter_name USE_TF_FILE 
 * @parameter_type bool 
 * @parameter_value true 
 * @parameter_list true:false 
 * @parameter_description If true, transfer function file is used. If false, microphone positions file and source positions file are used.
 * 
 * @parameter_name TF_FILENAME
 * @parameter_type string
 * @parameter_valid USE_TF_FILE=true
 * @parameter_description Name of a transfer function file. This is used only when USE_TF_FILE=true.
 * 
 * @parameter_name MIC_POSITIONS
 * @parameter_type string
 * @parameter_valid USE_TF_FILE=false
 * @parameter_description Name of a microphone positions file. This is used only when USE_TF_FILE=false.
 * 
 * @parameter_name SRC_POSITIONS
 * @parameter_type string
 * @parameter_valid USE_TF_FILE=false
 * @parameter_description Name of a source positions file This is used only when USE_TF_FILE=false.
 *
 * @parameter_name SPEED_OF_SOUND
 * @parameter_type float
 * @parameter_value 340.0
 * @parameter_valid USE_TF_FILE=false
 * @parameter_description Speed of sound (m/sec). This is used only when USE_TF_FILE=false.
 * 
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_description Debug option. If true, outputs debug to a standard output.
 * 
END*/

class LocalizeBFDS : public BufferedNode {
public:
	LocalizeBFDS(string nodeName, ParameterSet params)
	  : BufferedNode(nodeName, params), bfdsID(-1)
	{
		/*== Assign input/output ID ==*/

		this->inputID  = addInput("INPUT");
		this->outputID = addOutput("OUTPUT");

    	/*== Get parameters ==*/

		Vector<int> channelSelection;
		if (parameters.exist("TF_CHANNEL_SELECTION")) {
			channelSelection = object_cast<Vector<int> >(parameters.get("TF_CHANNEL_SELECTION"));      
		}

		int length = dereference_cast<int>(parameters.get("LENGTH"));
		if (length <= 0) {
			throw new NodeException(NULL, "LocalizeBFDS: LENGTH is invalid value.", __FILE__, __LINE__);
		}

		int samplingRate = dereference_cast<int>(parameters.get("SAMPLING_RATE"));
		if (samplingRate <= 0) {
			throw (new NodeException(NULL, "LocalizeBFDS: SAMPLING_RATE is invalid value.", __FILE__, __LINE__));
		}

		bool useTFfile = dereference_cast<bool>(parameters.get("USE_TF_FILE"));

		string filenameTF = "";
		if (parameters.exist("TF_FILENAME")) {
			filenameTF = object_cast<string>(parameters.get("TF_FILENAME"));
		}
		if (useTFfile && filenameTF.empty()) {
			throw new NodeException(NULL, "LocalizeBFDS: TF_FILENAME is not set.", __FILE__, __LINE__);
		}

		string filenameMICpos = "";
		if (parameters.exist("MIC_POSITIONS")) {
			filenameMICpos = object_cast<string>(parameters.get("MIC_POSITIONS"));
		}
		if (!useTFfile && filenameMICpos.empty()) {
			throw new NodeException(NULL, "LocalizeBFDS: MIC_POSITIONS is not set.", __FILE__, __LINE__);
		}

		string filenameSRCpos = "";
		if (parameters.exist("SRC_POSITIONS")) {
			filenameSRCpos = object_cast<string>(parameters.get("SRC_POSITIONS"));
		}
		if (!useTFfile && filenameSRCpos.empty()) {
			throw new NodeException(NULL, "LocalizeBFDS: SRC_POSITIONS is not set.", __FILE__, __LINE__);
		}

		float speedSound = 0;
		if (parameters.exist("SPEED_OF_SOUND")) {
			speedSound = dereference_cast<float>(parameters.get("SPEED_OF_SOUND"));
		}
		if (!useTFfile && speedSound <= 0) {
			throw new NodeException(NULL, "LocalizeBFDS: SPEED_OF_SOUND is invalid value.", __FILE__, __LINE__);
		}

		this->isDebug = dereference_cast<bool>(parameters.get("DEBUG"));

		/*== Initialize ==*/

		this->localizer.reset(new LocalizationBFDS(samplingRate, length, speedSound));

		this->numFrequency = length / 2 + 1;

		vector<int> chSelect;
		chSelect.resize(channelSelection.size());
		if (channelSelection.size() != 0) {
			for (int i = 0; i < channelSelection.size(); i++) {
				chSelect[i] = channelSelection[i];
			}
		}

		if (useTFfile) {
			// get transfer function and source positions
			if (!this->localizer->ReadTransferFunction(filenameTF, chSelect, this->numChannel, this->numSources)) {
				throw (new NodeException(NULL, "LocalizeBFDS: Transfer function file is invalid.", __FILE__, __LINE__));
			}
		} else {
			// get microphone positions
			if (!this->localizer->ReadXMLmicPos(filenameMICpos.c_str(), chSelect, this->numChannel)) {
				throw (new NodeException(NULL, "LocalizeBFDS: Microphone XML file is invalid.", __FILE__, __LINE__));
			}
			// get source positions
			if (!this->localizer->ReadXMLsrcPos(filenameSRCpos.c_str(), this->numSources)) {
				throw (new NodeException(NULL, "LocalizeBFDS: Source XML file is invalid.", __FILE__, __LINE__));
			}
		}
//cout << endl << "ch: " << numChannel << ", freq: " << numFrequency << ", source: " << numSource << endl << endl;

		// get BFDS filter

		if (useTFfile) {
			this->localizer->SetBfDsFilterfromTF(this->bfFilter);
		} else {
			this->localizer->SetBfDsFilterfromPOS(this->bfFilter);
		}

/*		if (this->isDebug) {
			printf("BFDS filter: ");
			for (int i = 0; i < this->numSources; i++) {
				for (int j = 0; j < this->numChannel; j++) {
					for (int k = 0; k < this->numFrequency; k++) {
						printf("%f+%fi ", 
						       this->bfFilter[i].getElementAt(j, k).real(), this->bfFilter[i].getElementAt(j, k).imag());
					}
				}
			}
			printf("\n");
		}
*/

		inOrder = true;
	}

	virtual void initialize()
	{
		this->BufferedNode::initialize();
	} 

	virtual int translateOutput (const string &outputName)
	{
		for (size_t i = 0; i < outputNames.size(); i++) {	
			if (outputNames[i] == outputName) {
				return i;
			}
		}
    
		if (outputName == "SPECTRUM") {
			return this->bfdsID = addOutput(outputName);
		} else {
//			throw new NodeException(this, "LocalizeBFDS: " + outputName+ " is not supported.", __FILE__, __LINE__);
			return addOutput(outputName);
		}
	}

	void calculate(int output_id, int count, Buffer &out)
	{
		/*== set input/output ==*/

		// input spectrum
		RCPtr<Matrix<complex<float> > > inputp = getInput(this->inputID, count);

		// output source
		RCPtr<Vector<ObjectRef> > sources(Vector<ObjectRef>::alloc(0));
		(*(outputs[this->outputID].buffer))[count] = sources;

		// output BF spectrum power
		RCPtr<Vector<float > > spec(new Vector<float >(this->numSources));
		if (this->bfdsID != -1) {
			(*(outputs[this->bfdsID].buffer))[count] = spec;
			for (unsigned int i = 0; i < this->numSources; i++) {
				(*spec)[i] = 0.0f;
			}
		}

		/*== input data ==*/

		// spectrum data

		if (count == 0) {
			if (this->numChannel != (*inputp).nrows()) {
				throw new NodeException(this, 
				      "LocalizeBFDS: Input channel size is different from parametr TF_CHANEL_SELECTION.", 
				      __FILE__, __LINE__);
			}
			if (this->numFrequency != (*inputp).ncols()) {
				throw new NodeException(this, 
				      "LocalizeBFDS: Input frame size is different from parameter LENGTH.", __FILE__, __LINE__);
			}
		}

		BF::Matrix x(this->numChannel, this->numFrequency);
		for (int j = 0; j < this->numChannel; j++) {
			for (int k = 0; k < this->numFrequency; k++) {
				x.setElementAt(j, k, (*inputp)(j, k));
			}
		}

		/*== localize ==*/

		int foundId;
		vector<float> bfPower;
		bfPower.resize(this->numSources); 

		this->localizer->Localize(x, this->bfFilter, bfPower, foundId);

		/*== output data ==*/

		// output source

		RCPtr<Source> src(new Source);

		const BF::SourceLocation &sl = this->localizer->GetSrcPositions()[foundId];

		src->x[0] = sl.x;
		src->x[1] = sl.y;
		src->x[2] = sl.z;
		src->power = bfPower[foundId];

		src->id = 0;
		src->tfindex[0] = foundId; //sl.sourceID ?
		src->tfindex[1] = 0;
		src->tfindex[2] = 0;
		src->tfindex_max[0] = this->localizer->GetSrcPositions().size() - 1;
		src->tfindex_max[1] = 0;
		src->tfindex_max[2] = 0;
		src->tfindex_min[0] = 0;
		src->tfindex_min[1] = 0;
		src->tfindex_min[2] = 0;
		src->compare_mode = Source::TFINDEX;

		if (this->isDebug) {
			printf("(Frame %d) Source: %d %f %f %f %f\n", count, foundId, src->x[0], src->x[1], src->x[2], src->power);
		}

		sources->push_back(src);

		// output BF spectrum power

		if (this->bfdsID != -1) {
			for (int i = 0; i < this->numSources; i++) {
				(*spec)[i] = bfPower[i];
			}
		}

		if (this->isDebug) {
			printf("Power of BF spectrum:\n");
			for (int i = 0; i < this->numSources; i++) {
				printf("%f ", bfPower[i]);
			}
			printf("\n");
		}

	}

private:
	// input/output

	int inputID;
	int outputID;
	int bfdsID;

	// parameter

	bool isDebug;

	// variable

	int numFrequency;
	int numChannel;
	int numSources;
	vector<BF::Matrix> bfFilter;

	std::unique_ptr<LocalizationBFDS> localizer;

	//IN_ORDER_NODE_SPEEDUP(LocalizeBFDS);
};

