/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Map.h>

#include "fft.hpp"

//using namespace std;
using namespace FD;

class MSLSExtraction;

DECLARE_NODE(MSLSExtraction)
/*Node
 *
 * @name MSLSExtraction
 * @category HARK:FeatureExtraction
 * @description Calculate a feature vector from a Mel-scale spectrum.
 *
 * @input_name FBANK
 * @input_type Map<int,ObjectRef>
 * @input_description Input Mel-scale spectrum. The key is source ID, and the value is the feature vector (Vector<float>).
 *
 * @input_name SPECTRUM
 * @input_type Map<int,ObjectRef>
 * @input_description Input spectrum. The key is source ID, and the value is the spectrum vector (Vector<complex<float> >).
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description MSLS feature vector. The key is source ID, and the value is MFCC feature vector (Vector<float>).
 *
 * @parameter_name FBANK_COUNT
 * @parameter_type int
 * @parameter_value 13
 * @parameter_description Size of the static part of MSLS feature vector. [default: 13]
 *
 * @parameter_name NORMALIZATION_MODE
 * @parameter_type string
 * @parameter_value CEPSTRAL
 * @parameter_list CEPSTRAL:SPECTRAL
 * @parameter_description The domain to perform normalization. CEPSTRAL or SPECTRAL. [default: CEPSTRAL]
 *
 * @parameter_name USE_LEGACY_MODE
 * @parameter_type bool
 * @parameter_value true
 * @parameter_list true:false
 * @parameter_valid NORMALIZATION_MODE=CEPSTRAL
 * @parameter_description For more than 14 dimensions must use false. This parameter is preparing only for compatibility with HARK 2.x or earlier. [default: true]
 *
 * @parameter_name USE_HTK_LIFTER
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_valid NORMALIZATION_MODE=CEPSTRAL
 * @parameter_description Use HTK liftering vector if true. [default: false]
 *
 * @parameter_name LIFTERING_COEF
 * @parameter_type int
 * @parameter_value 22
 * @parameter_valid NORMALIZATION_MODE=CEPSTRAL
 * @parameter_description The HTK liftering coefficient used in Cepstral mode. [default: 22]
 *
 * @parameter_name USE_POWER
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_description Use power feature if true. [default: false]
 *
 END*/

class MSLSExtraction : public BufferedNode {
  int fbankID;
  int spectrumID;
  int outputID;

  int fbank_count;

  int duration;
  enum {LEGACY_MODE, CEPSTRAL_MODE, SPECTRAL_MODE} normalization_mode;
  bool use_power;
  int dim_static;

  std::vector<float> lifter;
  std::vector<float> cepstrum;

public:
  MSLSExtraction(std::string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params), spectrumID(-1)
  {
    fbankID = addInput("FBANK");
    //spectrumID = addInput("SPECTRUM");
    outputID = addOutput("OUTPUT");

    fbank_count = dereference_cast<int>(parameters.get("FBANK_COUNT"));
    if (fbank_count>0) {
    }
    else{
      throw (new NodeException(NULL,"MSLS:FBANK_COUNT must be positive value.",__FILE__,__LINE__)) ;
    }
    std::string mode = object_cast<String>(parameters.get("NORMALIZATION_MODE"));
    bool use_lecacy_mode = dereference_cast<bool>(parameters.get("USE_LEGACY_MODE"));
    if (mode == "CEPSTRAL") {
      normalization_mode = use_lecacy_mode ? LEGACY_MODE : CEPSTRAL_MODE;
    }
    else if (mode == "SPECTRAL") {
      normalization_mode = SPECTRAL_MODE;
    }
    else{
      throw (new NodeException(NULL,"MSLS:NORMALIZATION_MODE is not CEPSTRAL nor SPECTRAL",__FILE__,__LINE__)) ;
    }
    bool use_htk_lifter = dereference_cast<bool>(parameters.get("USE_HTK_LIFTER"));
    int coef = dereference_cast<int>(parameters.get("LIFTERING_COEF"));
    use_power = dereference_cast<bool>(parameters.get("USE_POWER"));
    dim_static = fbank_count + (use_power ? 1 : 0);

    lifter.resize(fbank_count);
    if(use_htk_lifter){
      if(coef <= 0){ coef = 1; }
      for(int i = 0; i < lifter.size(); i++){
        lifter[i] = 1.0f + (coef / 2.0f) * std::sin((i + 1) * (M_PI / coef));
      }
    }
    else{
      for(int i = 0; i < lifter.size(); i++){
        lifter[i] = 1.0f;
      }
    }
    lifter[0] = 0.0f;
    cepstrum.resize(fbank_count);

    inOrder = true;
  }

  virtual int translateInput(const std::string &inputName)
  {
    for (unsigned int i = 0; i < inputs.size(); i++) {
      if (inputs[i].name == inputName) {
        return i;
      }
    }
    if (inputName == "SPECTRUM") {
      return spectrumID = addInput(inputName);
    }
    throw new NodeException(this, std::string("Unknown input in translateInput the name must be SPECTRUM: ") + inputName,
                            __FILE__, __LINE__);
  }

  virtual void verifyConnect()
  {
    if (use_power && spectrumID == -1) {
      throw new NodeException(this, "USE_POWER is true but SPECTRUM is not properly connected.",
                              __FILE__, __LINE__);
    }
    //BufferedNode::verifyConnect();
  }

  void calculate(int output_id, int count, Buffer &out)
  {
    RCPtr<Map<int, ObjectRef> > fbanks_ptr = getInput(fbankID, count);
    const Map<int, ObjectRef>& fbanks = *fbanks_ptr;
    RCPtr<Map<int, ObjectRef> > spectra(NULL);
    if(use_power){
      spectra = getInput(spectrumID, count);
    }

    Buffer& outputBuffer = *(outputs[outputID].buffer);
        
    Map<int,ObjectRef>& output = *new Map<int,ObjectRef>;
    outputBuffer[count] = &output;

    for (Map<int, ObjectRef>::const_iterator it = fbanks.begin();
         it != fbanks.end(); ++it) {
      const Vector<float>& in = object_cast<Vector<float> >(it->second);
      Vector<float>& fbank = *Vector<float>::alloc(dim_static * 2);
      output[it->first] = &fbank;

      int i = 0;
      while(i < fbank_count){
        fbank[i] = std::log(in[i] < 1.0f ? 1.0f : in[i]);
        i++;
      }
      while(i < fbank.size()){
        fbank[i] = 0.0f;
        i++;
      }

      switch (normalization_mode) {
      case LEGACY_MODE:
        LifterOld(fbank);
        break;
      case CEPSTRAL_MODE:
        Lifter(fbank);
        break;
      case SPECTRAL_MODE:
        SubtractMean(fbank);
        EmphasizeSpectralPeak(fbank);
        break;
      }

      if(use_power){
        Map<int, ObjectRef>::iterator itspec = spectra->find(it->first);
        float power = 0.0f;
        if(itspec != spectra->end()){
          Vector<std::complex<float> >& spec =
            object_cast<Vector<std::complex<float> > >(itspec->second);

          int k = 0;
          power += std::norm(spec[k++]);
          while(k < spec.size() - 1){
            power += std::norm(spec[k++]) * 2;
          }
          power += std::norm(spec[k]);
          power /= (spec.size() - 1) * 2;

          //// This code is fast, but there are slight differences in results.
          //int k = spec.size() - 1;
          //power += std::norm(spec[k--]);
          //while(k > 0){
          //  power += std::norm(spec[k--]) * 2;
          //}
          //power += std::norm(spec[k]);
          //power /= (spec.size() - 1) * 2;

        }
        fbank[dim_static - 1] = (power >= 1e-40 ? std::log(power) : 0.0f);
      }
    }
  }

  void Lifter(Vector<float>& fbank)
  {
    dct(&fbank[0], &cepstrum[0], fbank_count);
    for (int i = 0; i < fbank_count; i++){
      cepstrum[i] *= lifter[i];
    }
    idct(&cepstrum[0], &fbank[0], fbank_count);
  }

  void LifterOld(Vector<float>& fbank)
  {
    dct(&fbank[0], &cepstrum[0], fbank_count);
    cepstrum[0] = 0.0f;
    for(int i = 13; i < fbank_count; i++){
      cepstrum[i] = 0.0f;
    }
    idct(&cepstrum[0], &fbank[0], fbank_count);
  }

  void SubtractMean(Vector<float>& fbank)
  {
    float mean = 0.0f;

    for (int i = 0; i < fbank_count; i++) {
      mean += fbank[i];
    }
    mean /= fbank_count;

    for (int i = 0; i < fbank_count; i++) {
      fbank[i] -= mean;
    }
  }
    
  void dct(float* x, float* y, int n)
  {
    for (int i = 0; i < n; i++) {
      y[i] = 0;
      for (int j = 0; j < n; j++) {
        y[i] += x[j] * std::cos(M_PI * (j + 0.5f) * i / n);
      }
      y[i] *= 2.0f;
    }
  }

  void idct(float* x, float* y, int n)
  {
    const float a = 1.0 / n;
    for (int i = 0; i < n; i++) {
      y[i] = x[0] * 0.5f;
      for (int j = 1; j < n; j++) {
        y[i] += x[j] * std::cos(M_PI * j * (i + 0.5f) / n);
      }
      y[i] *= a;
    }
  }

  void EmphasizeSpectralPeak(Vector<float>& vec)
  {
    const float p = 0.9f;

    for (int i = vec.size() - 1; i > 0; i--) {
      vec[i] -= p * vec[i - 1];
    }
    // 旧wav2nsspecと合わせるための修正
    //vec[0] *= 1.0 - p;
  }

  IN_ORDER_NODE_SPEEDUP(MSLSExtraction);
};
