/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <iostream>

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>
#include <Map.h>
#include <operators.h>

using namespace std;
using namespace FD;

class MapOperator;

DECLARE_NODE(MapOperator)
/*Node
 *
 * @name MapOperator
 * @category HARK:MISC
 * @description Mathematical operations for Map<int, ObjectRef> which have the same ID. 
 *
 * @input_name INPUT1
 * @input_type Map<int,ObjectRef>
 * @input_description The key is source ID, and the value is arbitral.
 *
 * @input_name INPUT2
 * @input_type Map<int,ObjectRef>
 * @input_description The key is source ID, and the value is arbitral.
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description The key is source ID, and the value is the same type as INPUT1 and INPUT2.
 *
 * @parameter_name OPERATION_TYPE
 * @parameter_type string
 * @parameter_value Add
 * @parameter_list Add:Sub:Mul:Div:Equal:Max:Min:Concat
 * @parameter_description Operation Type. [default: Add]
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Show debug points. [default: false]
 *
 END*/

static const int OPERATION_ADD = 0;
static const int OPERATION_SUB = 1;
static const int OPERATION_MUL = 2;
static const int OPERATION_DIV = 3;
static const int OPERATION_EQUAL = 4;
static const int OPERATION_MAX = 5;
static const int OPERATION_MIN = 6;
static const int OPERATION_CONCAT = 7;

class MapOperator : public BufferedNode {
	int input1ID;
	int input2ID;
	int outputID;
	int operation_type;
	bool debug;
	String operation_name;

public:
	MapOperator(string nodeName, ParameterSet params)
		: BufferedNode(nodeName, params) {
		input1ID = addInput("INPUT1");
		input2ID = addInput("INPUT2");
    	outputID = addOutput("OUTPUT");
		operation_name = object_cast<String>(parameters.get("OPERATION_TYPE"));
		if      (operation_name == "Add")    { operation_type = OPERATION_ADD; }
		else if (operation_name == "Sub")    { operation_type = OPERATION_SUB; }
		else if (operation_name == "Mul")    { operation_type = OPERATION_MUL; }
		else if (operation_name == "Div")    { operation_type = OPERATION_DIV; }
		else if (operation_name == "Equal")  { operation_type = OPERATION_EQUAL; }
		else if (operation_name == "Max")    { operation_type = OPERATION_MAX; }
		else if (operation_name == "Min")    { operation_type = OPERATION_MIN; }
		else if (operation_name == "Concat") { operation_type = OPERATION_CONCAT; }
		else {
			cerr << "******* MapOperator: Input error" << endl;
			throw new NodeException(NULL, "MapOperator: OPERATION_TYPE should be \"Add\" or \"Sub\" or \"Mul\" or \"Div\"or \"Equal\" or \"Max\" or \"Min\" or \"Concat\".", __FILE__, __LINE__);
		}
		debug = dereference_cast<bool>(parameters.get("DEBUG"));
	}

	void calculate(int output_id, int count, Buffer &out) {
		if (debug == true) cout << "******* MapOperator debug point-Input check (count=" << count << ")" << endl;

		ObjectRef input1 = getInput(input1ID, count);
		ObjectRef input2 = getInput(input2ID, count);
		if ((typeid(*input1) == typeid(Map<int, ObjectRef>)) && (typeid(*input2) == typeid(Map<int, ObjectRef>))) {
			RCPtr<Map<int, ObjectRef> > inp1 = input1;
			const Map<int, ObjectRef>& in1 = *inp1;
			RCPtr<Map<int, ObjectRef> > inp2 = input2;
			const Map<int, ObjectRef>& in2 = *inp2;

			if (debug == true) {
				if (in1.empty() == true) {
					cout << "input1 map empty" << endl;
				}
				else {
					cout << "Input1 map begins from " << (in1.begin()->first) << endl;
					cout << "Number of channels: " << in1.size() << endl;
					cout << "Type of first channel of ObjectRef of input1 map is: ";
					if (typeid(*(in1.begin()->second)) == typeid(Float)) {
						cout << "Float" << endl;
					}
					else if (typeid(*(in1.begin()->second)) == typeid(Complex<float>)) {
						cout << "Complex<float>" << endl;
					}
					else if (typeid(*(in1.begin()->second)) == typeid(Vector<float>)) {
						cout << "Vector<float>";
						cout << ", size=" << (object_cast<Vector<float> >(in1.begin()->second)).size() << endl;
					}
					else if (typeid(*(in1.begin()->second)) == typeid(Vector<complex<float> >)) {
						cout << "Vector<complex<float> >";
						cout << ", size=" << (object_cast<Vector<complex<float> > >(in1.begin()->second)).size() << endl;
					}
					else if (typeid(*(in1.begin()->second)) == typeid(Matrix<float>)) {
						cout << "Matrix<float>";
						cout << ", rows=" << (object_cast<Matrix<float> >(in1.begin()->second)).nrows();
						cout << ", cols=" << (object_cast<Matrix<float> >(in1.begin()->second)).ncols() << endl;
					}
					else if (typeid(*(in1.begin()->second)) == typeid(Matrix<complex<float> >)) {
						cout << "Matrix<complex<float> >";
						cout << ", rows=" << (object_cast<Matrix<complex<float> > >(in1.begin()->second)).nrows();
						cout << ", cols=" << (object_cast<Matrix<complex<float> > >(in1.begin()->second)).ncols() << endl;
					}
					else {
						cout << "unidentified." << endl;
					}
				}

				if (in2.empty() == true) {
					cout << "input2 map empty" << endl;
				}
				else {
					cout << "Input2 map begins from " << (in2.begin()->first) << endl;
					cout << "Number of channels: " << in2.size() << endl;
					cout << "Type of first channel of ObjectRef of input2 map is: ";
					if (typeid(*(in2.begin()->second)) == typeid(Float)) {
						cout << "Float" << endl;
					}
					else if (typeid(*(in2.begin()->second)) == typeid(Complex<float>)) {
						cout << "FDComplex<Float>" << endl;
					}
					else if (typeid(*(in2.begin()->second)) == typeid(Vector<float>)) {
						cout << "Vector<float>";
						cout << ", size=" << (object_cast<Vector<float> >(in2.begin()->second)).size() << endl;
					}
					else if (typeid(*(in2.begin()->second)) == typeid(Vector<complex<float> >)) {
						cout << "Vector<complex<float> >";
						cout << ", size=" << (object_cast<Vector<complex<float> > >(in2.begin()->second)).size() << endl;
					}
					else if (typeid(*(in2.begin()->second)) == typeid(Matrix<float>)) {
						cout << "Matrix<float>";
						cout << ", rows=" << (object_cast<Matrix<float> >(in2.begin()->second)).nrows();
						cout << ", cols=" << (object_cast<Matrix<float> >(in2.begin()->second)).ncols() << endl;
					}
					else if (typeid(*(in2.begin()->second)) == typeid(Matrix<complex<float> >)) {
						cout << "Matrix<complex<float> >";
						cout << ", rows=" << (object_cast<Matrix<complex<float> > >(in2.begin()->second)).nrows();
						cout << ", cols=" << (object_cast<Matrix<complex<float> > >(in2.begin()->second)).ncols() << endl;
					}
					else {
						cout << "unidentified." << endl;
					}
				}
			}

			if (typeid(*(in1.begin()->second)) != typeid(*(in2.begin()->second))) {
				cerr << "******* MapOperator: Types of ObjectRef of input1 and input2 are not same." << endl;
				throw(new NodeException(NULL, "Exception caught in calculate[MapOperator]", __FILE__, __LINE__));
			}

			Map<int,ObjectRef>& output = *new Map<int,ObjectRef>;
			out[count] = &output;

			if (in1.empty() == false) {
				for (Map<int, ObjectRef>::const_iterator it1 = in1.begin(); it1 != in1.end(); ++it1) {
					ObjectRef oOut;
					ObjectRef oIn1 = it1->second;

					// make output clone of input1
					if (typeid(*oIn1) == typeid(Vector<float>)) {
						const Vector<float>& spec = object_cast<Vector<float> >(oIn1);
						RCPtr<Vector<float> > fbank(new Vector<float>(spec.size()));
						for (int i = 0; i < spec.size(); i++) {
							(*fbank)[i] = spec[i];
						}
						oOut = ObjectRef(fbank);
					}
					else if(typeid(*oIn1) == typeid(Vector<complex<float> >)) {
						const Vector<complex<float> >& spec = object_cast<Vector<complex<float> > >(oIn1);
						RCPtr<Vector<complex<float> > > fbank(new Vector<complex<float> >(spec.size()));
						for (int i = 0; i < spec.size(); i++) {
							(*fbank)[i] = spec[i];
						}
						oOut = ObjectRef(fbank);
					}
					else if (typeid(*oIn1) == typeid(Matrix<float>)) {
						const Matrix<float>& spec = object_cast<Matrix<float> >(oIn1);
						RCPtr<Matrix<float> > fbank(new Matrix<float>(spec.nrows(), spec.ncols()));
						for (int row = 0; row < spec.nrows(); row++) {
							for (int col = 0; col < spec.ncols(); col++) {
								(*fbank)(row, col) = spec(row, col);
							}
						}
						oOut = ObjectRef(fbank);
					}
					else if (typeid(*oIn1) == typeid(Matrix<complex<float> >)) {
						const Matrix<complex<float> >& spec = object_cast<Matrix<complex<float> > >(oIn1);
						RCPtr<Matrix<complex<float> > > fbank(new Matrix<complex<float> >(spec.nrows(), spec.ncols()));
						for (int row = 0; row < spec.nrows(); row++) {
							for (int col = 0; col < spec.ncols(); col++) {
								(*fbank)(row, col) = spec(row, col);
							}
						}
						oOut = ObjectRef(fbank);
					}
					else {
						oOut = oIn1;
					}

					Map<int, ObjectRef>::const_iterator it2 = in2.find(it1->first);

					if (it2 != in2.end()) {
						ObjectRef oIn2 = it2->second;
					
						if      (operation_type == OPERATION_ADD)    oOut = oIn1 + oIn2;
						else if (operation_type == OPERATION_SUB)    oOut = oIn1 - oIn2;
						else if (operation_type == OPERATION_MUL)    oOut = oIn1 * oIn2;
						else if (operation_type == OPERATION_DIV)    oOut = oIn1 / oIn2;
						else if (operation_type == OPERATION_EQUAL)  oOut = oIn1 == oIn2;
						else if (operation_type == OPERATION_MAX)    oOut = max(oIn1, oIn2);
						else if (operation_type == OPERATION_MIN)    oOut = min(oIn1, oIn2);
						else if (operation_type == OPERATION_CONCAT) oOut = concat(oIn1, oIn2);
					}

					output[it1->first] = oOut;
				}
			}
		}
		else {
			cerr << "******* MapOperator: Only Map<int, ObjectRef> is supported." << endl;
			throw(new NodeException(NULL, "Exception caught in calculate[MapOperator]", __FILE__, __LINE__));
		}

		if (debug == true) cout << "******* MapOperator debug point-Finish" << endl;
  }

	NO_ORDER_NODE_SPEEDUP(MapOperator);

};
