/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <iostream>

#include <BufferedNode.h>
#include <Buffer.h>
#include <Matrix.h>
#include <Map.h>

using namespace std;
using namespace FD;

class MapToMatrix;

DECLARE_NODE(MapToMatrix)
/*Node
 *
 * @name MapToMatrix
 * @category HARK:MISC
 * @description Convert Map<int, ObjectRef> to Matrix<float> or Matrix<complex<float> >. If the input is Map<int, Matrix<float> >, the output is Matrix<float>. If the input is Map<int, Matrix<complex<float> > >, the output is Matrix<complex<float> >.
 *
 * @input_name INPUT
 * @input_type Map<int,ObjectRef>
 * @input_description The key is source ID, and the value is Matrix<float> or Matrix<complex<float> >.
 *
 * @output_name OUTPUT
 * @output_type any	
 * @output_description Matrix<float> or Matrix<complex<float> >.
 *
 * @parameter_name METHOD
 * @parameter_type string
 * @parameter_list min_key:max_key:average:summation
 * @parameter_value min_key
 * @parameter_description Method of transform Map to Matrix. [default: min_key]
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Show debug points. [default: false]
 *
 END*/

static const int METHOD_MIN_KEY = 0;
static const int METHOD_MAX_KEY = 1;
static const int METHOD_AVERAGE = 2;
static const int METHOD_SUMMATION = 3;

class MapToMatrix : public BufferedNode {
	int inputID;
	int outputID;
	int method;
	bool debug;
  
public:
	MapToMatrix(string nodeName, ParameterSet params)
		: BufferedNode(nodeName, params) {
		inputID = addInput("INPUT");
		outputID = addOutput("OUTPUT");
		string method_name = object_cast<String>(parameters.get("METHOD"));
		if (method_name == "min_key") {
			method = METHOD_MIN_KEY;
		}
		else if (method_name == "max_key") {
			method = METHOD_MAX_KEY;
		}
		else if (method_name == "average") {
			method = METHOD_AVERAGE;
		}
		else if (method_name == "summation") {
			method = METHOD_SUMMATION;
		}
		else {
			cerr << "******* MapToMatrix: Input error" << endl;
			throw new NodeException(NULL, "MapToMatrix: METHOD should be \"min_key\" or \"max_key\" or \"average\" or \"summation\".", __FILE__, __LINE__);
		}
		debug = dereference_cast<bool>(parameters.get("DEBUG"));
	}

	void calculate(int output_id, int count, Buffer &out) {
		RCPtr<Map<int, ObjectRef> > input_ptr =  getInput(inputID, count);
		const Map<int, ObjectRef> input_map = *input_ptr;

		if (debug == true ) {
			cout << "******* MapToMatrix debug point-Input check (count=" << count << ")" << endl;
			if (input_map.empty() == true) {
				cout << "map empty" << endl;
			}
			else {
				cout <<  "Input map begins from " << (input_map.begin()->first) << endl;
				cout <<  "Input map's ObjectRef's first channel's type is: ";
				if (typeid(*(input_map.begin()->second)) == typeid(Matrix<float>))                cout << "Matrix<float>" << endl;
				else if (typeid(*(input_map.begin()->second)) == typeid(Matrix<complex<float> >)) cout << "Matrix<complex<float> >" << endl;
				else                                                                              cout << "unidentified." << endl;
			}
		}

		if (input_map.empty() == true) {
			RCPtr<Matrix<float> > output(new Matrix<float>);
			out[count] = output;
		}
		else {
			int maxid = -1;
			int minid = 10000;    
			Map<int, ObjectRef>::const_iterator it = input_map.begin();

			if (typeid(*(it->second)) == typeid(Matrix<float>)) {
				const Matrix<float> &mtFirst = object_cast<Matrix<float> >(it->second);
				RCPtr<Matrix<float> > output(new Matrix<float>(mtFirst.nrows(), mtFirst.ncols()));
				out[count] = output;

				if (method == METHOD_AVERAGE || method == METHOD_SUMMATION) { // initialize zero
					for (int row = 0; row < mtFirst.nrows(); row++) {
						for (int col = 0; col < mtFirst.ncols(); col++) {
							(*output)(row, col) = 0.0f;
						}
					}
				}

				for (; it != input_map.end(); ++it) {
					const Matrix<float> &mtFltInput = object_cast<Matrix<float> >(it->second);
					if ((mtFltInput.nrows() != mtFirst.nrows()) || (mtFltInput.ncols() != mtFirst.ncols())) {
						break;
					}

					if (method == METHOD_MIN_KEY) {
						if (minid > it->first) {
							minid = it->first;
							for (int row = 0; row < mtFltInput.nrows(); row++) {
								for (int col = 0; col < mtFltInput.ncols(); col++) {
									(*output)(row, col) = mtFltInput(row, col);
								}
							}
						}
					}
					else if (method == METHOD_MAX_KEY) {
						if (maxid < it->first) {
							minid = it->first;
							for (int row = 0; row < mtFltInput.nrows(); row++) {
								for (int col = 0; col < mtFltInput.ncols(); col++) {
									(*output)(row, col) = mtFltInput(row, col);
								}
							}
						}
					}
					else { // method == METHOD_AVERAGE, METHOD_SUMMATION
						for (int row = 0; row < mtFltInput.nrows(); row++) {
							for (int col = 0; col < mtFltInput.ncols(); col++) {
								(*output)(row, col) += mtFltInput(row, col);
							}
						}
					}
				}

				if (it != input_map.end()) {
					cerr << "******* MapToMatrix: Input size of ObjectRef is not constant." << endl;
					throw(new NodeException(NULL, "Exception caught in calculate[MapToMatrix]", __FILE__, __LINE__));
				}
				if (method == METHOD_AVERAGE) {
					for (int row = 0; row < mtFirst.nrows(); row++) {
						for (int col = 0; col < mtFirst.ncols(); col++) {
							(*output)(row, col) /= input_map.size();
						}
					}
				}
			}
			else if (typeid(*(it->second)) == typeid(Matrix<complex<float> >)) {
				const Matrix<complex<float> > &mtFirst = object_cast<Matrix<complex<float> > >(it->second);
				RCPtr<Matrix<complex<float> > > output(new Matrix<complex<float> >(mtFirst.nrows(), mtFirst.ncols()));
				out[count] = output;

				if (method == METHOD_AVERAGE || method == METHOD_SUMMATION) { // initialize zero
					for (int row = 0; row < mtFirst.nrows(); row++) {
						for (int col = 0; col < mtFirst.ncols(); col++) {
							(*output)(row, col) = complex<float>(0.0f, 0.0f);
						}
					}
				}

				for (; it != input_map.end(); ++it) {
					const Matrix<complex<float> > &mtCmpInput = object_cast<Matrix<complex<float> > >(it->second);
					if ((mtCmpInput.nrows() != mtFirst.nrows()) || (mtCmpInput.ncols() != mtFirst.ncols())) {
						break;
					}

					if (method == METHOD_MIN_KEY) {
						if (minid > it->first) {
							minid = it->first;
							for (int row = 0; row < mtCmpInput.nrows(); row++) {
								for (int col = 0; col < mtCmpInput.ncols(); col++) {
									(*output)(row, col) = mtCmpInput(row, col);
								}
							}
						}
					}
					else if (method == METHOD_MAX_KEY) {
						if (maxid < it->first) {
							minid = it->first;
							for (int row = 0; row < mtCmpInput.nrows(); row++) {
								for (int col = 0; col < mtCmpInput.ncols(); col++) {
									(*output)(row, col) = mtCmpInput(row, col);
								}
							}
						}
					}
					else { // method == METHOD_AVERAGE, METHOD_SUMMATION
						for (int row = 0; row < mtCmpInput.nrows(); row++) {
							for (int col = 0; col < mtCmpInput.ncols(); col++) {
								(*output)(row, col) += mtCmpInput(row, col);
							}
						}
					}
				}

				if (it != input_map.end()) {
					cerr << "******* MapToMatrix: Input size of ObjectRef is not constant." << endl;
					throw(new NodeException(NULL, "Exception caught in calculate[MapToMatrix]", __FILE__, __LINE__));
				}
				if (method == METHOD_AVERAGE) {
					for (int row = 0; row < mtFirst.nrows(); row++) {
						for (int col = 0; col < mtFirst.ncols(); col++) {
							(*output)(row, col) /= input_map.size();
						}
					}
				}
			}
			else {
				cerr << "******* MapToMatrix: Input type of ObjectRef is unidentified." << endl;
				throw(new NodeException(NULL, "Exception caught in calculate[MapToMatrix]", __FILE__, __LINE__));
			}
		}

		if (debug == true) cout << "******* MapToMatrix debug point-Finish" << endl;
	}

	IN_ORDER_NODE_SPEEDUP(MapToMatrix);
};
