/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <iostream>
#include <BufferedNode.h>
#include <Buffer.h>
#include <Matrix.h>
#include <Vector.h>

using namespace FD;
using namespace std;

class MatrixToVector;

DECLARE_NODE(MatrixToVector)
/*Node
 *
 * @name MatrixToVector
 * @category HARK:MISC
 * @description Convert Matrix<float> to Vector<float> and Matrix<complex<float> > to Vector<complex<float> >.
 *
 * @input_name INPUT
 * @input_type any
 * @input_description Matrix<float> or Matrix<complex<float> >.
 *
 * @output_name OUTPUT
 * @output_type any
 * @output_description Vector<float> or Vector<complex<float> >.
 *
 * @parameter_name METHOD
 * @parameter_type string
 * @parameter_list reshape:accumulate
 * @parameter_value reshape
 * @parameter_description Select method to transform matrix to vector. [default: reshape]
 *
 * @parameter_name RESHAPE_ORDER
 * @parameter_type string
 * @parameter_list row:column
 * @parameter_value row
 * @parameter_valid METHOD=reshape
 * @parameter_description Method of reshaping the input Matrix. If row is selected, the input Matrix is reshaped to row Vectors. If column is selected, the input Matrix is reshaped to column Vectors. [default: row]
 *
 * @parameter_name ACCUMULATE_METHOD
 * @parameter_type string
 * @parameter_list row_sum:col_sum:row_avg:col_avg
 * @parameter_value row_sum
 * @parameter_valid METHOD=accumulate
 * @parameter_description Method of accumulating the input Matrix, for example, row_sum converts NxM Matrix to size of N Vector by summation. [default: row_sum]
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Show debug points. [default: false]
 *
 END*/

static const int METHOD_RESHAPE = 0;
static const int METHOD_ACCUMULATE = 1;

static const int ROW_MAJOR_ORDER = 0;
static const int COLUMN_MAJOR_ORDER = 1;

static const int ROW_SUM_PATTERN = 0;
static const int COL_SUM_PATTERN = 1;
static const int ROW_AVG_PATTERN = 2;
static const int COL_AVG_PATTERN = 3;

class MatrixToVector : public BufferedNode {
	int inputID;
	int outputID;
	int method;
    int order;
	int pattern;
	bool debug;
    
public:
	MatrixToVector(string nodeName, ParameterSet params)
		: BufferedNode(nodeName, params) {
		inputID = addInput("INPUT");
		outputID = addOutput("OUTPUT");
		string method_name = object_cast<String>(parameters.get("METHOD"));
		if (method_name == "reshape") {
			method = METHOD_RESHAPE;
			string order_name = object_cast<String>(parameters.get("RESHAPE_ORDER"));
			if (order_name == "row")         order = ROW_MAJOR_ORDER;
			else if (order_name == "column") order = COLUMN_MAJOR_ORDER;
			else {
				cerr << "******* MatrixToVector: Input error" << endl;
				throw new NodeException(NULL, "MatrixToVector: RESHAPE_ORDER should be \"row\" or \"column\".", __FILE__, __LINE__);
			}
		}
		else if (method_name == "accumulate") {
			method = METHOD_ACCUMULATE;
			string calc_pattern = object_cast<String>(parameters.get("ACCUMULATE_METHOD"));
			if (calc_pattern == "row_sum")      pattern = ROW_SUM_PATTERN;
			else if(calc_pattern == "col_sum")  pattern = COL_SUM_PATTERN;
			else if (calc_pattern == "row_avg") pattern = ROW_AVG_PATTERN;
			else if (calc_pattern == "col_avg") pattern = COL_AVG_PATTERN;
			else {
				cerr << "******* MatrixToVector: Input error" << endl;
				throw new NodeException(NULL, "MatrixToVector: ACCUMULATE_METHOD should be \"row_sum\" or \"col_sum\" or \"row_avg\" or \"col_avg\".", __FILE__, __LINE__);
			}
		}
		else {
			cerr << "******* MatrixToVector: Input error" << endl;
			throw new NodeException(NULL, "MatrixToVector: METHOD should be \"reshape\" or \"accumulate\".", __FILE__, __LINE__);
		}
		debug = dereference_cast<bool>(parameters.get("DEBUG"));
	}

	void calculate(int output_id, int count, Buffer &out) {
		ObjectRef input = getInput(inputID, count);
		if (debug == true) {
			cout << "******* MatrixToVector debug point-Input check (count=" << count << ")" << endl;
            cout << "Input ObjectRef's type is: ";
			if (typeid(*input) == typeid(Matrix<float>))                cout << "Matrix<float>" << endl;
			else if (typeid(*input) == typeid(Matrix<complex<float> >)) cout << "Matrix<complex<float> >" << endl;
			else                                                        cout << "unidentified." << endl;
		}

		if (typeid(*input) == typeid(Matrix<float>)) {
			const Matrix<float> &mtFltInput = object_cast<Matrix<float> >(input);

			if (method == METHOD_ACCUMULATE) {
				if (pattern == ROW_SUM_PATTERN || pattern == ROW_AVG_PATTERN) {
					/*ObjectRef orOutputTemp = ObjectRef(new Vector<float>(mtFltInput.nrows()));
					Vector<float> &vcFltOutput = object_cast<Vector<float> >(orOutputTemp);
					out[count] = orOutputTemp;*/
					/*RCPtr<Vector<float> > vcFltOutput(new Vector<float>(mtFltInput.nrows()));*/
					RCPtr<Vector<float> > vcFltOutput(new Vector<float>(mtFltInput.nrows(), 0.0f));
					out[count] = vcFltOutput;

					for (int row = 0; row < mtFltInput.nrows(); row++) {
						/*vcFltOutput[row] = 0.0f;*/
						/*(*vcFltOutput)[row] = 0.0f;*/
						for (int col = 0; col < mtFltInput.ncols(); col++) {
							/*vcFltOutput[row] += mtFltInput(row, col);*/
							(*vcFltOutput)[row] += mtFltInput(row, col);
						}
					}
					if (pattern == ROW_AVG_PATTERN) {
						for (int row = 0; row < mtFltInput.nrows(); row++) {
							/*vcFltOutput[row] /= mtFltInput.ncols();*/
							(*vcFltOutput)[row] /= mtFltInput.ncols();
						}
					}
				}
				else { // if (pattern == COL_SUM_PATTERN || pattern == COL_AVG_PATTERN) {
					/*ObjectRef orOutputTemp = ObjectRef(new Vector<float>(mtFltInput.ncols()));
					Vector<float> &vcFltOutput = object_cast<Vector<float> >(orOutputTemp);
					out[count] = orOutputTemp;*/
					/*RCPtr<Vector<float> > vcFltOutput(new Vector<float>(mtFltInput.ncols()));*/
					RCPtr<Vector<float> > vcFltOutput(new Vector<float>(mtFltInput.ncols(), 0.0f));    
					out[count] = vcFltOutput;

					for (int col = 0; col < mtFltInput.ncols(); col++) {
						/*vcFltOutput[col] = 0.0f;*/
						/*(*vcFltOutput)[col] = 0.0f;*/
						for (int row = 0; row < mtFltInput.nrows(); row++) {
							/*vcFltOutput[col] += mtFltInput(row, col);*/
							(*vcFltOutput)[col] += mtFltInput(row, col);
						}
					}
					if (pattern == COL_AVG_PATTERN) {
						for (int col = 0; col < mtFltInput.ncols(); col++) {
							/*vcFltOutput[col] /= mtFltInput.nrows();*/
							(*vcFltOutput)[col] /= mtFltInput.nrows();
						}
					}
				}
			}
			else { // method == METHOD_RESHAPE
				/*ObjectRef orOutputTemp = ObjectRef(new Vector<float>(mtFltInput.nrows() * mtFltInput.ncols()));
				Vector<float> &vcFltOutput = object_cast<Vector<float> >(orOutputTemp);
				out[count] = orOutputTemp;*/
				RCPtr<Vector<float> > vcFltOutput(new Vector<float>(mtFltInput.nrows() * mtFltInput.ncols()));    
				out[count] = vcFltOutput;

				if (order == ROW_MAJOR_ORDER) {
					for (int row = 0; row < mtFltInput.nrows(); row++) {
						for (int col = 0; col < mtFltInput.ncols(); col++) {
							/*vcFltOutput[row * mtFltInput.ncols() + col] = mtFltInput(row, col);*/
							(*vcFltOutput)[row * mtFltInput.ncols() + col] = mtFltInput(row, col);
						}
					}
				}
				else { // order == COLUMN_MAJOR_ORDER
					for (int col = 0; col < mtFltInput.ncols(); col++) {
						for (int row = 0; row < mtFltInput.nrows(); row++) {
							/*vcFltOutput[row + col * mtFltInput.nrows()] = mtFltInput(row, col);*/
							(*vcFltOutput)[row + col * mtFltInput.nrows()] = mtFltInput(row, col);
						}
					}
				}
			}
		}
		else if (typeid(*input) == typeid(Matrix<complex<float> >)) {
			const Matrix<complex<float> > &mtCmpInput = object_cast<Matrix<complex<float> > >(input);

			if (method == METHOD_ACCUMULATE) {
				if (pattern == ROW_SUM_PATTERN || pattern == ROW_AVG_PATTERN) {
					/*ObjectRef orOutputTemp = ObjectRef(new Vector<complex<float> >(mtCmpInput.nrows()));
					Vector<complex<float> > &vcCmpOutput = object_cast<Vector<complex<float> > >(orOutputTemp);
					out[count] = orOutputTemp;*/
					/*RCPtr<Vector<complex<float> > > vcCmpOutput(new Vector<complex<float> >(mtCmpInput.nrows()));*/
					RCPtr<Vector<complex<float> > > vcCmpOutput(new Vector<complex<float> >(mtCmpInput.nrows(), complex<float>(0.0f, 0.0f))); 
					out[count] = vcCmpOutput;

					for (int row = 0; row < mtCmpInput.nrows(); row++) {
						/*vcCmpOutput[row] = complex<float>(0.0f, 0.0f);*/
						/*(*vcCmpOutput)[row] = complex<float>(0.0f, 0.0f);*/
						for (int col = 0; col < mtCmpInput.ncols(); col++) {
							/*vcCmpOutput[row] += mtCmpInput(row, col);*/
							(*vcCmpOutput)[row] += mtCmpInput(row, col);
						}
					}
					if (pattern == ROW_AVG_PATTERN) {
						for (int row = 0; row < mtCmpInput.nrows(); row++) {
							/*vcCmpOutput[row] /= mtCmpInput.ncols();*/
							(*vcCmpOutput)[row] /= mtCmpInput.ncols();
						}
					}
				}
				else { // if (pattern == COL_SUM_PATTERN || pattern == COL_AVG_PATTERN) {
					/*ObjectRef orOutputTemp = ObjectRef(new Vector<complex<float> >(mtCmpInput.ncols()));
					Vector<complex<float> > &vcCmpOutput = object_cast<Vector<complex<float> > >(orOutputTemp);
					out[count] = orOutputTemp;*/
					/*RCPtr<Vector<complex<float> > > vcCmpOutput(new Vector<complex<float> >(mtCmpInput.ncols()));*/
					RCPtr<Vector<complex<float> > > vcCmpOutput(new Vector<complex<float> >(mtCmpInput.ncols(), complex<float>(0.0f, 0.0f)));
					out[count] = vcCmpOutput;

					for (int col = 0; col < mtCmpInput.ncols(); col++) {
						/*vcCmpOutput[col] = complex<float>(0.0f, 0.0f);*/
						/*(*vcCmpOutput)[col] = complex<float>(0.0f, 0.0f);*/
						for (int row = 0; row < mtCmpInput.nrows(); row++) {
							/*vcCmpOutput[col] += mtCmpInput(row, col);*/
							(*vcCmpOutput)[col] += mtCmpInput(row, col);
						}
					}
					if (pattern == COL_AVG_PATTERN) {
						for (int col = 0; col < mtCmpInput.ncols(); col++) {
							/*cCmpOutput[col] /= mtCmpInput.nrows();*/
							(*vcCmpOutput)[col] /= mtCmpInput.nrows();
						}
					}
				}
			}
			else { // method == METHOD_RESHAPE
				/*ObjectRef orOutputTemp = ObjectRef(new Vector<complex<float> >(mtCmpInput.nrows() * mtCmpInput.ncols()));
				Vector<complex<float> > &vcCmpOutput = object_cast<Vector<complex<float> > >(orOutputTemp);
				out[count] = orOutputTemp;*/
				RCPtr<Vector<complex<float> > > vcCmpOutput(new Vector<complex<float> >(mtCmpInput.nrows() * mtCmpInput.ncols()));    
				out[count] = vcCmpOutput;

				if (order == ROW_MAJOR_ORDER) {
					for (int row = 0; row < mtCmpInput.nrows(); row++) {
						for (int col = 0; col < mtCmpInput.ncols(); col++) {
							/*vcCmpOutput[row * mtCmpInput.ncols() + col] = mtCmpInput(row, col);*/
							(*vcCmpOutput)[row * mtCmpInput.ncols() + col] = mtCmpInput(row, col);
						}
					}
				}
				else { // order == COLUMN_MAJOR_ORDER
					for (int col = 0; col < mtCmpInput.ncols(); col++) {
						for (int row = 0; row < mtCmpInput.nrows(); row++) {
							/*vcCmpOutput[row + col * mtCmpInput.nrows()] = mtCmpInput(row, col);*/
							(*vcCmpOutput)[row + col * mtCmpInput.nrows()] = mtCmpInput(row, col);
						}
					}
				}
			}
		}
		else {
			cerr << "******* MatrixToVector: Input type is unidentified." << endl;
			throw(new NodeException(NULL, "Exception caught in calculate[MatrixToVector]", __FILE__, __LINE__));
		}

        if (debug == true) cout << "******* MatrixToVector debug point-Finish" << endl;
    }

    NO_ORDER_NODE_SPEEDUP(MatrixToVector);
};
