/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <fstream>
#include "BufferedNode.h"
#include "Exception.h"
#include "Map.h"
#include "Vector.h"
#include "Source.h"
#include <set>

#define USE_TRIPLET_CRITERION 1

using namespace std;

namespace HARK {

class InilikeFileParser {
public:
    typedef std::vector<std::pair<std::string, std::string> > section_type;
    typedef std::map<std::string, section_type> content_type;

public:
    InilikeFileParser() :
            _path() {
    }
    InilikeFileParser(const string &path) :
            _path(path) {
    }
    ~InilikeFileParser() {
    }

    const section_type &get_section(const string &name) const {
        content_type::const_iterator it;
        if ((it = this->_content.find(name)) == this->_content.end()) {
            throw new NodeException(NULL, "Section '" + name + "' is not defined", __FILE__, __LINE__);
        } else {
            return it->second;
        }
    }

    const content_type &load(const string &path) {
        this->_path = path;
        return this->parse();
    }

    static string trim(const string &s, const char *ws) {
        std::string::size_type si = s.find_first_not_of(ws);
        std::string::size_type ei = s.find_last_not_of(ws);
        if (si != std::string::npos && ei != std::string::npos) {
            return s.substr(si, ei - si + 1);
        } else {
            return s;
        }
    }

    const content_type &parse() {
        std::ifstream is(this->_path.c_str());
        if (!is) {
            throw new NodeException(NULL, "Failed to open config file", __FILE__, __LINE__);
        }

        const char ws[] = " \t\r";

        size_t nline = 0;
        std::string line, section;
        while (std::getline(is, line)) {
            nline++;
            std::string::size_type index;
            if ((index = line.find('#')) != std::string::npos) {
                line = line.substr(0, index);
            }
            if ((index = line.find_first_not_of(ws)) != std::string::npos) {
                line = line.substr(index);
            }
            if ((index = line.find_last_not_of(ws)) != std::string::npos) {
                line = line.substr(0, index + 1);
            }

            if (line.empty()) {
                continue;
            }

            std::string::size_type len = line.size();
            if (line[0] == '[') {
                if (line[len - 1] == ']') {
                    section = trim(line.substr(1, len - 2), ws);
                    if (section.empty()) {
                        throw new NodeException(NULL, "Empty section name", __FILE__, __LINE__);
                    }
                    cout << "Debug: section='" << section << "'" << endl;
                } else {
                    throw new NodeException(NULL, "Unexpected section header format '" + line + "'", __FILE__,
                            __LINE__);
                }

            } else {
                if ((index = line.find('=')) == std::string::npos) {
                    throw new NodeException(NULL, "Unexpected line format '" + line + "'", __FILE__, __LINE__);
                }

                std::string key = trim(line.substr(0, index), ws);
                std::string val = trim(line.substr(index + 1), ws);
                cout << "Debug: key='" << key << "', val='" << val << "'" << endl;
                if (section.empty()) {
                    throw new NodeException(NULL, "Unexpected line before section '" + line + "'", __FILE__, __LINE__);
                }

                section_type &section_map = this->_content[section];
                section_map.push_back(std::pair<std::string, std::string>(key, val));
            }
        }

        if (false) { //debug
            content_type::const_iterator cit;
            section_type::const_iterator sit;
            for (cit = this->_content.begin(); cit != this->_content.end(); ++cit) {
                const section_type &section = cit->second;
                for (sit = section.begin(); sit != section.end(); ++sit) {
                    cout << "MultiLocalizationIntegrator: " << cit->first << "." << sit->first << " = " << sit->second
                            << std::endl;
                }
            }
        }
    }

private:
    std::string _path;
    content_type _content;
};

class MultiLocalizationIntegrator;

DECLARE_NODE (MultiLocalizationIntegrator)
/*Node
 *
 * @name MultiLocalizationIntegrator
 * @category HARK:Localization
 * @description Integrate output of multiple localization processes.
 *
 * @output_name SRC_OUTPUT
 * @output_type Vector<ObjectRef>
 * @output_description
 *
 * @output_name SPEC_OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description
 *
 * @parameter_name CONFIG
 * @parameter_type string
 * @parameter_description Path to configuration file
 *
 END*/

class MultiLocalizationIntegrator: public BufferedNode {
    struct Criterion {
        bool includes;
        float start, end;
#if defined(USE_TRIPLET_CRITERION)
        int sourceID;
#endif
    };

    struct InputInfo {
        int spec;
        int source;
        std::string name;
        std::vector<Criterion> criteria;

        InputInfo() :
                spec(-1), source(-1) {
        }
    };

    // flowdesigner stuff
    string config;
    int outputSpecID;
    int outputSourceID;

    // config
    int sourceCount;
    vector<InputInfo> inputInfo;

    // workset
    int sourceIdCount;
    map<pair<int, int>, int> sourceIdMap;

public:
    MultiLocalizationIntegrator(string nodeName, ParameterSet params) :
            BufferedNode(nodeName, params) {
        this->sourceCount = 1;
        this->sourceIdCount = 0;
        this->outputSpecID = this->addOutput("SPEC_OUTPUT");
        this->outputSourceID = this->addOutput("SRC_OUTPUT");
        this->config = object_cast < String > (params.get("CONFIG"));
    }

    virtual int translateInput(const std::string &inputName) {
        for (unsigned int i = 0; i < inputs.size(); i++) {
            if (this->inputs[i].name == inputName) {
                return i;
            }
        }

        char *ep = NULL;
        int id = 0;
        if (inputName.find("SPEC") == 0 & (id = strtoul(inputName.c_str() + 4, &ep, 10), ep != NULL) && *ep == '\0'
                && id > 0) {
            if (this->inputInfo.size() < id) {
                this->inputInfo.resize(id);
            }
            return this->inputInfo[id - 1].spec = this->addInput(inputName);

        } else if (inputName.find("SRC_INFO") == 0 && (id = strtoul(inputName.c_str() + 8, &ep, 10), ep != NULL)
                && *ep == '\0' && id > 0) {
            if (this->inputInfo.size() < id) {
                this->inputInfo.resize(id);
            }
            return this->inputInfo[id - 1].source = this->addInput(inputName);

        } else {
            throw new NodeException(NULL, "Unexpected INPUT name " + inputName + ".", __FILE__, __LINE__);
        }
    }

    Criterion parse_criterion(const std::string &str, bool includes, bool triplet) {
        Criterion c;

        std::stringstream ss(str);
        std::string s;
        if (triplet) {
            if (std::getline(ss, s, ',')) {
                std::stringstream(s) >> c.sourceID;
            } else {
                throw new NodeException(NULL, "Failed to parse criterion '" + str + "'", __FILE__, __LINE__);
            }
        }
        if (std::getline(ss, s, ',')) {
            std::stringstream(s) >> c.start;
            c.start -= 360 * floor((c.start + 180) / 360);
        } else {
            throw new NodeException(NULL, "Failed to parse criterion '" + str + "'", __FILE__, __LINE__);
        }
        if (std::getline(ss, s, ',')) {
            std::stringstream(s) >> c.end;
            c.end -= 360 * floor((c.end + 180) / 360);
        } else {
            throw new NodeException(NULL, "Failed to parse criterion '" + str + "'", __FILE__, __LINE__);
        }

        c.includes = triplet || includes;
        return c;
    }

    void initialize() {
        for (size_t i = 0; i < this->inputInfo.size(); i++) {
            if (this->inputInfo[i].spec < 0) {
                std::stringstream ss;
                ss << "Input terminal with name SPEC" << (i + 1) << " not defined.";
                throw new NodeException(NULL, ss.str(), __FILE__, __LINE__);
            }
            if (this->inputInfo[i].source < 0) {
                std::stringstream ss;
                ss << "Input terminal with name SRC_INFO" << (i + 1) << " not defined.";
                throw new NodeException(NULL, ss.str(), __FILE__, __LINE__);
            }
        }

        // 角度設定情報等初期化
        InilikeFileParser ifp;
        ifp.load(this->config);

        const InilikeFileParser::section_type &section = ifp.get_section("main");

        InilikeFileParser::section_type::const_iterator sit;
        size_t index = 0;
        for (sit = section.begin(); sit != section.end(); ++sit) {
            if (sit->first == "sources") {
                std::stringstream(sit->second) >> this->sourceCount;
            } else {
                // unexpected key.  ignored
            }
        }
        for (sit = section.begin(); sit != section.end(); ++sit) {
            if (sit->first == "input") {
                InputInfo &input = this->inputInfo[index++];

                const std::string inputName = sit->second;
                input.name = inputName;

                const InilikeFileParser::section_type &inputSection = ifp.get_section(inputName);
                InilikeFileParser::section_type::const_iterator vit;
                for (vit = inputSection.begin(); vit != inputSection.end(); ++vit) {
                    const std::string &vname = vit->first;
                    const std::string &vval = vit->second;

                    if (vname == "include") {
                        input.criteria.push_back(parse_criterion(vval, true, false));
#if defined(USE_TRIPLET_CRITERION)
                    } else if (vname == "source") {
                        Criterion c = parse_criterion(vval, true, true);
                        if (c.sourceID >= this->sourceCount) {
                            std::stringstream ss;
                            ss << "Too large source-id value '" << c.sourceID << "'";
                            throw new NodeException(NULL, ss.str(), __FILE__, __LINE__);
                        } else {
                            input.criteria.push_back(c);
                        }
#else
                    } else if(vname == "exclude") {
                        input.criteria.push_back(parse_criterion(vval, false, false));
#endif
                    } else if (vname == "x") {
                        // ignored
                    } else if (vname == "y") {
                        // ignored
                    } else {
                        throw new NodeException(NULL, "Unexpected value '" + vname + "'", __FILE__, __LINE__);
                    }
                }

            } else {
                // unexpected key.  ignored

            }
        }

        { //debug
            for (size_t i = 0; i < this->inputInfo.size(); i++) {
                const InputInfo &ii = this->inputInfo[i];

                cout << "MultiLocalizationIntegrator: Input[" << i << "]" << endl;
                cout << "  name:          " << ii.name << endl;
                cout << "  sourceInputID: " << ii.source << endl;
                cout << "  specInputID:   " << ii.spec << endl;
                cout << "  criteria:" << endl;

                for (std::vector<Criterion>::const_iterator it = ii.criteria.begin(); it != ii.criteria.end(); ++it) {
                    cout << "    " << (it->includes ? "include" : "exclude") << ": (" << it->start << ", " << it->end
                            << "), source=" << it->sourceID << endl;
                }

//		for(int t = -10; t <= 10; t++){
//		    float a = t * 18.0;
//		    bool  x = check_if_acceptable(ii, cos(a / 180 * M_PI), sin(a / 180 * M_PI));
//		    cout << "    test: " << a << ": " << x << endl;
//		}
            }
        }

        // lookAhead 等初期化
        int ids[] = { this->outputSpecID, this->outputSourceID };
        this->initializeLookAheadBack(ids, sizeof(ids) / sizeof(*ids));

        this->BufferedNode::initialize();
    }

    void initializeLookAheadBack(const int ids[], size_t count) {
        int max_lookAhead = 0, max_lookBack = 0;
        for (size_t i = 0; i < count; i++) {
            if (max_lookAhead < this->outputs[ids[i]].lookAhead) {
                max_lookAhead = this->outputs[ids[i]].lookAhead;
            }
            if (max_lookBack < this->outputs[ids[i]].lookBack) {
                max_lookBack = this->outputs[ids[i]].lookBack;
            }
        }
        for (size_t i = 0; i < count; i++) {
            this->outputs[ids[i]].lookAhead = max_lookAhead + 1;
            this->outputs[ids[i]].lookBack = max_lookBack + 1;
        }
    }

    bool check_if_acceptable(const InputInfo &inputInfo, const Source &src, int *sourceID) {
        return check_if_acceptable(inputInfo, src.x[0], src.x[1], sourceID);
    }

    bool check_if_acceptable(const InputInfo &inputInfo, float x, float y, int *sourceID) {
        float a = atan2(y, x) * 180 / M_PI;

        for (std::vector<Criterion>::const_iterator it = inputInfo.criteria.begin(); it != inputInfo.criteria.end();
                ++it) {
            const Criterion &c = *it;
            if (((/* c.start <  c.end && */(c.start <= a && a <= c.end))
                    || (c.start >= c.end && (a >= c.start || a <= c.end)))) {
                if (sourceID) {
                    *sourceID = c.sourceID;
                }
                return c.includes;
            }
        }

#if defined(USE_TRIPLET_CRITERION)
        return false;
#else
        return true;
#endif	
    }

    void calculate(int output_id, int count, Buffer &out) {
        //cout << "MultiLocalizationIntegrator: count=" << count << endl;

        /* 出力へのデータ領域の割り当て */
        Buffer &outputSpecBuffer = *this->outputs[this->outputSpecID].buffer;
        Buffer &outputSrcBuffer = *this->outputs[this->outputSourceID].buffer;

        RCPtr < Map<int, ObjectRef> > pspectrumMap(new Map<int, ObjectRef>());
        RCPtr < Vector<ObjectRef> > psources(Vector < ObjectRef > ::alloc(0));
        outputSpecBuffer[count] = pspectrumMap;
        outputSrcBuffer[count] = psources;

        Map<int, ObjectRef> &spectrumMap = *pspectrumMap;
        Vector < ObjectRef > &sources = *psources;

        /* 出現中の source-id の集合を保持 */
        set<pair<int, int> > existingSourceSet;
        for (std::map<pair<int, int>, int>::const_iterator it = this->sourceIdMap.begin();
                it != this->sourceIdMap.end(); ++it) {
            existingSourceSet.insert(it->first);
        }

        /* 入力から音源情報・スペクトログラムを取得 */
        for (size_t i = 0; i < this->inputInfo.size(); i++) {
            int inputSpecID = this->inputInfo[i].spec;
            int inputSrcID = this->inputInfo[i].source;

            RCPtr < Vector<ObjectRef> > psrcs = this->getInput(inputSrcID, count);

            Map<int, ObjectRef> &specMap = object_cast < Map<int, ObjectRef> > (this->getInput(inputSpecID, count));
            Vector < ObjectRef > &srcs = *psrcs;

            for (size_t j = 0; j < srcs.size(); j++) {
                Source &src = object_cast < Source > (srcs[j]);
                Vector < complex<float> > &ispec = object_cast < Vector<complex<float> > > (specMap[src.id]);

                int outputSourceId = -1;
                pair<int, int> idpair(i, src.id);
                std::map<pair<int, int>, int>::iterator it;
                if ((it = this->sourceIdMap.find(idpair)) == this->sourceIdMap.end()) {
                    /* 新しい音源 */
                    /* 音源方向が条件を満たしているか？ */
                    int sourceID = 0;
                    if (this->check_if_acceptable(this->inputInfo[i], src, &sourceID)) {
                        /* 方向チェック OK */
                        outputSourceId = ++this->sourceIdCount * this->sourceCount + sourceID;
                        cout << "MultiLocalizationIntegrator: Source (" << i << ", " << src.id << "), x=(" << src.x[0]
                                << ", " << src.x[1] << ", " << src.x[2] << ", "
                                << (atan2(src.x[1], src.x[0]) * 180 / M_PI) << "), is accepted" << endl;
                    } else {
                        /* 方向チェック NG */
                        cout << "MultiLocalizationIntegrator: Source (" << i << ", " << src.id << "), x=(" << src.x[0]
                                << ", " << src.x[1] << ", " << src.x[2] << ", "
                                << (atan2(src.x[1], src.x[0]) * 180 / M_PI) << "), is discarded" << endl;
                        outputSourceId = -1;
                    }
                    this->sourceIdMap[idpair] = outputSourceId;
                    cout << "MultiLocalizationIntegrator: New source: (" << idpair.first << ", " << idpair.second
                            << "), out=" << outputSourceId << endl;

                } else {
                    /* 継続音源 */
                    outputSourceId = it->second;
                    existingSourceSet.erase(idpair);
                }

                if (outputSourceId >= 0) {
                    /* 代替音源情報 */
                    RCPtr < Source > outputSource(new Source(src));
                    outputSource->id = outputSourceId;

                    /* 出力 */
                    ObjectRef ospecp = ObjectRef(new Vector<complex<float> >());
                    Vector < complex<float> > &ospec = object_cast < Vector<complex<float> > > (ospecp);
                    ospec = ispec;

                    sources.push_back(outputSource);
                    spectrumMap[outputSourceId] = ospecp;
                }
            }
        }

        /* 消失した source-id を削除 */
        for (std::set<pair<int, int> >::const_iterator it = existingSourceSet.begin(); it != existingSourceSet.end();
                ++it) {
            cout << "MultiLocalizationIntegrator: Lost source: (" << it->first << ", " << it->second << "), out="
                    << this->sourceIdMap[*it] << endl;
            this->sourceIdMap.erase(*it);
        }
    }

    NO_ORDER_NODE_SPEEDUP (MultiLocalizationIntegrator);
};

} //namespace FD
