/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <iostream>
#include <fstream>
#include <iomanip>
#include <limits>

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Map.h>

#include "format.h"

using namespace FD;

class SaveMapFrames;

DECLARE_NODE(SaveMapFrames);
/*Node
 *
 * @name SaveMapFrames
 * @category HARK:MISC
 * @description Save map data of frames to file.
 *
 * @input_name INPUT
 * @input_type any
 * @input_description Data Stream
 *
 * @output_name OUTPUT
 * @output_type any
 * @output_description The same as input.
 *
 * @parameter_name FILENAME
 * @parameter_type string
 * @parameter_description The file name for saving map data.
 *
 * @parameter_name OUTPUTTYPE
 * @parameter_type string
 * @parameter_list TEXT:RAW
 * @parameter_value TEXT
 * @parameter_description Output file format type [default: TEXT]
 *
 * @parameter_name TEXTFORMAT
 * @parameter_type string
 * @parameter_valid OUTPUTTYPE=TEXT
 * @parameter_list FIXED:SCIENTIFIC
 * @parameter_value FIXED
 * @parameter_description Output file format type [default: FIXED]
 *
 END*/

template <class T>
inline int printLine(const Vector<T> &v, std::ofstream &out, bool is_fixed=true) {
    if(std::is_floating_point<T>::value){
        if(is_fixed){
            out << std::fixed << std::setprecision(std::numeric_limits<T>::max_digits10);
        }
        else{
            out << std::scientific << std::setprecision(std::numeric_limits<T>::max_digits10);
        }
    }
    else if(std::numeric_limits<T>::is_integer){
        out << std::dec;
    }
    else{
        out << "Vector's elements must be integer or floating point types. It is undefined behavior." << std::endl;
        return -1;
    }
    for(size_t i=0; i < v.size(); i++){
        out << v[i] << " ";
    }
    out << std::endl;
    return 0;
}

template <class T>
inline int printLine(const Vector<std::complex<T> > &v, std::ofstream &out, bool is_fixed=true) {
    if(!std::is_floating_point<T>::value){
        out << "Complex numbers are not floating point types. It is undefined behavior." << std::endl;
        return -1;
    }
    if(is_fixed){
        out << std::fixed << std::setprecision(std::numeric_limits<T>::max_digits10);
    }
    else{
        out << std::scientific << std::setprecision(std::numeric_limits<T>::max_digits10);
    }
    for(size_t i=0; i < v.size(); i++){
        out << v[i].real() << " " << v[i].imag() << " ";
    }
    out << std::endl;
    return 0;
}

class SaveMapFrames : public BufferedNode {
    int inputID;
    int outputID;

    typedef enum {TEXT, RAW} OutputMode;
    typedef enum {FIXED, SCIENTIFIC} TextMode;

    OutputMode mode;
    TextMode tmode;
    std::string outputType;
    std::string textFormat;
    std::string filename;
    std::string formatted_filename;
    bool will_format_filename;
    std::ofstream fout_in;
    std::ofstream fout;
    int dimension;
    std::string containertype;
    std::vector<std::string> list;

    public:
    SaveMapFrames(std::string nodeName, ParameterSet params) : BufferedNode(nodeName, params), mode(TEXT), dimension(1), containertype("Vector")
    {
        inputID = addInput("INPUT");
        outputID = addOutput("OUTPUT");
        outputType = object_cast<std::string>(parameters.get("OUTPUTTYPE"));
        textFormat = object_cast<std::string>(parameters.get("TEXTFORMAT"));
        filename = object_cast<std::string>(parameters.get("FILENAME"));

        if (filename.empty()) {
            throw new NodeException(NULL, std::string("FILENAME is empty."), __FILE__, __LINE__);
        }

        formatted_filename.clear();
        will_format_filename = false;

        
        if(filename.find("{datatype", 0) != std::string::npos ||
           filename.find("{colsize", 0) != std::string::npos ||
           filename.find("{dim", 0) != std::string::npos ||
           filename.find("{containertype", 0) != std::string::npos) {
            will_format_filename = true;
        }
        
        if (outputType.compare("RAW") == 0) {
            mode = RAW;
        } else {
            mode = TEXT;
        }
        if (textFormat.compare("FIXED") == 0) {
            tmode = FIXED;
        } else {
            tmode = SCIENTIFIC;
        }
    }

    ~SaveMapFrames()
    {
        fout.close();
    }

    virtual int translateInput(const std::string &inputName) {
        int ret = -1;
        if (inputName == "INPUT") {
            for (unsigned int i = 0; i < inputs.size(); i++) {
                if (inputs[i].name == inputName) {
                    ret = i; break;
                }
            }
        } else {
            throw new NodeException(this, inputName+ " is not supported.", __FILE__, __LINE__);
        }
        return ret;
    }

    // Rosette Lopez : 03132018 Currently not used.  This is for TAG file naming for the main file
    std::string formatFilename(const char* type, int size) {
        std::ostringstream new_filename;
        std::string temp_fname(filename);

        if(will_format_filename) {
            new_filename << fmt::format(temp_fname,
                                              fmt::arg("datatype", type),
                                              fmt::arg("colsize", size),
                                              fmt::arg("dim", dimension),
                                              fmt::arg("containertype", containertype)
                                                ); //format basename
        } else { //use original format if filename is not formatted
            new_filename << temp_fname;
        }
        return new_filename.str();
    }

    std::string formatFilename(const char* type, int size, int mapKey) {
        std::ostringstream new_filename;
        std::ostringstream key;
        std::string temp_fname(filename);
        key << "_" << std::setfill('0') << std::setw(4) << mapKey;

        std::size_t pos = temp_fname.find("_{");
        if (pos != std::string::npos) {
            pos = pos - 1;
            temp_fname.insert(pos, key.str());
        } else {
            temp_fname.append(key.str());
        }

        if(mode == RAW)
            new_filename << temp_fname << "_" << containertype << "_" << type << "_col" << size << ".raw";
        else
            new_filename << temp_fname << "_" << containertype << "_" << type << "_col" << size << ".txt";

        return new_filename.str();
    }

    void handle_files (std::string mname, std::string fname, const int count)
    {

        // List file name into a main/source file
        if (fout.is_open()) {
            fout << fname.c_str() << " ";
        }

        if ( !(std::find(list.begin(), list.end(), fname) != list.end())){
            list.push_back(fname);

            // Open internal file, no append mode since this is the first time the file is opened
            fout_in.open(fname.c_str(), std::ios::binary);
        }
        else {
            //Open internal file, append new data to existing file
            fout_in.open(fname.c_str(), std::ios::binary | std::ios::app);
        }

        if (fout_in.fail()) {
            throw new NodeException(this, std::string("Can't open file'") + fname + "'.", __FILE__, __LINE__);
        }
    }

    void calculate(int output_id, int count, Buffer &out) {
        ObjectRef input_ptr;
        input_ptr = getInput(inputID, count);
        out[count] = input_ptr;

        // Write into file this iteration frame value
        if (!fout.is_open()) {
            fout.open(filename.c_str(), std::fstream::out);
        }
        if (fout.fail()) {
            throw new NodeException(this, std::string("Can't open file'") + filename + "'.", __FILE__, __LINE__);
        } else {
            fout << count << " ";
        }
        

        if (typeid(*input_ptr) == typeid(Map<int,ObjectRef>)) {
            RCPtr <Map<int, ObjectRef> > input = getInput(inputID, count);
            for (Map<int, ObjectRef>::const_iterator it = input->begin(); it != input->end(); ++it) {
                std::string internal_filename("");

                if (typeid(*(it->second)) == typeid(Vector<float>)) {
                    const Vector<float> &tmpvec = object_cast<Vector<float> >(it->second);

                    // Check files if is previously created.  If it is new, store it to filename list for the main file
                    if (!fout_in.is_open()) {
                        internal_filename = formatFilename("float", tmpvec.size(), it->first);
                        handle_files(filename, internal_filename, count);
                    } 

                    // Write Data into file
                    if (mode == RAW) {
                        fout_in.write((const char *) &tmpvec[0], sizeof(float) * tmpvec.size());
                    } else {
                        printLine(tmpvec, fout_in, tmode == FIXED);
                    }
                } else if (typeid(*(it->second)) == typeid(Vector<std::complex<float> > )) {
                    const Vector<std::complex<float> > &tmpvec = object_cast<Vector<std::complex<float> > >(it->second);

                    // Check files if is previously created.  If it is new, store it to filename list for the main file
                    if (!fout_in.is_open()) {
                        internal_filename = formatFilename("complex_float", tmpvec.size(), it->first);
                        handle_files(filename, internal_filename, count);
                    }

                    // Write Data into file
                    if (mode == RAW) {
                        fout_in.write((const char *) &tmpvec[0], sizeof(float) * tmpvec.size() * 2);
                    } else {
                        printLine(tmpvec, fout_in, tmode == FIXED);
                    }
                } else {
                    // Input File Type is not Map <int, ObjectRef>
                    throw new NodeException(this, "Invalid Input Type.", __FILE__, __LINE__);
                }

                // End Internal File, next call might have data stored to another file
                if (fout_in.is_open()) {
                    fout_in.close();
                }
            }
            if (fout.is_open()) { 
                fout << std::endl;
            }
        } else {
            throw new NodeException(this, "Invalid Input Type", __FILE__, __LINE__);
        }
    }
};

