/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <fstream>
#include <iomanip>
#include <limits>

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>

#include "format.h"

using namespace FD;

class SaveMatrixFrames;

DECLARE_NODE(SaveMatrixFrames);
/*Node
 *
 * @name SaveMatrixFrames
 * @category HARK:MISC
 * @description Save matrix data of frames to file.
 *
 * @input_name INPUT
 * @input_type any
 * @input_description Data Stream
 *
 * @output_name OUTPUT
 * @output_type any
 * @output_description The same as input.
 *
 * @parameter_name FILENAME
 * @parameter_type string
 * @parameter_description The file name for saving matrix data.
 *
 * @parameter_name OUTPUTTYPE
 * @parameter_type string
 * @parameter_list TEXT:RAW
 * @parameter_value TEXT
 * @parameter_description Output file format type [default: TEXT]
 *
 * @parameter_name TEXTFORMAT
 * @parameter_type string
 * @parameter_valid OUTPUTTYPE=TEXT
 * @parameter_list FIXED:SCIENTIFIC
 * @parameter_value FIXED
 * @parameter_description Output file format type [default: FIXED]
 *
 END*/

template <class T>
inline int printLine(const RCPtr<Matrix<T> > &v, std::ofstream &out, bool is_fixed=true) {
    if(std::is_floating_point<T>::value){
        if(is_fixed){
            out << std::fixed << std::setprecision(std::numeric_limits<T>::max_digits10);
        }
        else{
            out << std::scientific << std::setprecision(std::numeric_limits<T>::max_digits10);
        }
    }
    else if(std::numeric_limits<T>::is_integer){
        out << std::dec;
    }
    else{
        out << "Vector's elements must be integer or floating point types. It is undefined behavior." << std::endl;
        return -1;
    }
    for(int row = 0; row < (*v).nrows(); row++){
        for(int col = 0; col < (*v).ncols(); col++){
            out << (*v)(row, col) << " ";
        }
    }
    out << std::endl;
    return 0;
}

template <class T>
inline int printLine(const RCPtr<Matrix<std::complex<T> > > &v, std::ofstream &out, bool is_fixed=true) {
    if(!std::is_floating_point<T>::value){
        out << "Complex numbers are not floating point types. It is undefined behavior." << std::endl;
        return -1;
    }
    if(is_fixed){
        out << std::fixed << std::setprecision(std::numeric_limits<T>::max_digits10);
    }
    else{
        out << std::scientific << std::setprecision(std::numeric_limits<T>::max_digits10);
    }
    for(int row = 0; row < (*v).nrows(); row++){
        for(int col = 0; col < (*v).ncols(); col++){
            out << (*v)(row, col).real() << " " << (*v)(row, col).imag() << " ";
        }
    }
    out << std::endl;
    return 0;
}

class SaveMatrixFrames : public BufferedNode {
    int inputID;
    int outputID;

    typedef enum {TEXT, RAW} OutputMode;
    typedef enum {FIXED, SCIENTIFIC} TextMode;

    OutputMode mode;
    TextMode tmode;
    std::string outputType;
    std::string textFormat;
    std::string filename;
    std::string formatted_filename;
    bool will_format_filename;
    std::ofstream fout;
    int dimension;

    public:
    SaveMatrixFrames(std::string nodeName, ParameterSet params) : BufferedNode(nodeName, params), mode(TEXT), dimension(2)
    {
        inputID = addInput("INPUT");
        outputID = addOutput("OUTPUT");
        outputType = object_cast<std::string>(parameters.get("OUTPUTTYPE"));
        textFormat = object_cast<std::string>(parameters.get("TEXTFORMAT"));
        filename = object_cast<std::string>(parameters.get("FILENAME"));
        if (filename.empty()) {
            throw new NodeException(NULL, std::string("FILENAME is empty."), __FILE__, __LINE__);
        }

        formatted_filename.clear();
        will_format_filename = false;

        if(filename.find("{datatype", 0) != std::string::npos ||
           filename.find("{colsize", 0) != std::string::npos ||
           filename.find("{rowsize", 0) != std::string::npos ||
           filename.find("{dim", 0) != std::string::npos) {
            will_format_filename = true;
        }

        if (outputType.compare("RAW") == 0) {
            mode = RAW;
        } else {
            mode = TEXT;
        }
        if (textFormat.compare("FIXED") == 0) {
            tmode = FIXED;
        } else {
            tmode = SCIENTIFIC;
        }
    }

    ~SaveMatrixFrames()
    {
       fout.close();
    }

    virtual int translateInput(const std::string &inputName) {
        int ret = -1;
        if (inputName == "INPUT") {
            for (unsigned int i = 0; i < inputs.size(); i++) {
                if (inputs[i].name == inputName) {
                    ret = i; break;
                }
            }
        } else {
            throw new NodeException(this, inputName+ " is not supported.", __FILE__, __LINE__);
        }
        return ret;
    }

    std::string formatFilename(const char* type, int row, int col) {
        std::ostringstream new_filename;

        if(will_format_filename) {
            new_filename << fmt::format(filename,
                                              fmt::arg("datatype", type),
                                              fmt::arg("rowsize", row),
                                              fmt::arg("colsize", col),
                                              fmt::arg("dim", dimension)); //format basename
        } else { //use original format if filename is not formatted
            new_filename << filename;
        }
        return new_filename.str();
    }

    void calculate(int output_id, int count, Buffer &out) {
        ObjectRef input_ptr;
        input_ptr = getInput(inputID, count);
        out[count] = input_ptr;

        if (typeid(*input_ptr) == typeid(Matrix<float>)) {
            RCPtr <Matrix<float> > input = getInput(inputID, count);
            if (!fout.is_open()) {
                formatted_filename = formatFilename("float", (*input).nrows(), (*input).ncols());
                fout.open(formatted_filename.c_str(), std::ios::binary);
                if (fout.fail()) {
                    throw new NodeException(this, std::string("Can't open file'") + filename + "'.", __FILE__, __LINE__);
                }
            }

            if (mode == RAW) {
                const Matrix<float> &feature = object_cast<Matrix<float> >(input);
                int row = feature.nrows();
                for (int line = 0; line < row; ++line) {
                    fout.write((const char *) &feature(line, 0), sizeof(float) * feature.ncols());
                }
            } else {
                printLine(input, fout, tmode == FIXED);
            }
        }
        else if (typeid(*input_ptr) == typeid(Matrix<std::complex<float> >)) {
            RCPtr <Matrix<std::complex<float> > > input = getInput(inputID, count);
            if (!fout.is_open()) {
                formatted_filename = formatFilename("complex_float", (*input).nrows(), (*input).ncols());
                fout.open(formatted_filename.c_str(), std::ios::binary);
                if (fout.fail()) {
                    throw new NodeException(this, std::string("Can't open file'") + filename + "'.", __FILE__, __LINE__);
                }
            }
            if (mode == RAW) {
                const Matrix <std::complex<float> > &feature = object_cast <Matrix<std::complex<float> > >(input);
                int row = feature.nrows();
                for (int line = 0; line < row; ++line) {
                    fout.write((const char *) &feature(line, 0), sizeof(float) * feature.ncols() * 2);
                }
            } else {
                printLine(input, fout, tmode == FIXED);
            }
        }
        else {
            throw new NodeException(this, "Invalid Input Type.", __FILE__, __LINE__);
        }
    }
};

