/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include "BufferedNode.h"
#include "Buffer.h"
#include "Vector.h"
#include "Map.h"
#include "Source.h"
#include <iostream>
#include <cstring>
#include <fstream>
#include <string>
#include <sstream>


//#include "fft.hpp"
#include "SpectralMeanMod.hpp"

using namespace std;
using namespace HARK;

class SpectralMeanNormalizationIncremental;

DECLARE_NODE(SpectralMeanNormalizationIncremental)
/*Node
 *
 * @name SpectralMeanNormalizationIncremental
 * @category HARK:FeatureExtraction
 * @description Perform Spectral Mean Normalization (SMN). A spectral mean is calculated per direction (resolution: 10 degrees).
 *
 * @input_name INPUT
 * @input_type Map<int,ObjectRef>
 * @input_description Input feature vectors. The key is source ID, and the value is a feature vector (Vector<float>).
 *
 * @input_name SM_HISTORY
 * @input_type Map<int,ObjectRef>
 * @input_description Optional input. ObjectRef should be Vector<Vector<float> > of the spectral history for each source.
 *
 * @input_name NOT_EOF
 * @input_type bool
 * @input_description Optional input. Use this when you specify OUTPUT_FNAME or SM_EXPORT_FILENAME.
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description Feature vectors after SMN.
 *
 * @parameter_name FBANK_COUNT
 * @parameter_type int
 * @parameter_value 13
 * @parameter_description Size of a static part of a feature vector. [default: 13]
 *
 * @parameter_name PERIOD
 * @parameter_type int
 * @parameter_value 20
 * @parameter_description Period to calculate SM. If 1, SM calculated in every frame. If 0, SM_ALGORITHM is used for all frames. [default: 20]
 *
 * @parameter_name SM_ALGORITHM
 * @parameter_type string
 * @parameter_value INCREMENTAL
 * @parameter_list INCREMENTAL:PREV_SM:ZERO:FILE
 * @parameter_description Algorithm to decide the initial SM until reaching PERIOD + IGNORE_FRAMES. If INCREMENTAL, SM is calculated in every frame. If PREV_SM, SM is taken from the previous source. If ZERO, SM is zeroed. If FILE, SM is loaded from SM_FILENAME.
 *
 * @parameter_name SM_FILENAME
 * @parameter_type string
 * @parameter_value 
 * @parameter_valid SM_ALGORITHM=FILE
 * @parameter_description csv file name of the initial SM until reaching PERIOD + IGNORE_FRAMES.
 *
 * @parameter_name SM_HISTORY_FILENAME
 * @parameter_type string
 * @parameter_value 
 * @parameter_description csv file name of the initial spectral history for an initial SM
 *
 * @parameter_name IGNORE_FRAMES
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description Number of first ingnored frames to calculate SM [default: 0]
 *
 * @parameter_name BASENAME
 * @parameter_type string
 * @parameter_value smn_
 * @parameter_description Basename of SMN files. The filename will be BASENAME+id+.csv [default: smn_]
 *
 * @parameter_name OUTPUT_FNAME
 * @parameter_type string
 * @parameter_value out.txt
 * @parameter_description 
 *
 * @parameter_name SM_EXPORT_FILENAME
 * @parameter_type string
 * @parameter_value 
 * @parameter_description csv file name of the export SM which follows SM_EXPORT_ALGORITHM. If vacant, the exporting is disabled.
 *
 * @parameter_name SM_EXPORT_ALGORITHM
 * @parameter_type string
 * @parameter_value LAST_SRC
 * @parameter_list LAST_SRC:SRC_AVERAGE:FRAME_AVERAGE
 * @parameter_description Algorithm to decide the SM for exporting. If LAST_SRC, SM of the last source is saved. If SRC_AVERAGE, SM of all the sources are averaged and saved. If FRAME_AVERAGE, SM of all the frames in all sources are averaged and saved.
 *
 END*/

class SpectralMeanNormalizationIncremental : public BufferedNode
{
  int inputID;
  int smhistoryID;
  int noteofID;
  int outputID;
  int fbank_count;
  int period;
  string sm_algorithm;
  vector<float> infileSM;
  bool b_sm_history_filename;
  Vector<Vector<float> > historySM;
  int ignore_frames;
  vector<float>  prevSM;
  vector<float>  curSM;
  map<int, SpectralMeanMod> mapSM;
  string basename;
  bool b_basename;
  map<int, shared_ptr<ofstream> > streams;
  string output_fname;
  bool b_output_fname;
  ofstream fout2;
  bool b_sm_export;
  string sm_export_filename;
  string sm_export_algorithm;
  vector<int> sm_export_nframe;
  vector<vector<float> > sm_export_sm;
 
public:
  SpectralMeanNormalizationIncremental(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params), smhistoryID(-1), noteofID(-1)
  {
    inputID     = addInput("INPUT");
    outputID    = addOutput("OUTPUT");
    fbank_count = dereference_cast<int>(parameters.get("FBANK_COUNT"));
    period = dereference_cast<int>(parameters.get("PERIOD"));
    sm_algorithm = object_cast<string>(parameters.get("SM_ALGORITHM"));
    if(sm_algorithm == "FILE"){
      string sm_filename = object_cast<string>(parameters.get("SM_FILENAME"));
      ifstream fin(sm_filename.c_str());
      if(fin.fail()){
	throw new NodeException(NULL, string("Can't open '")+sm_filename+string("'."), __FILE__, __LINE__);
      } 
      if(fin.eof()){
	throw new NodeException(NULL, string("No_coef file defined '")+sm_filename+string("'."), __FILE__, __LINE__);
      }
      string str;
      getline(fin, str);
      string token;
      istringstream stream( str );
      infileSM.resize(0);
      while( getline( stream, token, ',' ) ) {
	float val;
	stringstream ss(token);
	ss >> val;
	infileSM.push_back(val);
      }
      if(fbank_count != infileSM.size()){
	throw new NodeException(NULL, string("Vector size of SM_FILENAME mismatches to FBANK_COUNT'"), __FILE__, __LINE__);
      } 
    }
    b_sm_history_filename = false;
    if(parameters.exist("SM_HISTORY_FILENAME") == true){
      b_sm_history_filename = true;
      string sm_history_filename = object_cast<string>(parameters.get("SM_HISTORY_FILENAME"));
      ifstream fin(sm_history_filename.c_str());
      if(fin.fail()){
	throw new NodeException(NULL, string("Can't open '")+sm_history_filename+string("'."), __FILE__, __LINE__);
      } 
      if(fin.eof()){
	throw new NodeException(NULL, string("No_coef file defined '")+sm_history_filename+string("'."), __FILE__, __LINE__);
      }
      string str;
      historySM.resize(0);
      while( getline(fin, str) ) {
	string token;
	istringstream stream( str );
	Vector<float> tmpV;
	tmpV.resize(0);
	while( getline( stream, token, ',' ) ) {
	  float val;
	  stringstream ss(token);
	  ss >> val;
	  tmpV.push_back(val);
	}
	historySM.push_back(tmpV);
      }
      for (int i = 0; i < historySM.size(); i++) {      
	if(fbank_count != historySM[i].size()){
	  throw new NodeException(NULL, string("Vector size of SM_FILENAME mismatches to FBANK_COUNT'"), __FILE__, __LINE__);
	} 
      }
    }
    ignore_frames = dereference_cast<int>(parameters.get("IGNORE_FRAMES"));
    if(parameters.exist("BASENAME") == true){	
      basename = object_cast<string>(parameters.get("BASENAME"));
      b_basename = true;
    } else {
      b_basename = false;
    }
    if(parameters.exist("OUTPUT_FNAME") == true){	
      output_fname = object_cast<string>(parameters.get("OUTPUT_FNAME"));
      fout2.open(output_fname.c_str());
      if(fout2.fail())
	{
	  throw new NodeException(NULL, string("Can't open file'")+output_fname+"'.", __FILE__, __LINE__);
	}
      b_output_fname = true;
    } else {
      b_output_fname = false;
    }
    prevSM.resize(fbank_count);
    for (int i = 0; i < fbank_count; i++) {
      prevSM[i] = 0.0f;
    }	 	
    curSM.resize(fbank_count);

    b_sm_export = false;
    if(parameters.exist("SM_EXPORT_FILENAME") == true){	
      b_sm_export = true;
      sm_export_filename = object_cast<string>(parameters.get("SM_EXPORT_FILENAME"));
      sm_export_algorithm = object_cast<string>(parameters.get("SM_EXPORT_ALGORITHM"));
      sm_export_nframe.resize(0);
      sm_export_sm.resize(0);
    }
    
    inputsCache[inputID].lookAhead = 1;
    inOrder = true;
  }

  ~SpectralMeanNormalizationIncremental() {
    if(b_sm_export){
      shared_ptr<ofstream> fout;
      fout.reset(new ofstream(sm_export_filename.c_str()));
      vector<float> exportSM(fbank_count);
      if(sm_export_algorithm == "LAST_SRC"){
	for (int i = 0; i < fbank_count; i++) {
	  exportSM[i] = prevSM[i];	
	}
      }
      if(sm_export_algorithm == "SRC_AVERAGE"){
	for (int i = 0; i < fbank_count; i++) {
	  float sum = 0.0f;
	  for (int j = 0; j < sm_export_sm.size(); j++) {
	    sum += sm_export_sm[j][i];
	  }
	  exportSM[i] = sum / (float)sm_export_sm.size();	
	}
      }
      if(sm_export_algorithm == "FRAME_AVERAGE"){
	for (int i = 0; i < fbank_count; i++) {
	  float sum = 0.0f;
	  float nframes = 0.0f;
	  for (int j = 0; j < sm_export_sm.size(); j++) {
	    sum += sm_export_sm[j][i] * (float)sm_export_nframe[j];
	    nframes += (float)sm_export_nframe[j];
	  }
	  exportSM[i] = sum / nframes;	
	}	
      }      
      *fout << exportSM[0];
      for (int i = 1; i < fbank_count; i++) {
	*fout << ", " << exportSM[i];
      }
    }
  }

  virtual int translateInput(const std::string &inputName) {
    int ret = -1;
    if (inputName == "NOT_EOF") {
      noteofID = addInput(inputName);
      inputsCache[noteofID].lookAhead = 1;
      ret = noteofID;
    }
    else if (inputName == "SM_HISTORY") {
      smhistoryID = addInput(inputName);
      ret = smhistoryID;
    }
    else if (inputName == "INPUT") {
      for (unsigned int i = 0; i < inputs.size(); i++) {
	if (inputs[i].name == inputName) {
	  ret = i; break;
	}
      }
    }
    else {
      throw new NodeException(this, inputName+ " is not supported.", __FILE__, __LINE__);
    }
    return ret;
  }


  void calculate(int output_id, int count, Buffer &out)
  {

    RCPtr<Map<int, ObjectRef> > input1 = getInput(inputID, count);
    const Map<int, ObjectRef>& in1 = *input1;
      
    RCPtr<Map<int, ObjectRef> > input2 = getInput(inputID, count + 1);
    const Map<int, ObjectRef>& in2 = *input2;
     
    bool eofn = true;
    if(noteofID != -1){
      ObjectRef peofn = getInput(noteofID, count+1);
      eofn = dereference_cast<bool>(peofn);
    }

    RCPtr<Map<int, ObjectRef> > smhistory_map(NULL);
    if(smhistoryID != -1){
        smhistory_map = getInput(smhistoryID, count);
    }

    Map<int,ObjectRef>& output = *new Map<int,ObjectRef>;
    out[count] = &output;

    for (Map<int, ObjectRef>::const_iterator it = in1.begin(); it != in1.end(); ++it) {

      RCPtr<Vector<float> > fbank = it->second;
      if (fbank->size() < fbank_count) {
	throw new NodeException(this, "Size of input feature vector is less than FBANK_COUNT.", __FILE__, __LINE__);
      }
	
      // Detect beggining of sources
      if (mapSM.find(it->first) == mapSM.end()) {
	SpectralMeanMod tmpSM;
	tmpSM.Initialize(fbank_count);
	if(b_sm_history_filename){	
	  for(int i = 0; i < historySM.size(); i++){
	    tmpSM.AddSpectrum(historySM[i]);
	  }
	}
	if(smhistoryID != -1){
	  if (smhistory_map->find(it->first) != smhistory_map->end()) {
	    Map<int, ObjectRef>::const_iterator smhistory_it = smhistory_map->find(it->first);
	    Vector<ObjectRef>& tmpVV = object_cast<Vector<ObjectRef> >(smhistory_it->second);
	    for (int i = 0; i < tmpVV.size(); i++) {
	      RCPtr<Vector<float> > tmpV = tmpVV[i];
	      tmpSM.AddSpectrum(*tmpV);
	    }
	  }
	}
	mapSM[it->first] = tmpSM;
      } 

      // Load spectrum
      map<int, SpectralMeanMod>::iterator it_mapSM = mapSM.find(it->first);
      it_mapSM->second.AddDuration();

      // Add Spectrum
      if(it_mapSM->second.GetDuration() > ignore_frames)
	it_mapSM->second.AddSpectrum(*fbank);

      // Calculate SM
      if(period != 0 && it_mapSM->second.GetSpecCount() % period == 0 && it_mapSM->second.GetSpecCount() != 0) 
	it_mapSM->second.CalculateMean();

      // Decide SM
      if(it_mapSM->second.GetDuration() < ignore_frames + period || period == 0){
	if (sm_algorithm == "INCREMENTAL"){
	  if(it_mapSM->second.GetDuration() <= ignore_frames){
	    for (int i = 0; i < fbank_count; i++) {
	      curSM[i] = 0.0f;
	    }	 	
	  } else {
	    it_mapSM->second.CalculateMean();
	    for (int i = 0; i < fbank_count; i++) {
	      curSM[i] = it_mapSM->second.GetSpectralMean()[i];
	    }
	  }
	} else if (sm_algorithm == "PREV_SM") {
	  for (int i = 0; i < fbank_count; i++) {
	    curSM[i] = prevSM[i];
	  }	 	
	} else if (sm_algorithm == "ZERO") {
	  for (int i = 0; i < fbank_count; i++) {
	    curSM[i] = 0.0f;
	  }	 	
	} else if (sm_algorithm == "FILE") {
	  for (int i = 0; i < fbank_count; i++) {
	    curSM[i] = infileSM[i];
	  }	 	
	}
      } else {
	for (int i = 0; i < fbank_count; i++) {
	  curSM[i] = it_mapSM->second.GetSpectralMean()[i];
	}	 	
      }

      // Logging
      if(b_basename){		
	shared_ptr<ofstream> fout;
	if (streams.find(it->first) == streams.end()) {
	  ostringstream name_str;
	  name_str << basename << it->first << ".csv";
	  fout.reset(new ofstream(name_str.str().c_str()));
	  streams[it->first] = fout;
	} else {
	  fout = streams[it->first];
	}
	it_mapSM->second.CalculateMeanTmp();
	*fout << count << ", " << it->first << ", " << it_mapSM->second.GetDuration() << ", " << it_mapSM->second.GetSpecCount();
	for (int i = 0; i < fbank_count; i++) {
	  *fout << ", " << it_mapSM->second.GetSpectralMeanTmp()[i];
	}
	for (int i = 0; i < fbank_count; i++) {
	  *fout << ", " << curSM[i];
	}
	*fout << endl;	      
      }		

      // Subtract SM
      RCPtr<Vector<float> > fbankout(Vector<float>::alloc(fbank->size()));
      for (int i = 0; i < fbank->size(); i++){
	if(i < fbank_count)
	  (*fbankout)[i] = (*fbank)[i] - curSM[i];
	else
	  (*fbankout)[i] = (*fbank)[i];
      }
      output[it->first] = fbankout;

      // Detect end of sources -> normalization and saving prevSM
      if (in2.find(it->first) == in2.end() || !eofn) {
	it_mapSM->second.CalculateMean();
	if(b_output_fname){		
	  fout2 << count << ", " << it->first << ", " << it_mapSM->second.GetDuration() << ", " << it_mapSM->second.GetSpecCount();
	  for (int i = 0; i < fbank_count; i++) {
	    fout2 << ", " << it_mapSM->second.GetSpectralMean()[i];
	  }	 
	  fout2 << endl;
	}
	for (int i = 0; i < fbank_count; i++) {
	  prevSM[i] = it_mapSM->second.GetSpectralMean()[i];
	}
	if(b_sm_export){
	  sm_export_nframe.push_back(it_mapSM->second.GetSpecCount());
	  sm_export_sm.push_back(prevSM);
	}	
	mapSM.erase(it_mapSM);
	if(b_basename){		
	  streams.erase(streams.find(it_mapSM->first));
	}
      }

    }

  }

  IN_ORDER_NODE_SPEEDUP(SpectralMeanNormalizationIncremental);
};
