/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <iostream>
#include <sstream>
#include <string>
#include <math.h>
#include <BufferedNode.h>
#include <Buffer.h>
#include <Source.h>
#include <Vector.h>
#include <Matrix.h>
#include <Map.h>

//using namespace std;
using namespace FD;

class TextConverter;

DECLARE_NODE(TextConverter);
/*Node
 *
 * @name TextConverter
 * @category HARK:MISC
 * @description This node parses arbitrary input and outputs JSON formatted text. This node's output can be used for TEXT input of HarkDataStreamSender.
 *
 * @input_name DATA
 * @input_type any
 * @input_description The name of the terminal becomes the key of dict type. For example, if it is a terminal named DATA, it will be output striing of the "{'DATA': <value(s)>}". In addition, if you no need DATA terminal, you can delete the DATA terminal which is prepared. Any number of inputs can be added.
 *
 * @output_name TEXT
 * @output_type string
 * @output_description Outputs JSON formatted text based on input. Multiple inputs are automatically parsed. One JSON text is output per frame.
 * 
 * @parameter_name ENABLE_DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Enable debug print. The same text as output from OUTPUT is printed with standard output.
 *
END*/

class TextConverter : public BufferedNode {
  int inputID;
  int outputID;
  bool enable_debug;

public:
  TextConverter(std::string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    outputID = addOutput("TEXT");
    enable_debug = dereference_cast<bool>(parameters.get("ENABLE_DEBUG"));
    inOrder = true;
  }

  virtual int translateInput (const std::string &inputName) {
    for (unsigned int i=0; i< inputs.size(); i++) {
      if (inputs[i].name == inputName) {
	return i;
      }
    }
    return addInput(inputName);
  }

  void calculate(int output_id, int count, Buffer &out)
  {
    std::ostringstream os;
    os << "{";
    for(int id=0; id<inputs.size(); ++id){
      ObjectRef in = getInput(id, count);
      if(id != 0){ os << ", "; }
      os << "'" << inputs[id].name << "': ";
      if(typeid(*in) == typeid(Bool)){
        const bool &val = object_cast<Bool>(in);
        os << (val ? "true" : "false");
      }
      else if(typeid(*in) == typeid(Int)){
        const int &val = object_cast<Int>(in);
        os << val;
      }
      else if(typeid(*in) == typeid(Float)){
        const float &val = object_cast<Float>(in);
        os << val;
      }
      else if(typeid(*in) == typeid(std::complex<float>)){
        const std::complex<float> &val = object_cast<std::complex<float> >(in);
        os << "(" << val.real() << ", " << val.imag() << ")";
      }
      else if(typeid(*in) == typeid(String)){
        const String &text = object_cast<String>(in);
        os << text;
      }
      else if(typeid(*in) == typeid(Vector<int>)){
        const Vector<int> &vec = object_cast<Vector<int> >(in);
        os << "[";
        for(int i=0; i<vec.size(); ++i){
          if(i != 0){ os << ", "; }
          os << vec[i];
        }
        os << "]";
      }
      else if(typeid(*in) == typeid(Vector<float>)){
        const Vector<float> &vec = object_cast<Vector<float> >(in);
        os << "[";
        for(int i=0; i<vec.size(); ++i){
          if(i != 0){ os << ", "; }
          os << vec[i];
        }
        os << "]";
      }
      else if(typeid(*in) == typeid(Vector<std::complex<float> >)){
        const Vector<std::complex<float> > &vec = object_cast<Vector<std::complex<float> > >(in);
        os << "[";
        for(int i=0; i<vec.size(); ++i){
          if(i != 0){ os << ", "; }
          os << "(" << vec[i].real() << ", " << vec[i].imag() << ")";
        }
        os << "]";
      }
      else if(typeid(*in) == typeid(Vector<ObjectRef>)){
        const Vector<ObjectRef> &vec = object_cast<Vector<ObjectRef> >(in);
        os << "{";
        for(int i=0; i<vec.size(); ++i){
          if(i != 0){ os << ", "; }
          RCPtr<Source> src = vec[i];
          os << "'" << src->id << "': {" << "'x': [" << src->x[0] << ", " << src->x[1] << ", " << src->x[2] << "], 'power': " << src->power << "}";
        }
        os << "}";
      }
      else if(typeid(*in) == typeid(Matrix<int>)){
        const Matrix<int> &mat = object_cast<Matrix<int> >(in);
        os << "[";
        for(int i=0; i<mat.nrows(); ++i){
          if(i != 0){ os << ", "; }
          os << "[";
          for(int j=0; j<mat.ncols(); ++j){
            if(j != 0){ os << ", "; }
            os << mat(i, j);
          }
          os << "]";
        }
        os << "]";
      }
      else if(typeid(*in) == typeid(Matrix<float>)){
        const Matrix<float> &mat = object_cast<Matrix<float> >(in);
        os << "[";
        for(int i=0; i<mat.nrows(); ++i){
          if(i != 0){ os << ", "; }
          os << "[";
          for(int j=0; j<mat.ncols(); ++j){
            if(j != 0){ os << ", "; }
            os << mat(i, j);
          }
          os << "]";
        }
        os << "]";
      }
      else if(typeid(*in) == typeid(Matrix<std::complex<float> >)){
        const Matrix<std::complex<float> > &mat = object_cast<Matrix<std::complex<float> > >(in);
        os << "[";
        for(int i=0; i<mat.nrows(); ++i){
          if(i != 0){ os << ", "; }
          os << "[";
          for(int j=0; j<mat.ncols(); ++j){
            if(j != 0){ os << ", "; }
            os << "(" << mat(i, j).real() << ", " << mat(i, j).imag() << ")";
          }
          os << "]";
        }
        os << "]";
      }
      else if(typeid(*in) == typeid(Map<int, ObjectRef>)){
        RCPtr<Map<int, ObjectRef> > map_in = in;
        Map<int, ObjectRef>::iterator it = map_in->begin();
        os << "{";
        if(typeid(*(it->second)) == typeid(Vector<int>)){
          for(; it != map_in->end(); ++it){
            const Vector<int> &vec = object_cast<Vector<int> >(it->second);
            if(it != map_in->begin()){ os << ", "; }
            os << "'" << it->first << "': [";
            for(int i=0; i<vec.size(); ++i){
              if(i != 0){ os << ", "; }
              os << vec[i];
            }
            os << "]";
          }
        }
        else if(typeid(*(it->second)) == typeid(Vector<float>)){
          for(; it != map_in->end(); ++it){
            const Vector<float> &vec = object_cast<Vector<float> >(it->second);
            if(it != map_in->begin()){ os << ", "; }
            os << "'" << it->first << "': [";
            for(int i=0; i<vec.size(); ++i){
              if(i != 0){ os << ", "; }
              os << vec[i];
            }
            os << "]";
          }
        }
        else if(typeid(*(it->second)) == typeid(Vector<std::complex<float> >)){
          for(; it != map_in->end(); ++it){
            const Vector<std::complex<float> > &vec = object_cast<Vector<std::complex<float> > >(it->second);
            if(it != map_in->begin()){ os << ", "; }
            os << "'" << it->first << "': [";
            for(int i=0; i<vec.size(); ++i){
              if(i != 0){ os << ", "; }
              os << "(" << vec[i].real() << ", " << vec[i].imag() << ")";
            }
            os << "]";
          }
        }
        else if(typeid(*(it->second)) == typeid(Matrix<int>)){
          for(; it != map_in->end(); ++it){
            const Matrix<int> &mat = object_cast<Matrix<int> >(it->second);
            if(it != map_in->begin()){ os << ", "; }
            os << "'" << it->first << "': [";
            for(int i=0; i<mat.nrows(); ++i){
              if(i != 0){ os << ", "; }
              os << "[";
              for(int j=0; j<mat.ncols(); ++j){
                if(j != 0){ os << ", "; }
                os << mat(i, j);
              }
              os << "]";
            }
            os << "]";
          }
        }
        else if(typeid(*(it->second)) == typeid(Matrix<float>)){
          for(; it != map_in->end(); ++it){
            const Matrix<float> &mat = object_cast<Matrix<float> >(it->second);
            if(it != map_in->begin()){ os << ", "; }
            os << "'" << it->first << "': [";
            for(int i=0; i<mat.nrows(); ++i){
              if(i != 0){ os << ", "; }
              os << "[";
              for(int j=0; j<mat.ncols(); ++j){
                if(j != 0){ os << ", "; }
                os << mat(i, j);
              }
              os << "]";
            }
            os << "]";
          }
        }
        else if(typeid(*(it->second)) == typeid(Matrix<std::complex<float> >)){
          for(; it != map_in->end(); ++it){
            const Matrix<std::complex<float> > &mat = object_cast<Matrix<std::complex<float> > >(it->second);
            if(it != map_in->begin()){ os << ", "; }
            os << "'" << it->first << "': [";
            for(int i=0; i<mat.nrows(); ++i){
              if(i != 0){ os << ", "; }
              os << "[";
              for(int j=0; j<mat.ncols(); ++j){
                if(j != 0){ os << ", "; }
                os << "(" << mat(i, j).real() << ", " << mat(i, j).imag() << ")";
              }
              os << "]";
            }
            os << "]";
          }
        }
        os << "}";
      }
    }
    os << "}";
    if(enable_debug){ std::cout << os.str() << std::endl; }
    out[count] = ObjectRef(new String(os.str()));
  }
};

