/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include "BufferedNode.h"
#include "Buffer.h"
#include "Vector.h"
#include "Matrix.h"
#include "Source.h"
#include "Map.h"

#include <string.h>
#include <sys/time.h>

using namespace std;
using namespace HARK;

class ZeroCrossCounter {
private:
    std::vector<float> buffer;
    std::vector<bool> hasPower;
    std::vector<bool> hasZeroCross;
    int size;
    int offset;
    float threshold;
    int zeroCrossCount;

public:
    ZeroCrossCounter() {
    }

public:
    void reset(int size, float threshold) {
        this->buffer.resize(size);
        this->hasPower.resize(size);
        this->hasZeroCross.resize(size);
        this->buffer.clear();
        this->hasPower.clear();
        this->hasZeroCross.clear();
        this->size = size;
        this->offset = 0;
        this->threshold = threshold;
        this->zeroCrossCount = 0;
    }

    int update(const float *data, int count) {
        int i;
        for(i = 0; i < count; i++){
            int po = (this->offset == 0) ? (this->size - 1) : (this->offset - 1);
            this->buffer[this->offset] = data[i];
            this->hasPower[this->offset] = (fabs(this->buffer[this->offset]) > this->threshold);

            if(this->hasZeroCross[this->offset]){
                this->zeroCrossCount--;
            }
            this->hasZeroCross[this->offset] = (this->hasPower[po]
                    && (this->buffer[po] * this->buffer[this->offset]) < 0);
            if(this->hasZeroCross[this->offset]){
                this->zeroCrossCount++;
            }
            if(++this->offset >= this->size){
                this->offset = 0;
            }
        }

        return this->zeroCrossCount;
    }

    int getZeroCrossCount() const {
        return this->zeroCrossCount;
    }

};

class VAD_ZC;

DECLARE_NODE(VAD_ZC)
/*Node
 *
 * @name VAD_ZC
 * @category HARK:MISC
 * @description AD-in tool to record/split/send/receive speech data
 *
 * @input_name INPUT
 * @input_type Matrix<float>
 * @input_description Input PCM data. Each row corresponds to channel.
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description Output PCM data.  If VAD_ZC judges a channel has enough audio energy, it outpus Vector<float> with channel index as a key.
 *
 * @output_name SOURCES
 * @output_type Vector<ObjectRef>
 * @output_description Source information.  If OUTPUT contains PCM data, corresponding source information is output.
 *
 * @parameter_name ADVANCE
 * @parameter_type int
 * @parameter_value 160
 * @parameter_description The shift length beween adjacent frames (in samples)[default: 160].
 *
 * @parameter_name LENGTH
 * @parameter_type int
 * @parameter_value 512
 * @parameter_description PCM samples for each frame [default: 512].
 *
 * @parameter_name CHANNEL_COUNT
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description Number of input channels.
 *
 * @parameter_name SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_description Sampling rate (in samples) [default: 16000].
 *
 * @parameter_name STRIP_ZERO
 * @parameter_type bool
 * @parameter_value true
 * @parameter_description
 *
 * @parameter_name ZERO_MEAN
 * @parameter_type bool
 * @parameter_value false
 * @parameter_description Flag if need to remove DC component [default: false].
 *
 * @parameter_name LEVEL_THRESHOLD
 * @parameter_type float
 * @parameter_value 2000
 * @parameter_description Silence cut level threshold.  Corresponds to the adintool's option '-lv' (in 16bit PCM value) [default: 2000].
 *
 * @parameter_name ZEROCROSS_THRESHOLD
 * @parameter_type int
 * @parameter_value 60
 * @parameter_description Silence cut zero-cross number.  Corresponds to the adintool's option '-zc' [default: 60].
 *
 * @parameter_name HEAD_MARGIN
 * @parameter_type float
 * @parameter_value 300
 * @parameter_description Head margin length.  Corresponds to the adintool's option '--headmargin' (in msec) [default: 300].
 *
 * @parameter_name TAIL_MARGIN
 * @parameter_type float
 * @parameter_value 400
 * @parameter_description Tail margin length.  Corresponds to the adintool's option '--tailmargin' (in msec) [default: 400].
 *
 END*/

class VAD_ZC: public BufferedNode {

    // I/O id
    int inputID;
    int audioOutputID;
    int sourcesOutputID;

    // params
    int channel_count;
    int advance;
    int length;
    int sampling_rate;
    bool strip_zero;
    bool zero_mean;
    float level_threshold;
    int zerocross_threshold;
    float head_margin; // in millisec
    float tail_margin; // in millisec

    // workset
    int zerocross_threshold_count; // zerocross_threshold をフレーム数で表した値
    int source_count;       // みつかった音源の通し番号
    int head_margin_frames; // head_margin をフレーム数で表した値
    int tail_margin_frames; // tail_margin をフレーム数で表した値
    int last_frame_index;
    int channel_buffer_length;
    int channel_buffer_offset;
    map<int, Source> channel_sources;

    struct ChannelWorkset {
        vector<vector<float> > buffer; // channel_buffer[frame_offset][]
        ZeroCrossCounter zerocross_counter;
        int last_active_frame_index;
        float dc;
        bool dc_prepared;
    };
    vector<ChannelWorkset> channel_worksets;

public:
    VAD_ZC(string nodeName, ParameterSet params) :
            BufferedNode(nodeName, params) {
        this->inputID = this->addInput("INPUT");
        this->audioOutputID = this->addOutput("OUTPUT");
        this->sourcesOutputID = this->addOutput("SOURCES");

        this->channel_count = dereference_cast<int>(this->parameters.get("CHANNEL_COUNT"));
        this->advance = dereference_cast<int>(this->parameters.get("ADVANCE"));
        this->length = dereference_cast<int>(this->parameters.get("LENGTH"));
        this->sampling_rate = dereference_cast<int>(this->parameters.get("SAMPLING_RATE"));
        this->strip_zero = dereference_cast<bool>(this->parameters.get("STRIP_ZERO"));
        this->zero_mean = dereference_cast<bool>(this->parameters.get("ZERO_MEAN"));
        this->level_threshold = dereference_cast<float>(this->parameters.get("LEVEL_THRESHOLD"));
        this->zerocross_threshold = dereference_cast<int>(this->parameters.get("ZEROCROSS_THRESHOLD"));
        this->head_margin = dereference_cast<float>(this->parameters.get("HEAD_MARGIN"));
        this->tail_margin = dereference_cast<float>(this->parameters.get("TAIL_MARGIN"));

        this->source_count = 0;
        this->last_frame_index = 0;
        this->head_margin_frames = (int) (this->head_margin * this->sampling_rate / this->advance / 1000);
        this->tail_margin_frames = (int) (this->tail_margin * this->sampling_rate / this->advance / 1000);
        if(this->head_margin_frames == 0){
            this->head_margin_frames = 1;
        }

        this->channel_buffer_length = this->advance * this->head_margin_frames;
        this->channel_worksets.resize(this->channel_count);
        for(int i = 0; i < this->channel_count; i++){
            this->channel_worksets[i].buffer.resize(this->head_margin_frames);
            this->channel_worksets[i].zerocross_counter.reset(this->channel_buffer_length, this->level_threshold);
            for(int j = 0; j < this->head_margin_frames; j++){
                this->channel_worksets[i].buffer[j].resize(this->length);
            }
            this->channel_worksets[i].last_active_frame_index = -this->tail_margin_frames;
            this->channel_worksets[i].dc = 0;
            this->channel_worksets[i].dc_prepared = false;
        }
        this->zerocross_threshold_count = this->zerocross_threshold * this->channel_buffer_length / this->sampling_rate;
        //cout << "zerocross_num=" << this->zerocross_threshold << ", length=" << this->channel_buffer_length << ", zerocross_count=" << this->zerocross_threshold_count << endl;

        this->inputsCache[this->inputID].lookAhead = this->head_margin_frames;
        this->inputsCache[this->inputID].lookBack = this->head_margin_frames;
        this->inOrder = true;
    }

    ~VAD_ZC() {
    }

    virtual void initialize(void) {
        std::vector<OutputCacheInfo>::iterator it;

        int maxahead = 0, maxback = 0;
        for(it = this->outputs.begin(); it != this->outputs.end(); ++it){
            if(maxahead < it->lookAhead){
                maxahead = it->lookAhead;
            }
            if(maxback < it->lookBack){
                maxback = it->lookBack;
            }
        }
        for(it = this->outputs.begin(); it != this->outputs.end(); ++it){
            it->lookAhead = maxahead + 1;
            it->lookBack = maxback + 1;
        }

        this->BufferedNode::initialize();
    }

    void receive(int count){
        while(this->last_frame_index < count){
            //cout << "receive: " << this->last_frame_index << "/" << count << endl;
            /* cursor in channel_buffers */
            int buffer_cursor = this->last_frame_index % this->head_margin_frames;

            /* receive input object */
            ObjectRef in = getInput(this->inputID, this->last_frame_index);

            /* input entity is Matrix<float> type. */
            Matrix<float>& x = object_cast<Matrix<float> >(in);
            //std::cout << "in: " << x << std::endl;

            /* check inter-frame cosistency */
            if(this->channel_count != x.nrows()){
                throw new NodeException(this, "Channel count mismatch.", __FILE__, __LINE__);
            }
            if(this->length > 0 && this->length != x.ncols()){
                throw new NodeException(this, "Buffer length mismatch.", __FILE__, __LINE__);
            }
            else if(this->length == 0){
                this->length = x.ncols();
            }

            /* append to buffer */
            int channel_index;
            for(channel_index = 0; channel_index < channel_count; channel_index++){
                ChannelWorkset &channel_workset = this->channel_worksets[channel_index];

                if(count == 0 && this->zero_mean && !channel_workset.dc_prepared){
                    float dc = 0;
                    for(int k = 0; k < this->length; k++){
                        dc += x[channel_index][k];
                    }
                    channel_workset.dc = dc / this->length;
                    channel_workset.dc_prepared = true;
                }

                /* update buffer */
                for(int i = 0; i < this->length; i++){
                    channel_workset.buffer[buffer_cursor][i] = x[channel_index][i] - channel_workset.dc;
                }

                /* update zero-cross counter */
                channel_workset.zerocross_counter.update(x[channel_index] + (this->length - this->advance),
                        this->advance);
            }

            /* update cursor */
            this->last_frame_index++;
        }
    }

    void calculate(int output_id, int count, Buffer &out) {
        //std::cout << "calculate(): count=" << count << std::endl;

        /* 先読みが必要なので、head_margin_frames フレーム追加で読み込んでおく */
        this->receive(count + this->head_margin_frames);

        /* prepare output data buffer */
        /* - sources */
        RCPtr<Vector<ObjectRef> > sources(new Vector<ObjectRef>);

        /* - pcm data */
        RCPtr<Map<int, ObjectRef> > outputp(new Map<int, ObjectRef>());

        int channel_index;
        for(channel_index = 0; channel_index < channel_count; channel_index++){
            ChannelWorkset &channel_workset = this->channel_worksets[channel_index];

            int zc = 0; //debug
            bool source_exists = false;
            if((zc = channel_workset.zerocross_counter.getZeroCrossCount()) > this->zerocross_threshold_count){
                // output frames
                channel_workset.last_active_frame_index = count;
                source_exists = true;

            }
            else if(count < channel_workset.last_active_frame_index + this->tail_margin_frames){
                // output frames
                source_exists = true;

            }
            else{
                // no need to output
            }

            //std::cout << "source-exists[" << channel_index << "]: " << source_exists << ", zc=" << zc << "/" << this->zerocross_threshold_count << std::endl; //debug

            map<int, Source>::iterator it = this->channel_sources.find(channel_index);
            if(source_exists){
                RCPtr<Source> source(new Source);

                if(it != this->channel_sources.end()){
                    *source = it->second;
                }
                else{
                    int source_id = ++this->source_count;
                    source->x[0] = 0;
                    source->x[1] = 0;
                    source->x[2] = 0;
                    source->power = 0;
                    source->remaining = 0;
                    source->id = source_id * this->channel_count + channel_index;
                    this->channel_sources[channel_index] = *source;
                }
                source->count = count;
                source->last_found_count = count;

                /* target data to output */
                RCPtr<Vector<float> > outputVector(new Vector<float>(this->length));
                vector<float> &targetBuffer = channel_workset.buffer[count % this->head_margin_frames];
                (*outputVector).assign(targetBuffer.begin(), targetBuffer.end());
                (*sources).push_back(source);
                (*outputp)[source->id] = outputVector;
            }
            else{
                if(it != this->channel_sources.end()){
                    this->channel_sources.erase(it);
                }
            }
        }
        (*(this->outputs[this->sourcesOutputID].buffer))[count] = sources;
        (*(this->outputs[this->audioOutputID].buffer))[count] = outputp;

        //std::cout << "count: " << count << ", data="  << (*this->outputs[this->audioOutputID].buffer)[count] << std::endl;
    }

    IN_ORDER_NODE_SPEEDUP (VAD_ZC);
};
